'From Squeak3.4gamma of ''7 January 2003'' [latest update: #5169] on 9 February 2003 at 10:47:38 pm'!
"Change Set:		Closures-ajh
Date:			9 February 2003
Author:			Anthony Hannan (ajh18@cornell.edu)

This adds a new parser and parse nodes to the compiler that generate methods that create closures for blocks.  No new bytecodes are needed.  These new parse nodes are used when 'Preferences compileBlocksAsClosures' is true.  Etoys scripts/tiles do not work yet with the new parse nodes, so make sure this preference is false when working with them.

PostContextEnh-ajh must be filed in before this.
"!

Object variableSubclass: #BlockClosure
	instanceVariableNames: 'method '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Methods'!

!BlockClosure commentStamp: '<historical>' prior: 0!
A BlockClosure is a block of Smalltalk code (enclosed within []) that may be executed later by sending #valueWithArguments: (or one of its variants) to it.  A block can take arguments by specifying the names of the arguments in the beginning of the block, as in "[:arg1 :arg2 | ...]", and can have its own local temps, as in "[:arg1 |  | temp1 temp2 | ...]".  The block may reference variables outside its scope directly by name.  It also may return from its home context by using ^, otherwise, the value of the last statement is returned to the sender of valueWithArguments:.

Structure:

 method		CompiledMethod2
			Contains the block's code.  It has its own method separate from its home method.

 <indexable fields>  Objects
			Contains values of variables referenced outside its scope.  If a variable can change while being shared then it is wrapped in a SharedTempHolder, and the home context and each block references the holder and make changes to its contents.  If the block remotely returns (^) then its home context is also held in these fields.  If the block references self or one of its instance variables then self is also held in these fields.  See VarUse in the Compiler for how shared temps are determined.

See comment in class method #aboutClosures for more.
!

Exception subclass: #BlockNodeSignal
	instanceVariableNames: 'blockNode '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Translate'!
Link subclass: #BytecodeInstruction
	instanceVariableNames: 'owningSequence localSpBefore localBytecodeRange '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!

!BytecodeInstruction commentStamp: '<historical>' prior: 0!
My subclasses represent single abstract bytecode instructions like pushTemp: or sendLiteral:.!

WriteStream subclass: #BytecodeV3WriteStream
	instanceVariableNames: ''
	classVariableNames: 'BytecodeTable Bytecodes SpecialConstants SpecialReturnConstants SpecialSelectors '
	poolDictionaries: ''
	category: 'Compiler-Translate'!
Inspector subclass: #ClosureInspector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tools-Debugger'!

!Compiler commentStamp: '<historical>' prior: 0!
The compiler accepts Smalltalk source code and compiles it with respect to a given class. The user of the compiler may supply a context so that temporary variables are accessible during compilation. If there is an error, a requestor (usually a kind of StringHolderController) is sent the message notify:at:in: so that the error message can be displayed. If there is no error, then the result of compilation is a MethodNode2, which is the root of a parse tree whose nodes are kinds of ParseNode2s. The parse tree can be sent messages to (1) generate code for a CompiledMethod2 (this is done for compiling methods or evaluating expressions); (2) pretty-print the code (for formatting); or (3) produce a map from object code back to source code (used by debugger program-counter selection).

If "Preferences compileBlocksAsClosures" is true then compiling source code to bytecodes is done in four phases:

 Phase 1 - Parse source text into a parse tree.
	The Scanner tokenizes the text while its subclass Parser2 interprets these tokens into composite ParseNode2s resulting in a tree with MethodNode2 at the root.

 Phase 2 - Analyze var usage.
	#calcVarUsage: is sent to each node, accumulating the effect of nodes on each temporary variable.  Those that are referenced by embedded blocks (that are not inlined) and are changed later are marked #indirect.  When translated (next phase), indirect temps are initialized to SharedTempHolders and emit indirect instruction references to get at the holder's contents.  Each temp var, also, remembers when it is first used, so it can be initialized on first use, instead of all at the beginning.

 Phase 3 - Translate the parse tree into stack instructions.
	Sending #generateIR to a MethodNode2 causes #emit:forValue: to be sent to each node in the tree with an InstructionBuilder as its first argument.  Each node sends the appropriate stack-like message to the MethodBuilder, causing a sequence of StackInstructions to be built up as the parse tree is traversed.  Instructions are grouped into InstructionSequences (basic blocks).

 Phase 4 - Generate final bytecodes.
	Sending #compiledMethod to an InstructionBuilder causes each StackInstruction to write itself out as bytes via BytecodeWriteStream.  The result is a CompiledMethod2, which can be executed or installed in a class.

ajh 10/15/2002!

BytecodeInstruction subclass: #CreateBlockInstr
	instanceVariableNames: 'closureSize selector isSpecial '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
Decompiler subclass: #Decompiler2
	instanceVariableNames: 'outer keepNodesBasic '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Decompile'!
ParseNode subclass: #DecompilerConstructor2
	instanceVariableNames: 'tempNames capturedVars classEnv methodNode decompiler tempVarScopeDict '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Decompile'!
BytecodeInstruction subclass: #DupInstr
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
Object subclass: #FiniteAutomaton
	instanceVariableNames: 'states actions transitions '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Translate'!

!FiniteAutomaton commentStamp: '<historical>' prior: 0!
I represent a deterministic finite automaton (or finite state machine), with states, input actions, and transition rules.  See users for an example.!

Object subclass: #InstrTempVar
	instanceVariableNames: 'key fpOffset '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Translate'!

!InstrTempVar commentStamp: '<historical>' prior: 0!
I represent a named slot in the context stack.  See InstructionBuilder declareAsTemp:, pushTemp:, and storeTemp:.!

Object subclass: #InstructionBuilder
	instanceVariableNames: 'tempVars startSequence literals lastLiteral instrs jumpBackTargetStacks jumpAheadStacks nextSequenceNumber primitiveNode declareInstrs mapToObject rargTemps otherTemps contextVars '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Translate'!

!InstructionBuilder commentStamp: 'ajh 2/6/2003 13:14' prior: 0!
I provide a simple abstract bytecode interface for creating BytecodeInstructions.  These instructions can then be translated into a CompiledMethod.  For example, to create instructions that switches x values of two points (the receiver and the arg) and returns 'equal' if equal, do:

	InstructionBuilder new
		rargs: #(self otherPoint);		"receiver and args declarations"
		addTemps: #(z);					"temp declariations"
		pushTemp: #self;
		getField: 1;
		storeTemp: #z;						"z _ x."
		pushTemp: #otherPoint;
		send: #x;
		pushTemp: #self;
		setField: 1;						"x _ otherPoint x"
		doPop;
		pushTemp: #otherPoint;
		pushTemp: #z;
		send: #setX:;						"otherPoint setX: z"
		doPop;
		pushTemp: #z;
		pushTemp: #self;
		getField: 1;
		send: #=;						"z = x"
		jumpAheadTo: #notEqual if: false;
		pushLiteral: 'equal';				"ifTrue: ['equal']"
		jumpAheadTo: #end;
		jumpAheadTarget: #notEqual;
		pushLiteral: 'not equal';			"ifFalse: ['not equal']"
		jumpAheadTarget: #end;
		localReturnTop;
		yourself

To see a printout of the instructions, send #longPrintString to the resulting instructionBuilder.  To convert the instructions to a CompiledMethod, send #compiledMethodWith: #(0) to it.
!

LinkedList subclass: #InstructionSequence
	instanceVariableNames: 'method localParseStack parseStackIn parseStackOut bytecodesWithoutJump jumpBytecodes startBytecodePosition orderNumber '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Translate'!

!InstructionSequence commentStamp: '<historical>' prior: 0!
I hold a sequence of BytecodeInstructions where only the last instruction jumps or returns, the rest are guaranteed to execute in order (ie. a basic block).

Structure:

 method			InstructionBuilder
				The owning instruction builder, used to get at it literals list.

 localParseStack ParseStack
				Stack effect as if my instructions were the only instructions in the method.

 parseStackIn	ParseStack
				Cummulative stack effect of all instructions that may execute before me.

 parseStackIn	ParseStack
				Cummulative stack including my instructions.

 bytecodesWithoutJump  ByteArray
				Generated bytecodes for all my instructions instead of the last one if its a jump, because the jump bytecodes can't be calculated until all other sequences' bytecodes are calculated.

 jumpBytecodes	ByteArray
				Generated bytecodes for my last instruction if it is a jump.  Calculated after all other bytecodes are generated.

 startBytecodePosition  SmallInteger
				Bytecode index of my first bytecode.

 orderNumber	SmallInteger
				Relative position of this sequence with respect to all other sequences when generated.

Cummulative stack effect is calculated by aggregating instructionSequence effects into successor instructionSequence effects and repeating this until the cummulative effects no longer change (see InstructionBuilder aggregateParseStack).
!

ObjectMemory subclass: #Interpreter
	instanceVariableNames: 'activeContext theHomeContext method receiver instructionPointer stackPointer localIP localSP localHomeContext messageSelector argumentCount newMethod currentBytecode successFlag primitiveIndex methodCache atCache lkupClass reclaimableContextCount nextPollTick nextWakeupTick lastTick interruptKeycode interruptPending semaphoresToSignalA semaphoresUseBufferA semaphoresToSignalCountA semaphoresToSignalB semaphoresToSignalCountB savedWindowSize fullScreenFlag deferDisplayUpdates pendingFinalizationSignals compilerInitialized compilerHooks extraVMMemory newNativeMethod methodClass receiverClass interpreterVersion obsoleteIndexedPrimitiveTable obsoleteNamedPrimitiveTable interpreterProxy showSurfaceFn interruptCheckCounterFeedBackReset interruptChecksEveryNms externalPrimitiveTable '
	classVariableNames: 'ActiveProcessIndex AtCacheEntries AtCacheFixedFields AtCacheFmt AtCacheMask AtCacheOop AtCacheSize AtCacheTotalSize AtPutBase BlockArgumentCountIndex BlockMethodIndex BytecodeTable CacheProbeMax CallerIndex CharacterValueIndex CompilerHooksSize CrossedX DirBadPath DirEntryFound DirNoMoreEntries EndOfRun ExcessSignalsIndex FirstLinkIndex GenerateBrowserPlugin HeaderIndex HomeIndex InitialIPIndex InstanceSpecificationIndex InstructionPointerIndex JitterTable LastLinkIndex LiteralStart MaxExternalPrimitiveTableSize MaxPrimitiveIndex MessageArgumentsIndex MessageDictionaryIndex MessageLookupClassIndex MessageSelectorIndex MethodArrayIndex MethodCacheClass MethodCacheEntries MethodCacheEntrySize MethodCacheMask MethodCacheMethod MethodCacheNative MethodCachePrim MethodCacheSelector MethodCacheSize MethodIndex MillisecondClockMask MyListIndex NextLinkIndex PrimitiveExternalCallIndex PrimitiveTable PriorityIndex ProcessListsIndex ReceiverIndex SelectorStart SemaphoresToSignalSize SenderIndex StackPointerIndex StreamArrayIndex StreamIndexIndex StreamReadLimitIndex StreamWriteLimitIndex SuperclassIndex SuspendedContextIndex TempFrameStart ValueIndex XIndex YIndex '
	poolDictionaries: ''
	category: 'VMConstruction-Interpreter'!
BytecodeInstruction subclass: #JumpInstr
	instanceVariableNames: 'destinationSequence '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
JumpInstr subclass: #JumpIfInstr
	instanceVariableNames: 'jumpCondition otherwiseSequence '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
OrderedCollection subclass: #LiteralList
	instanceVariableNames: 'equalitySet '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Translate'!
Set subclass: #LiteralSet
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Translate'!

!LiteralSet commentStamp: '<historical>' prior: 0!
Literal objects are equal if they are #= plus they are the same class.  This set uses this rule for finding elements.  For example, #tony = 'tony' but they are not "literallyEqual".

	Set new add: 'tony'; add: #tony; yourself  "=> size = 1"
	LiteralSet new add: 'tony'; add: #tony; yourself  "=> size = 2"
!

BytecodeInstruction subclass: #LocalReturnInstruction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
LocalReturnInstruction subclass: #LocalReturnSelfInstr
	instanceVariableNames: 'selfVar '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
LocalReturnInstruction subclass: #LocalReturnSpecialConstantInstr
	instanceVariableNames: 'constant isSpecial isByte '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!

!LocalReturnSpecialConstantInstr commentStamp: 'ajh 1/15/2003 16:01' prior: 0!
A PushConstantInstr changes its class to LocalReturnSpecialConstantInstr if a local return immediately follows it and it contains a special constant (nil, false, true, -1, 0, 1, 2), so inst var schema has to match.!

LocalReturnInstruction subclass: #LocalReturnTopInstr
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
Object subclass: #MethodClosure
	instanceVariableNames: 'receiver method '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Contexts'!

!MethodClosure commentStamp: 'ajh 2/8/2003 23:14' prior: 0!
A MethodClosure is a method and receiver ready to be executed.  It is analogous to a MessageSend except the method is directly supplied instead of its selector.  This is useful for executing methods without installing them in the method dictionary (do-its).  It is called a closure because, like a BlockClosure, a MethodClosure holds the values (the receiver) it will need to execute.!

OrderedCollection subclass: #OrderedLiterals
	instanceVariableNames: 'equalitySet '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Translate'!

!OrderedLiterals commentStamp: '<historical>' prior: 0!
I order literals so they can be optimally positioned for compact bytecodes (sendLiteral0Range, sendLiteral1Range, etc).  I do not order unnecessarily so the literals can preserve as much as their access order as possible.  I can be configured by modifying the methods in 'configuration' category.

Test:
	| litList |
	litList _ OrderedLiterals new.
	litList addAll: #(odd + 'Hello' - 'World' 4 even add: at:ifAbsent: 5 replaceFrom:with:to: 6 at:ifPresent: jumpTo:if:otherwise: jumpTo:if:).
	litList
!

Object subclass: #ParseEnv
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Parse'!

!ParseEnv commentStamp: '<historical>' prior: 0!
The Parser2 looks up non-local variables (and some day selectors) in the ParseEnv held by the MethodNode2 in its classEnv field.  The interface is:
	lookupVar: name
	possibleVarsFor: name continued: list
	classEncoding
	instVarNodeAt: index (for decompiling only)
!

ParseEnv subclass: #ClassEnv
	instanceVariableNames: 'receiverVar class '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Parse'!

!ClassEnv commentStamp: '<historical>' prior: 0!
I lookup a class's instance and pool variables, include global variables.!

ParseEnv subclass: #DoItClosureEnv
	instanceVariableNames: 'receiverVar closure classEnv '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Parse'!

!DoItClosureEnv commentStamp: '<historical>' prior: 0!
I lookup a BlockClosure's free variables by name, and delegate to a ClassEnv if 'self' is one of my free variables.!

ParseEnv subclass: #DoItContextEnv
	instanceVariableNames: 'receiverVar context receiverEnv '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Parse'!

!DoItContextEnv commentStamp: '<historical>' prior: 0!
I lookup a MethodContext's temporary variables by name, and delegate to a ClassEnv for the receiver's variables or a DoItClosureEnv if I am executing a block.!

Error subclass: #ParseError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Parse'!

!ParseError commentStamp: '<historical>' prior: 0!
I signal bad syntax during parsing.!

Object subclass: #ParseNode2
	instanceVariableNames: 'comment sourceRange bytecodeRange '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!ParseNode2 commentStamp: '<historical>' prior: 0!
I am the abstract class of all parse tree nodes.  All ParseNodes are interchangeable with each other meaning they are all expression nodes.  VarNodes are not part of this hierarchy, VarReferenceNodes and VarAssignmentNodes are used instead.  They wrap the varNode and provide separate points of reference (source range and bytecodeRange).!

ParseNode2 subclass: #ArrayNode
	instanceVariableNames: 'elementNodes '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!ArrayNode commentStamp: '<historical>' prior: 0!
I represent the brace construct: {...}

I compile into either a fast short form for 4 elements or less:
	Array braceWith: a with: b ... 
or a long form of indefinite length:
	(Array braceStream: N) nextPut: a; nextPut: b; ...; braceArray.
!

ParseNode2 subclass: #BlockNode2
	instanceVariableNames: 'receiverTempVar argTempVars thisContextTempVar otherTempVars statementNodes primitive selector varUsage ir compiledMethod sourceText '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!BlockNode2 commentStamp: '<historical>' prior: 0!
I represent a block or a method.  I have my own scope with my own args, temporary variables, and statements, but also with access to my outer scopes's variables.  The value of my last statement is the value of myself as a whole.  A return (^) in one of my statements will cause execution to return out to the sender of my outer most block (method).!

ParseNode2 subclass: #CascadeMessageNode
	instanceVariableNames: 'receiverNode messageNodes '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!CascadeMessageNode commentStamp: '<historical>' prior: 0!
I represent the cascading messages construct:
	<rcvr> <message1>; <message2>; ... <messageN>

I hold the receiver while my messages have nil receivers.!

ParseNode2 subclass: #ConstantNode
	instanceVariableNames: 'object '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!ConstantNode commentStamp: '<historical>' prior: 0!
I represent a literal constant such as 'a' or 2.!

BlockNode2 subclass: #DialectMethodNode2
	instanceVariableNames: 'dialect '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Parse'!

!DialectMethodNode2 commentStamp: '<historical>' prior: 0!
The purpose of this class is to carry along with the information in a regular method node the further information that it was parsed from an alternate dialect of Squeak.  Which dialect that was is carried as a symbol in the dialect variable.!

ParseNode2 subclass: #DummyJumpBack
	instanceVariableNames: 'destinationIp '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Decompile'!
ParseNode2 subclass: #DummyJumpForward
	instanceVariableNames: 'destinationIp '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Decompile'!
ParseNode2 subclass: #MessageNode2
	instanceVariableNames: 'receiverNode selector argumentNodes '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!MessageNode2 commentStamp: '<historical>' prior: 0!
I represent a message send composed of the receiver, selector, and arguments.  If I represent one of the special inlined message sends (eg. ifTrue:), I calculate var usage and emit code specially.!

ParseNode2 subclass: #NoopNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Decompile'!
Object subclass: #ParseStack
	instanceVariableNames: 'position length '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Translate'!
BytecodeInstruction subclass: #PopInstr
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
Object subclass: #PrimitiveNode
	instanceVariableNames: 'primitiveNum spec '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!PrimitiveNode commentStamp: '<historical>' prior: 0!
I represent a primitive.  I am more than just a number if I am a named primitive.!

Object subclass: #PseudoPoolVariable
	instanceVariableNames: 'name getterBlock setterBlock '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Parse'!

!PseudoPoolVariable commentStamp: '<historical>' prior: 0!
The values of pool and global variables (traditionally Associations) are fetched by sending #poolValue and set by sending #setInPool: which send #poolValue:.  These sends are automatically added in by the Compiler (see PoolVarNode {code generation}).  So any object can act like a pool variable.  This class allows getter and setter blocks for poolValue and poolValue:.!

BytecodeInstruction subclass: #PushConstantInstr
	instanceVariableNames: 'constant isSpecial isByte '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
BytecodeInstruction subclass: #PushInstVarInstr
	instanceVariableNames: 'instVarIndex '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
BytecodeInstruction subclass: #PushThisContextInstr
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
BytecodeInstruction subclass: #RemoteReturnTopInstr
	instanceVariableNames: 'isSpecial '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
ParseNode2 subclass: #ReturnNode2
	instanceVariableNames: 'valueNode contextVar '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!ReturnNode2 commentStamp: '<historical>' prior: 0!
I represent the return construct:
	^ <expression>

I explicitly hold the context variable of the block/method node I will return out of.!

Object subclass: #Scanner
	instanceVariableNames: 'source mark hereChar aheadChar token tokenType currentComment buffer typeTable '
	classVariableNames: 'TypeTable '
	poolDictionaries: ''
	category: 'Compiler-Parse'!
Scanner subclass: #Parser2
	instanceVariableNames: 'here hereType hereMark hereEnd prevMark prevEnd requestor parseNode failBlock requestorOffset tempsMark doitFlag nestedBlockNodes homeBlockNode parseEnv '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Parse'!

!Parser2 commentStamp: '<historical>' prior: 0!
I parse Smalltalk syntax tokens fed to me by my superclass, and create a parse tree with MethodNode2 at its root.  I look one token ahead.  I look up non-local variable names in the ParseEnv supplied to me.!

Parser2 subclass: #DialectParser2
	instanceVariableNames: 'dialect '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Parse'!

!DialectParser2 commentStamp: '<historical>' prior: 0!
This Parser is part of a package designed to allow for experiments with alternatives to ST-80 syntax.  The particular alternative offered by this parser eliminates colons, left-arrows and up-arrows, and adds prefix keywords for common control constructs.

ST-80									SQ-00
-------									-------
a + b between: c and: c + 4				a + b between (c) and (c + 4)
a _ 3.0									Set a to 3.0
^ self size + 3							Return self size + 3
a > b									Test (a > b)
	ifTrue: ['greater']						ifTrue ['greater']
	ifFalse: ['less']							ifFalse ['less']
1 to: 9 do:								Repeat (1) to (9) do
	[:i | Transcript cr; show: i]				[Set i. | Transcript cr; show (i)]

The use of prefix keywords is currently ad-hoc;  in other words they are built into the parser, and there is not a way to define a method pattern to include a prefix keyword.  Most of the work has been done to support this, though, as selectors can now have the form
	#:prefix:kwd1:kwd2:
and they will respond appropriately to #keywords and #numArgs.

A test method in the class ensures that every method in the system can be pretty-printed in the alternative syntax, and that compiling the resulting text produces exactly the same bytecodes as the original method.!

BytecodeInstruction subclass: #SendSelectorInstr
	instanceVariableNames: 'selector isSpecial '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
SendSelectorInstr subclass: #SendSelectorSuperInstr
	instanceVariableNames: 'behavior '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
Object subclass: #SharedTempHolder
	instanceVariableNames: 'value '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Contexts'!

!SharedTempHolder commentStamp: '<historical>' prior: 0!
Holds a changing temp that is shared among BlockClosures/homeContext.  The home method and block method(s) know to access the var indirectly through this holder.  See Compiler and VarUse for more on when these are used.

Structure:
 value		Object	- holds my value that may change!

BytecodeInstruction subclass: #StoreInstVarInstr
	instanceVariableNames: 'instVarIndex pop '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
BytecodeInstruction subclass: #TempVarInstruction
	instanceVariableNames: 'tempVar '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
TempVarInstruction subclass: #DeclareTempInstr
	instanceVariableNames: 'deallocInstr '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
TempVarInstruction subclass: #PushTempInstr
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
TempVarInstruction subclass: #StoreTempInstr
	instanceVariableNames: 'pop '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Instructions'!
Warning subclass: #UndefinedVarSignal
	instanceVariableNames: 'refNode '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Parse'!
ParseNode2 subclass: #VarAssignmentNode
	instanceVariableNames: 'var valueNode '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!VarAssignmentNode commentStamp: '<historical>' prior: 0!
I represent a (var _ expr) construct.!

Object subclass: #VarNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!VarNode commentStamp: '<historical>' prior: 0!
I am the abstract class for all variables nodes.  I am separate from the ParseNode hierarchy, and only get included in the parse tree via VarReferenceNode or VarAssignmentNode.!

VarNode subclass: #InstVarNode
	instanceVariableNames: 'instVarIndex name selfVar '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!InstVarNode commentStamp: '<historical>' prior: 0!
I represent an instance variable of 'self'.!

VarNode subclass: #MessageVarNode
	instanceVariableNames: 'name getMessageNode setMessageNode '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!MessageVarNode commentStamp: '<historical>' prior: 0!
I represent a variable that must get and set it value via a message send.  This is used for do-it expressions to get temp or closure values from an existing context or closure.!

VarNode subclass: #PoolVarNode
	instanceVariableNames: 'poolVariable '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!PoolVarNode commentStamp: '<historical>' prior: 0!
I represent a class, pool, or global variable.!

VarNode subclass: #TempVarNode
	instanceVariableNames: 'owningBlockNode name '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!TempVarNode commentStamp: '<historical>' prior: 0!
I represent a temporary variable (including args, receiver, and thisContext) of a block/method node.!

ParseNode2 subclass: #VarReferenceNode
	instanceVariableNames: 'var '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!VarReferenceNode commentStamp: '<historical>' prior: 0!
I represent a variable reference (just the name of the variable in place of an expression in Smalltalk).!

VarReferenceNode subclass: #DummyTempRefNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!
VarReferenceNode subclass: #SuperNode
	instanceVariableNames: 'varClass '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Nodes'!

!SuperNode commentStamp: '<historical>' prior: 0!
I represent a special variable reference to 'self' called 'super', used when the message sent to it is meant to be looked up in the superclass of the class the sending method is defined in (the sending method class is held in my varClass).!

OrderedCollection subclass: #VarUsageSet
	instanceVariableNames: 'blockNode currentBlockStatement '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Translate'!

!VarUsageSet commentStamp: 'ajh 1/15/2003 13:48' prior: 0!
I hold an ordered collection of varUses for a blockNode, one for each var it touches.  The order is in first seen order and determines the order of temp vars in the context stack.!

Object subclass: #VarUse
	instanceVariableNames: 'var useState firstUse firstUseStatement lastUse lastUseStatement '
	classVariableNames: 'AddUseAutomaton MergeUseAutomaton '
	poolDictionaries: ''
	category: 'Compiler-Translate'!

!VarUse commentStamp: '<historical>' prior: 0!
I remember the cummulative effect that nodes of a BlockNode have on a TempVarNode.  The resulting effect can be one of the following:

	#none			not read or written.
	#read			read by local nodes but never written.
	#write			written by local nodes, may have also been read by local nodes.
	#readBlock		read by embedded block closure, may have also been read by local nodes, but never written.
	#wrRdBlock		written by local nodes first then read by embedded block closure(s).
	#writeBlock		written by embedded block closure, may have also been read by local nodes earlier.
	#indirect		written by embedded block closure then read or written by local nodes or another embedded block closure, or read by embedded block closure then written by local nodes or another embedded block closure.

#indirect indicates that the variable needs to be put in a SharedTempHolder and accessed indirectly.

My AddUseAutomaton (a FiniteAutomaton) encodes the state transitions described above.  My MergeUseAutomaton (another FiniteAutomaton) has the same states but defines how to merge var effects from two branches that come together.  It basically chooses the most progressed state except for a couple of cases where an even more progressed state is chosen.  See my class #initialize method for these state machine definitions.

I also remember the first node that access my var.  This is used to generate temp initialization at the time it is first used instead of all in the beginning.  I also remember the last node that accesses my var.  This is not used now, but can be used in the future to nil out or pop temp slots that will no longer be used.  This would be advantageous if we want to release objects early instead of waiting until the method returns.
!


!ProtoObject methodsFor: 'testing' stamp: 'ajh 12/4/2001 12:44'!
ifNotNil: ifNotNilBlock
	"Evaluate the block, unless I'm == nil (q.v.)"

	^ ifNotNilBlock valueWithPossibleArgs: {self}! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 2/3/2003 23:44'!
getInstVarAt: instVarIndex
	"The compiler generates this message when assigning to instance variables of objects that have been captured by a block closure.  Do not override or change this unless you know what your doing"

	^ self instVarAt: instVarIndex! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 1/16/2003 01:25'!
remoteReturnTo: contextTag
	"Generated from compiler to return to the home of a block.  self is the value return and contextTag is the tag of the home context.  Do not change or override this unless you know what you are doing."

	contextTag return: self! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 1/18/2003 13:17'!
setInPool: poolVariable
	"The compiler generates this message to place values in global variables.  Do not override or change this unless you know what your doing"

	poolVariable poolValue: self.
	^ self! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 1/15/2003 21:24'!
storeIn: object instVar: instVarIndex
	"The compiler generates this message when assigning to instance variables of objects that have been captured by a block closure.  Do not override or change this unless you know what your doing"

	object instVarAt: instVarIndex put: self! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 1/15/2003 21:23'!
storeInTempHolder: sharedTempHolder
	"The compiler generates this message to place temp values in holders so they can be shared amongst closures.  Do not override or change this unless you know what your doing"

	sharedTempHolder tempValue: self! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 1/15/2003 21:23'!
wrapInTempHolder
	"The compiler generates this message to place temp values in holders so they can be shared amongst closures.  Do not override or change this unless you know what your doing"

	| holder |
	holder _ SharedTempHolder new.
	holder value: self.
	^ holder! !


!Object methodsFor: 'binding' stamp: 'ajh 2/6/2003 13:24'!
hasLiteralSuchThat: testBlock
	"This is the end of the imbedded structure path so return false."

	^ false! !

!Object methodsFor: 'binding' stamp: 'ajh 1/31/2003 12:04'!
hasLiteralThorough: literal
	"Answer true if literal is identical to any literal in this array, even if imbedded in further structures.  This is the end of the imbedded structure path so return false."

	^ false! !


!Array methodsFor: 'private' stamp: 'ajh 2/6/2003 13:22'!
hasLiteralSuchThat: testBlock

	| lit |
	1 to: self size do: [:index |
		(testBlock value: (lit _ self at: index)) ifTrue: [^ true].
		(lit hasLiteralSuchThat: testBlock) ifTrue: [^ true]].
	^ false! !

!Array methodsFor: 'private' stamp: 'ajh 1/31/2003 12:02'!
hasLiteralThorough: literal
	"Answer true if literal is identical to any literal in this array, even if imbedded in further array structures or closure methods"

	| lit |
	1 to: self size do: [:index |
		(lit _ self at: index) == literal ifTrue: [^ true].
		(lit hasLiteralThorough: literal) ifTrue: [^ true]].
	^ false! !


!Behavior methodsFor: 'accessing' stamp: 'ajh 2/7/2003 20:17'!
decompilerClass
	"Answer a decompiler class appropriate for compiled methods of this class."

	^ Preferences compileBlocksAsClosures
		ifTrue: [Decompiler2]
		ifFalse: [Decompiler]! !

!Behavior methodsFor: 'testing method dictionary' stamp: 'ajh 2/3/2003 17:40'!
thoroughWhichSelectorsReferTo: literal special: specialFlag byte: specialByte
	"Answer a set of selectors whose methods access the argument as a 
	literal. Dives into the compact literal notation, making it slow but 
	thorough "

	| who |
	who _ Set new.
	self selectorsAndMethodsDo:
		[:sel :method |
		((method hasLiteralThorough: literal) or: [specialFlag and: [method scanFor: specialByte]])
			ifTrue:
				[((literal isVariableBinding) not
					or: [method sendsToSuper not
					or: [method literals last ~= literal]])
						ifTrue: [who add: sel]]].
	^ who! !

!Behavior methodsFor: 'testing method dictionary' stamp: 'ajh 2/3/2003 17:36'!
whichSelectorsReferTo: literal special: specialFlag byte: specialByte
	"Answer a set of selectors whose methods access the argument as a literal."

	| who |
	who _ Set new.
	self selectorsAndMethodsDo: 
		[:sel :method |
		((method hasLiteralThorough: literal) or: [specialFlag and: [method scanFor: specialByte]])
			ifTrue:
				[((literal isVariableBinding) not
					or: [method sendsToSuper not
					or: [method literals last ~= literal]])
						ifTrue: [who add: sel]]].
	^ who! !

!Behavior methodsFor: 'compiling' stamp: 'ajh 10/15/2002 04:10'!
parseEnv

	^ ClassEnv new class: self! !


!BlockClosure methodsFor: 'initializing' stamp: 'ajh 5/28/2001 18:37'!
method: compiledMethod
	"compiledMethod will be the code I execute when I'm evaluated"

	method _ compiledMethod! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 1/16/2002 18:32'!
copyForSaving
	"obsolete"! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 1/16/2002 18:32'!
fixTemps
	"obsolete"! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 1/31/2003 12:53'!
reentrant! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 7/15/2001 16:13'!
valueError

	self error: 'Incompatible number of args'! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 7/26/2002 11:47'!
valueUnpreemptively
	"Evaluate the receiver (block), without the possibility of preemption by higher priority processes. Use this facility VERY sparingly!!"
	"Think about using Block>>valueUninterruptably first, and think about using Semaphore>>critical: before that, and think about redesigning your application even before that!! 
	After you've done all that thinking, go right ahead and use it..."

	| activeProcess oldPriority result |
	activeProcess _ Processor activeProcess.
	oldPriority _ activeProcess priority.
	activeProcess priority: Processor highestPriority.
	result _ self ensure: [activeProcess priority: oldPriority].
	"Yield after restoring priority to give the preempted processes a chance to run"
	Processor yield.
	^result! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 8/9/2002 13:11'!
veryDeepInner: deepCopier
	"Do not copy my method (which can be shared because CompiledMethod2 are basically treated as immutables) or my home context (MethodContexts are treated as immutables too)"

	super veryDeepInner: deepCopier.
	method _ method.
! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/13/2002 13:04'!
ifError: errorHandlerBlock
	"Evaluate the block represented by the receiver, and normally return it's value.  If an error occurs, the errorHandlerBlock is evaluated, and it's value is instead returned.  The errorHandlerBlock must accept zero, one, or two parameters (the error message and the receiver)."
	"Examples:
		[1 whatsUpDoc] ifError: [:err :rcvr | 'huh?'].
		[1 / 0] ifError: [:err :rcvr |
			'ZeroDivide' = err
				ifTrue: [Float infinity]
				ifFalse: [self error: err]]
"

	^ self on: Error do: [:ex |
		errorHandlerBlock valueWithPossibleArgs: {ex description. ex receiver}]! !

!BlockClosure methodsFor: 'evaluating'!
pvtEXECUTE
	^ t1! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 7/15/2001 15:57'!
timeToRun
	"Answer the number of milliseconds taken to execute this block."

	^ Time millisecondsToRun: self
! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:22'!
value
	"Primitive. Evaluate the block represented by the receiver. Fail if the 
	block expects any arguments or if the block is already being executed. 
	Optional. No Lookup. See Object documentation whatIsAPrimitive."

	<primitive: 186>
	^ self valueWithArguments: #()! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:22'!
value: arg 
	"Primitive. Evaluate the block represented by the receiver. Fail if the 
	block expects other than one argument or if the block is already being 
	executed. Optional. No Lookup. See Object documentation 
	whatIsAPrimitive."

	<primitive: 186>
	^ self valueWithArguments: (Array with: arg)! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:26'!
value: arg1 value: arg2 
	"Primitive. Evaluate the block represented by the receiver. Fail if the 
	block expects other than two arguments or if the block is already being 
	executed. Optional. See Object documentation whatIsAPrimitive."

	<primitive: 186>
	^ self valueWithArguments: (Array with: arg1 with: arg2)! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:26'!
value: arg1 value: arg2 value: arg3 
	"Primitive. Evaluate the block represented by the receiver. Fail if the 
	block expects other than three arguments or if the block is already being 
	executed. Optional. See Object documentation whatIsAPrimitive."

	<primitive: 186>
	^ self valueWithArguments: 
		(Array
			with: arg1
			with: arg2
			with: arg3)! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:26'!
value: arg1 value: arg2 value: arg3 value: arg4 
	"Primitive. Evaluate the block represented by the receiver. Fail if the 
	block expects other than three arguments or if the block is already being 
	executed. Optional. See Object documentation whatIsAPrimitive."

	<primitive: 186>
	^ self valueWithArguments: 
		(Array
			with: arg1
			with: arg2
			with: arg3
			with: arg4)! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:36'!
valueWithArguments: anArray 
	"Primitive. Evaluate the block represented by the receiver. The argument 
	is an Array whose elements are the arguments for the block. Fail if the 
	length of the Array is not the same as the the number of arguments that 
	the block was expecting. Fail if the block is already being executed. 
	Essential. See Object documentation whatIsAPrimitive."

	| dict index |
	<primitive: 187>
	#pvtEXECUTE flushCache.
	dict _ self class methodDict.
	index _ dict findElementOrNil: #pvtEXECUTE.
	dict methodArray at: index put: method.
	^ self perform: #pvtEXECUTE withArguments: anArray

	"Do not remove #pvtEXECUTE method just replace it and flush method cache.  Also avoid using an ensure block since that takes time also.  The primitive, if available, will execute the method directly without flushing the cache and adding to the method dictionary"
! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 14:44'!
valueWithPossibleArgs: anArray 

	| n |
	(n _ self numArgs) = 0 ifTrue: [^ self value].
	n = anArray size ifTrue: [^ self valueWithArguments: anArray].
	^ self valueWithArguments: (n > anArray size
		ifTrue: [anArray, (Array new: n - anArray size)]
		ifFalse: [anArray copyFrom: 1 to: n])! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 2/6/2003 13:24'!
hasLiteralSuchThat: testBlock

	(testBlock value: method) ifTrue: [^ true].
	^ method hasLiteralSuchThat: testBlock! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 1/31/2003 16:59'!
hasLiteralThorough: literal
	"Answer true if literal is identical to any literal imbedded in my method"

	method == literal ifTrue: [^ true].
	^ method hasLiteralThorough: literal! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 1/21/2003 13:16'!
isBlock

	^ true! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 1/28/2003 22:19'!
longPrintOn: stream

	self method blockNode printOn: stream indent: 0.
	stream cr.
	self size = 0 ifTrue: [^ self].
	stream cr; nextPut: $".
	stream nextPutAll: self closureVarsAndValues.
	stream skip: -1; nextPut: $"; cr.
! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 5/21/2001 14:01'!
method

	^ method! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 5/28/2001 14:37'!
numArgs

	^ method numArgs! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 9/10/2002 16:53'!
printOn: aStream 

	super printOn: aStream.
	aStream space; nextPutAll: self identityHashPrintString! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 8/16/2002 15:54'!
vars

	| vars |
	vars _ Array new: self size.
	1 to: self size do: [:i | vars at: i put: (self at: i)].
	^ vars! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 8/16/2002 15:50'!
varsDo: block

	1 to: self size do: [:i | block value: (self at: i)]! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
repeat
	"Evaluate the receiver repeatedly, ending only if the block explicitly returns."

	[self value. true] whileTrue! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
repeatWithGCIf: testBlock
	| ans |
	"run the receiver, and if testBlock returns true, garbage collect and run the receiver again"
	ans _ self value.
	(testBlock value: ans) ifTrue: [ Smalltalk garbageCollect. ans _ self value ].
	^ans! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
whileFalse
	"Ordinarily compiled in-line, and therefore not overridable.
	This is in case the message is sent to other than a literal block.
	Evaluate the receiver, as long as its value is false."
 
	^ [self value] whileFalse: []! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
whileFalse: aBlock 
	"Ordinarily compiled in-line, and therefore not overridable.
	This is in case the message is sent to other than a literal block.
	Evaluate the argument, aBlock, as long as the value of the receiver is false."

	^ [self value] whileFalse: [aBlock value]! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
whileTrue
	"Ordinarily compiled in-line, and therefore not overridable.
	This is in case the message is sent to other than a literal block.
	Evaluate the receiver, as long as its value is true."
 
	^ [self value] whileTrue: []! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
whileTrue: aBlock 
	"Ordinarily compiled in-line, and therefore not overridable.
	This is in case the message is sent to other than a literal block.
	Evaluate the argument, aBlock, as long as the value of the receiver is true."

	^ [self value] whileTrue: [aBlock value]! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 7/15/2001 16:14'!
assert
	self assert: self! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 1/21/2003 17:47'!
ensure: aBlock
	"Evaluate a termination block after evaluating the receiver, regardless of whether the receiver's evaluation completes."

	| returnValue |
	<primitive: 198>
	returnValue _ self value.
	"aBlock wasn't nil when execution of this method began; it is nil'd out by the unwind machinery, and that's how we know it's already been evaluated ... otherwise, obviously, it needs to be evaluated"
	aBlock == nil ifFalse: [aBlock value].
	^ returnValue! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 1/21/2003 17:50'!
ifCurtailed: aBlock
	"Evaluate the receiver with an abnormal termination action."

	<primitive: 198>
	^ self value! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 2/1/2003 00:30'!
on: exception do: handlerAction
	"Evaluate the receiver in the scope of an exception handler."

	| handlerActive |
	<primitive: 199>  "just a marker, fail and execute the following"
	handlerActive _ true.
	^ self value! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 1/31/2003 20:41'!
onDNU: selector do: handleBlock
	"Catch MessageNotUnderstood exceptions but only those of the given selector (DNU stands for doesNotUnderstand:)"

	^ self on: MessageNotUnderstood do: [:exception |
		exception message selector = selector
			ifTrue: [handleBlock valueWithPossibleArgs: {exception}]
			ifFalse: [exception pass]
	  ]! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 7/26/2002 11:49'!
valueUninterruptably
	"Prevent remote returns from escaping the sender.  Even attempts to terminate (unwind) this process will be halted and the process will resume here.  A terminate message is needed for every one of these in the sender chain to get the entire process unwound."

	^ self ifCurtailed: [^ self]! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 1/27/2003 18:51'!
callCC
	"Call with current continuation, ala Scheme.
	Evaluate self against a copy of the sender's call stack, which can be resumed later"

	^ self value: thisContext sender asContinuation! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 7/15/2001 16:03'!
fork
	"Create and schedule a Process running the code in the receiver."

	^ self newProcess resume! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 10/16/2002 11:14'!
forkAndWait
	"Suspend current process while self runs"

	| semaphore |
	semaphore _ Semaphore new.
	[self ensure: [semaphore signal]] fork.
	semaphore wait.
! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 9/29/2001 21:00'!
forkAt: priority 
	"Create and schedule a Process running the code in the receiver at the given priority. Answer the newly created process."

	^ self newProcess
		priority: priority;
		resume! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 2/6/2003 13:33'!
newProcess
	"Answer a Process running the code in the receiver. The process is not 
	scheduled."

	<primitive: 19> "Simulation guard"
	^ Process
		forContext: 
			[self value.
			Processor terminateActive]
		priority: Processor activePriority! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 7/27/2002 12:26'!
simulate
	"Like run except interpret self using Smalltalk instead of VM.  It is much slower."

	^ self newProcess simulate! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 8/16/2002 15:51'!
closureAt: index
	"Answer the value of the closure variable at index"

	| v |
	v _ self at: index. 
	^ (v isMemberOf: SharedTempHolder)
		ifTrue: [v value]
		ifFalse: [v]! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 2/6/2002 14:36'!
closureAt: index put: value
	"Set the value of the closure variable at index"

	| v |
	v _ self at: index. 
	(v isKindOf: SharedTempHolder) ifTrue: [^ v value: value].
	^ self at: index put: value! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 1/28/2003 22:24'!
closureVarNames

	^ method blockNode closureVarNames! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 1/18/2002 11:24'!
closureVarsAndValues
	"Return a string of my vars with their current values"

	| aStream |
	aStream _ WriteStream on: (String new: 100).
	self closureVarNames doWithIndex: [:title :index |
		aStream nextPutAll: title; nextPut: $:; space; tab.
		(self closureAt: index) printOn: aStream.
		aStream cr].
	^ aStream contents! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 2/6/2003 13:27'!
hasMethodReturn
	"Answer whether the receiver has a return ('^') in its code."

	^ self method remoteReturns! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 2/3/2003 19:16'!
inspectorClass

	^ ClosureInspector! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 10/15/2002 04:10'!
parseEnv

	^ DoItClosureEnv new closure: self! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 2/6/2003 12:48'!
searchImageForHomeMethod

	Smalltalk allObjectsDo: [:obj |
		obj class == CompiledMethod ifTrue: [
			(obj pointsTo: self) ifTrue: [^ obj searchImageForHomeMethod]
		] ifFalse: [obj class == BlockClosure ifTrue: [
			(obj method == self and: [obj size = 0])
				ifTrue: [^ obj searchImageForHomeMethod]
		]]
	].
	^ method  "must be a loner block method"! !

!BlockClosure methodsFor: 'Camp Smalltalk' stamp: 'rw 1/23/2002 00:27'!
sunitEnsure: aBlock

        ^self ensure: aBlock! !

!BlockClosure methodsFor: 'Camp Smalltalk' stamp: 'rw 1/23/2002 00:28'!
sunitOn: anException do: aHandlerBlock

        ^self on: anException do: aHandlerBlock! !

!BlockClosure methodsFor: 'comparing' stamp: 'ajh 10/4/2002 17:09'!
= other

	self class == other class ifFalse: [^ false].
	self size = other size ifFalse: [^ false].
	1 to: self size do: [:i |
		(self at: i) = (other at: i) ifFalse: [^ false].
	].
	^ method = other method! !

!BlockClosure methodsFor: 'comparing' stamp: 'ajh 10/4/2002 17:12'!
hash

	^ method hash! !


!BlockNodeSignal methodsFor: 'as yet unclassified' stamp: 'ajh 2/1/2003 00:37'!
blockNode

	^ blockNode! !

!BlockNodeSignal methodsFor: 'as yet unclassified' stamp: 'ajh 2/1/2003 00:36'!
blockNode: aBlockNode2

	blockNode _ aBlockNode2! !


!BytecodeInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 9/28/2002 00:24'!
addTo: instructionSequence

	| obj |
	instructionSequence add: self.
	owningSequence _ instructionSequence.
	self recordStackEffect: instructionSequence localParseStack.

	"Associate self with mapToObject"
	obj _ owningSequence method mapToObject.
	obj ifNotNil: [obj mapInstr: self].
! !

!BytecodeInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 17:40'!
bytecodes

	| stream |
	stream _ Compiler bytecodeWriteStream on: (ByteArray new: 10).
	self bytecodesOn: stream.
	^ stream contents! !

!BytecodeInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 6/17/2002 18:37'!
bytecodesOn: stream
	"Write my bytecodes on stream"

	self subclassResponsibility! !

!BytecodeInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 9/27/2002 21:10'!
mapBytecodesOn: stream
	"Write my bytecodes on stream and remember their positions"

	| p |
	p _ stream position.
	self bytecodesOn: stream.
	localBytecodeRange _ p + 1 to: stream position.
! !

!BytecodeInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 8/6/2002 19:38'!
method

	^ owningSequence method! !

!BytecodeInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 10/3/2002 20:34'!
printOn: stream seqMap: seqToIndexMap

	stream tab.
	self printOn: stream.
	stream cr.
! !

!BytecodeInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 9/27/2002 21:38'!
recordStackEffect: parseStack
	"Remember local sp if needed then push and pop parseStack to show effect of this instruction"

	localSpBefore _ parseStack position.
	self stackEffect: parseStack! !

!BytecodeInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 8/6/2002 19:48'!
spBefore

	^ owningSequence parseStackIn position + localSpBefore! !

!BytecodeInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 6/5/2002 11:52'!
stackEffect: parseStack
	"Push and pop parseStack to show effect of this instruction"

	self subclassResponsibility! !

!BytecodeInstruction methodsFor: 'testing' stamp: 'ajh 6/13/2002 00:04'!
isJumpOrReturn

	^ false! !

!BytecodeInstruction methodsFor: 'testing' stamp: 'ajh 1/15/2003 15:33'!
isPushSpecialConstant

	^ false! !

!BytecodeInstruction methodsFor: 'testing' stamp: 'ajh 6/13/2002 10:27'!
isReturnInstr

	^ false! !

!BytecodeInstruction methodsFor: 'mapping' stamp: 'ajh 9/27/2002 21:35'!
bytecodeRange

	| start |
	start _ owningSequence startBytecodePosition.
	^ localBytecodeRange first + start to: localBytecodeRange last + start! !


!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/15/2003 22:36'!
dupTop

	self nextPut: (Bytecodes at: #duplicateTopBytecode)! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/26/2003 15:01'!
jump: distance if: condition

	| hi |
	distance > 0 ifFalse: [self error: 'positive if jumps only'].
	condition ifTrue: [
		hi _ distance // 256.
		hi < 8 ifFalse: [self error: 'true jump too big'].
		self nextPut: (Bytecodes at: #longJumpIfTrue) first + hi.
		self nextPut: distance \\ 256.
	] ifFalse: [
		distance <= 8 ifTrue: [
			self nextPut: (Bytecodes at: #shortConditionalJump) first + distance - 1.
		] ifFalse: [
			hi _ distance // 256.
			hi < 8 ifFalse: [self error: 'false jump too big'].
			self nextPut: (Bytecodes at: #longJumpIfFalse) first + hi.
			self nextPut: distance \\ 256.
		].
	]	
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/16/2003 00:45'!
jumpBackward: distance

	| dist |
	dist _ 1024 - distance - 2.
	dist < 0 ifTrue: [self error: 'back jump to big'].
	self nextPut: (Bytecodes at: #longUnconditionalJump) first + (dist // 256).
	self nextPut: dist \\ 256.
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/16/2003 00:52'!
jumpForward: distance

	distance > 0 ifFalse: [self error: 'must jump forward at least 1'].
	distance <= 8 ifTrue: [
		self nextPut: (Bytecodes at: #shortUnconditionalJump) first + distance - 1.
	] ifFalse: [
		distance > 1023 ifTrue: [self error: 'forward jump too big'].
		self nextPut: (Bytecodes at: #longUnconditionalJump) first + (distance // 256) + 4.
		self nextPut: distance \\ 256.
	].
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/16/2003 00:53'!
localReturn: fpOffset const: specialConstant

	self nextPut: (Bytecodes at: #returnTrue) + (SpecialReturnConstants indexOf: specialConstant) - 1.
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/16/2003 00:54'!
localReturnSelf: fpOffset

	self nextPut: (Bytecodes at: #returnReceiver).
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/16/2003 00:55'!
localReturnTop: fpOffset

	self nextPut: (Bytecodes at: #returnTopFromMethod).
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/16/2003 00:56'!
popTop

	self nextPut: (Bytecodes at: #popStackBytecode).
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/16/2003 00:56'!
pushConst: specialConstant

	self nextPut: (Bytecodes at: #pushConstantTrueBytecode) + (SpecialConstants indexOf: specialConstant) - 1
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/17/2003 23:48'!
pushInstVar: index

	| interval |
	interval _ Bytecodes at: #pushReceiverVariableBytecode.
	index <= interval size ifTrue: [
		^ self nextPut: (interval at: index).
	].
	index <= 64 ifTrue: [
		self nextPut: (Bytecodes at: #extendedPushBytecode).
		^ self nextPut: (0 "instVar" << 6) + index - 1.
	].
	index <= 256 ifFalse: [self error: 'can''t reference more than 256 inst vars'].
	self nextPut: (Bytecodes at: #doubleExtendedDoAnythingBytecode).
	self nextPut: 2 "pushInstVar" << 5.
	self nextPut: index - 1.
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/16/2003 01:14'!
pushLiteral: index

	| interval |
	interval _ Bytecodes at: #pushLiteralConstantBytecode.
	(index <= interval size) ifTrue: [
		^ self nextPut: (interval at: index)
	].
	index <= 64 ifTrue: [
		self nextPut: (Bytecodes at: #extendedPushBytecode).
		^ self nextPut: 2 "lit constant" << 6 + index - 1
	].
	index > 256 ifTrue: [self error: 'too many literals (>256)'].
	self nextPut: (Bytecodes at: #doubleExtendedDoAnythingBytecode).
	self nextPut: 3 "lit constant" << 5.
	self nextPut: index - 1.
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/17/2003 23:51'!
pushReceiver

	self nextPut: (Bytecodes at: #pushReceiverBytecode).
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/17/2003 23:51'!
pushTemp: index

	| interval |
	index = 0 ifTrue: [^ self pushReceiver].
	interval _ Bytecodes at: #pushTemporaryVariableBytecode.
	index <= interval size ifTrue: [
		^ self nextPut: (interval at: index).
	].
	index <= 64 ifFalse: [self error: 'too many temp vars (>64)'].
	self nextPut: (Bytecodes at: #extendedPushBytecode).
	self nextPut: (1 "temp" << 6) + index - 1.
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/21/2003 00:08'!
pushThisContext: fpOffset

	self nextPut: (Bytecodes at: #pushActiveContextBytecode)
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/17/2003 22:12'!
sendLiteral: litIndex numArgs: numArgs

	(litIndex <= 16 and: [numArgs <= 2]) ifTrue: [
		"short send"
		^ self nextPut: (Bytecodes at: #sendLiteralSelectorBytecode) first + (numArgs * 16) + litIndex - 1
	].
	(litIndex <= 32 and: [numArgs <= 7]) ifTrue: [
		"extended (2-byte) send"
		self nextPut: (Bytecodes at: #singleExtendedSendBytecode)..
		^ self nextPut: numArgs * 32 + litIndex - 1
	].
	(litIndex <= 64 and: [numArgs <= 3]) ifTrue: [
		"new extended (2-byte)"
		self nextPut: (Bytecodes at: #secondExtendedSendBytecode).
		^ self nextPut: numArgs * 64 + litIndex - 1
	].
	"long (3-byte) send"
	self nextPut: (Bytecodes at: #doubleExtendedDoAnythingBytecode).
	self nextPut: numArgs.
	self nextPut: litIndex - 1.
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/17/2003 22:16'!
sendLiteral: selectorIndex superOf: behaviorIndex numArgs: numArgs

	(selectorIndex <= 32 and: [numArgs <= 7]) ifTrue: [
		"extended (2-byte) send"
		self nextPut: (Bytecodes at: #singleExtendedSuperBytecode)..
		^ self nextPut: numArgs * 32 + selectorIndex - 1
	].
	"long (3-byte) send"
	self nextPut: (Bytecodes at: #doubleExtendedDoAnythingBytecode).
	self nextPut: 1 << 5 "super" + numArgs.
	self nextPut: selectorIndex - 1.
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/21/2003 12:36'!
sendLiteral: selectorIndex toSuperOfLastLiteralNumArgs: numArgs

	(selectorIndex <= 32 and: [numArgs <= 7]) ifTrue: [
		"extended (2-byte) send"
		self nextPut: (Bytecodes at: #singleExtendedSuperBytecode)..
		^ self nextPut: numArgs * 32 + selectorIndex - 1
	].
	"long (3-byte) send"
	self nextPut: (Bytecodes at: #doubleExtendedDoAnythingBytecode).
	self nextPut: 1 << 5 "super" + numArgs.
	self nextPut: selectorIndex - 1.
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/17/2003 22:18'!
sendSpecial: specialSelector

	self nextPut: (Bytecodes at: #bytecodePrimAdd) + (SpecialSelectors at: specialSelector)! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/17/2003 23:46'!
storeInstVar: index pop: bool

	| interval |
	interval _ Bytecodes at: #storeAndPopReceiverVariableBytecode.
	(bool and: [index <= interval size]) ifTrue: [
		^ self nextPut: (interval at: index).
	].
	index <= 64 ifTrue: [
		self nextPut: (Bytecodes at: (bool
			ifTrue: [#extendedStoreAndPopBytecode]
			ifFalse: [#extendedStoreBytecode])).
		^ self nextPut: (0 "instVar" << 6) + index - 1.
	].
	index <= 256 ifFalse: [self error: 'can''t reference more than 256 inst vars'].
	self nextPut: (Bytecodes at: #doubleExtendedDoAnythingBytecode).
	self nextPut: (5 "instVar" + bool asBit) << 5.
	self nextPut: index - 1.
! !

!BytecodeV3WriteStream methodsFor: 'instructions' stamp: 'ajh 1/17/2003 23:39'!
storeTemp: index pop: bool

	| interval |
	interval _ Bytecodes at: #storeAndPopTemporaryVariableBytecode.
	(bool and: [index <= interval size]) ifTrue: [
		^ self nextPut: (interval at: index).
	].
	index <= 64 ifFalse: [self error: 'too many temps (>64)'].
	self nextPut: (Bytecodes at: (bool
		ifTrue: [#extendedStoreAndPopBytecode]
		ifFalse: [#extendedStoreBytecode])).
	self nextPut: (1 "temp" << 6) + index - 1.
! !


!Class methodsFor: 'compiling' stamp: 'ajh 6/18/2002 14:48'!
holder

	^ Smalltalk associationAt: name ifAbsent: [nil -> self]! !

!Class methodsFor: 'compiling' stamp: 'ajh 9/21/2001 11:11'!
possiblePoolVariablesFor: misspelled continuedFrom: oldResults

	| results |
	results _ misspelled correctAgainstDictionary: self classPool continuedFrom: oldResults.
	self sharedPools do: [:pool | 
		results _ misspelled correctAgainstDictionary: pool continuedFrom: results].
	superclass
		ifNil: [^ results]
		ifNotNil: [^ superclass possiblePoolVariablesFor: misspelled continuedFrom: results]! !

!Class methodsFor: 'compiling' stamp: 'ajh 9/20/2001 22:51'!
scopeHasPoolVariable: varName ifTrue: assocBlock
	"Look up the first argument, varName, in the context of the receiver. If it is there,
	pass the association to the second argument, assocBlock, and answer true."

	| assoc |
	"First look in classVar dictionary."
	(assoc _ self classPool associationAt: varName ifAbsent: [])
		ifNotNil: [assocBlock value: assoc.  ^ true].

	"Next look in shared pools."
	self sharedPools do: [:pool | 
		(assoc _ pool associationAt: varName ifAbsent: [])
			ifNotNil: [assocBlock value: assoc.  ^true]].

	"Finally look higher up the superclass chain and fail at the end."
	superclass == nil
		ifTrue: [^ false]
		ifFalse: [^ superclass scopeHasPoolVariable: varName ifTrue: assocBlock].
! !


!BlockClosure class methodsFor: 'documentation' stamp: 'ajh 2/6/2003 15:08'!
aboutClosures"

These changesets add a closure compiler to the standard Squeak image (without replacing the existing compiler).  A new 'general' preference called #compileBlocksAsClosures determines which compiler to use for new methods.  When true the closure compiler is used, when false the existing compiler is used (default).

Etoy scripts do not yet work with the closure compiler so be sure to work with them with the above preference set to false.

The closure compiler compiles non-inlined blocks as separate compiledMethods.  These block methods are held in the literals of the home method and sent the #createBlock... message at runtime to create BlockClosures.  Outer temps referenced by the block are given as args to the #createBlock... message which adds them to the closure it creates, which consists of the method in its sole named field and the captured temps in its indexable fields.  When #value... is sent to a BlockClosure its method is executed in a new MethodContext with the BlockClosure remaining as the receiver.  The block method accesses captured temps in indexed fields as if they were named instance variables.

Captured temps that may change after being captured are wrapped in SharedTempHolders and all methods (home and block) access their values indirectly through the holders.  If a captured temp may not change (most common) then no holder is created saving time.

BlockClosures that return to its home context hold the home context as one of its captured temps.  Actually, to support safe continuations, it holds the tag of its home context.  A ContextTag is a unique object for each context that stays the same for all copies of each context.  When a BlockClosure attempts to return to its home context it searches the sender chain for the context that has the right ContextTag and returns to it.  This works with copied contexts (continuations) as well as original contexts.

Thanks to captured temps and context tags, block closures are totally separate from their home context.  They are reentrant and each activation has its own block-local temps.

BlockClosures are slower primarily because each activation is a new context.  BlockContexts are faster because they reuse the same context.  An optimizing VM (Jitter, VI4) that stack allocates contexts should rectify this problem.  For an example of its slowness (and general slowness of context activation) execute '[(1 to: 100000) do: [:i | i]] timeToRun' under each compiler (toggle the preference).

The above example will be really slow for the closure compiler until you generate interp.c and recompile your VM to include a couple of new primitives (186 and 187) added with the closure compiler.  They implement BlockClosure>>value... directly.  Without them the failure code has to add the closure method to the BlockClosure method dictionary then call #perform:.

Closure Compiler Implementation

The back-end of the new compiler has been totally rewritten.  All of the new compiler classes have been put into its own system category called 'Compiler-'.  Scanner is reused as is, Parser2 is basically the same as Parser, and the ParseNode2s are different but similar to the originals, but their emit code is totally different.  They emit to an InstructionBuilder by sending it abstract bytecode messages (see InstructionBuilder comment and its 'instructions' method category).

But before the parse nodes can emit temp vars, var usage has to be analyzed to determine those that need to be wrapped in SharedTempHolders (see calcVarUsage).  A FiniteAutomaton is used to advance the useState of each var depending on read/write and whether it is in a closure or not (see VarUse).  Those that reach the isIndirect state will generate #wrapInTempHolder to initialize the shared temp.

The InstructionBuilder builds BytecodeInstruction objects and groups them by basic block (InstructionSequence).  A basic block is a sequence of instructions that are always executed in order, only the last instruction jumps or returns.  Some optimizations are done to the instructions (see simplify) like collapsing constant conditional jumps generated by inlined and:/or: message nodes.  Bytecodes are then generated by sending an abstract bytecode messages from each intruction to a BytecodeV3WriteStream.  The BytecodeV3WriteStream encodes these messages into bytes on its byte array.

Care was taken to keep the parse tree and the InstructionBuilder simple so other languages could potentially be compiled to bytecodes, including different versions of our own Smalltalk language.  The programmer can choose to parse to the existing parse nodes or build his own parse nodes and emit to the InstructionBuilder.  The parse tree was kept simple and easy to build by: avoiding back pointers; keeping all blocks the same whether inlined or not; keeping all messages the same whether inlined or not; and keeping all temp vars the same whether captured in a closure or not.  These distinction are calculated during emit and not stored.

Decompiler2 is a subclass of Decompiler and DecompilerConstructor2 is a new class originally copied from the original but significantly changed.  The Decompiler2 interprets bytecodes and builds the parse tree directly (no intermediate representation like the InstructionBuilder).  There may still be a few bugs when decompile optimized code like collapsed jumps, but this should not get in your way since most the parse tree is gnearted from the source when it is available.

Other General Enhancements

Context Control Enhancements - Methods like #return:, #restart, and #jump have been added to ContextPart for simpler control of the current thread.

Process Control Enhancements - Methods like #completeStep: have been added to Process for simpler control of suspended processes.  'completeStep: aContext' will resume the thread them jump back to the triggering thread when aContext get control back.  This allows quick stepping while in the context of the suspended process (i.e. its exception handlers will be active instead of the triggering thread's).

Debugger Enhancements - The debugger uses the new process methods described above.  Also, quickStep has been fixed to execute in the context of the debugged process so earlier exception handlers will be active.

ContextTags - 'methodContext contextTag' will create a unique ContextTag object for that methodContext and hold it in it receiverMap instance variable which was previously unused.  Copies of the methodContext will hold the same contextTag.  Holding a contextTag instead of a context directly is safer in light of context copies (see continuations below).  A context for a context tag is found in the current thread just be traversing the senders until one is found with that contextTag.

Exceptions Refactored - Exceptions now hold initialContext and handlerContext tags instead of the actual contexts so contexts can be copied safely (described above).  Also, searching for handler contexts has been simplified utilizing Context control behavior described above.

Continuations - 'Continuation current' copies thisContext stack and wraps it in a Continuation object.  Also, '[:cc | ...] callCC' copies the sender stack into a Continuation and returns it.  Evaluating a continuation copies the stack again and resumes that copy.  The current process stack is abandoned.  When a BlockContext on the stack is copied it points to copied home context if it is also on the stack.

Miscellaneous - Changeset 1 contains general additions to basic objects, collections, and streams, that are used by the rest of changes described above.

Closure Test Cases - Thanks to Boris Gaertner and Rob Withers for writing test cases for block closures.  See BlockClosuresTestCase (of course they all pass).

"! !

!BlockClosure class methodsFor: 'compiling' stamp: 'ajh 2/4/2003 16:11'!
initialize

	Preferences
		addPreference: #compileBlocksAsClosures
		category: #general
		default: false
		balloonHelp: 'If true, blocks ([...]) will be compiled as BlockClosures instead of BlockContexts.  BlockClosures are independent of their home context but are a little slower (for now).  Only newly compiled methods will be affected by this preference.  Make sure this is false when using Etoy scripts; they have not been adapted to the new compiler yet.'
! !

!BlockClosure class methodsFor: 'compiling' stamp: 'ajh 8/9/2002 01:52'!
varOffset

	^ self instSize! !


!BytecodeV3WriteStream class methodsFor: 'accessing' stamp: 'ajh 1/15/2003 17:32'!
bytecodeDict

	^ Bytecodes! !

!BytecodeV3WriteStream class methodsFor: 'accessing' stamp: 'ajh 1/15/2003 17:32'!
bytecodeTable

	^ BytecodeTable! !

!BytecodeV3WriteStream class methodsFor: 'accessing' stamp: 'ajh 1/15/2003 17:51'!
receiverFrameIndex: numRargs

	^ 0! !

!BytecodeV3WriteStream class methodsFor: 'accessing' stamp: 'ajh 1/15/2003 17:32'!
specialConstants

	^ SpecialConstants! !

!BytecodeV3WriteStream class methodsFor: 'accessing' stamp: 'ajh 1/15/2003 17:32'!
specialReturnConstants

	^ SpecialReturnConstants! !

!BytecodeV3WriteStream class methodsFor: 'accessing' stamp: 'ajh 1/15/2003 17:32'!
specialSelectors

	^ SpecialSelectors! !

!BytecodeV3WriteStream class methodsFor: 'accessing' stamp: 'ajh 1/15/2003 18:49'!
specialSelectorsArray

	^ #(#+ 1 #- 1 #< 1 #> 1 #<= 1 #>= 1 #= 1 #~= 1 #* 1 #/ 1 #\\ 1 #@ 1 #bitShift: 1 #// 1 #bitAnd: 1 #bitOr: 1 #at: 1 #at:put: 2 #size 0 #next 0 #nextPut: 1 #atEnd 0 #== 1 #class 0 #blockCopy: 1 #value 0 #value: 1 #do: 1 #new 0 #new: 1 #x 0 #y 0)! !

!BytecodeV3WriteStream class methodsFor: 'initialize' stamp: 'ajh 1/15/2003 18:47'!
bytecodeTableFrom: specArray
	"SpecArray is an array of either
	(index selector) or
	(index1 index2 selector)."

	| contiguous |
	Bytecodes _ IdentityDictionary new: 256.
	BytecodeTable _ Array new: 256.
	contiguous _ 0.
	specArray do: [ :spec |
		(spec at: 1) = contiguous ifFalse: [self error: 'Non-contiguous table entry'].
		spec size = 2 ifTrue: [
			Bytecodes at: (spec at: 2) put: (spec at: 1).
			BytecodeTable at: (spec at: 1) + 1 put: (spec at: 2).
			contiguous _ contiguous + 1.
		] ifFalse: [
			spec size = 3 ifFalse: [self error: 'bad spec size'].
			Bytecodes at: (spec at: 3) put: ((spec at: 1) to: (spec at: 2)).
			(spec at: 1) to: (spec at: 2) do: [ :i |
				BytecodeTable at: i + 1 put: (spec at: 3).
			].
			contiguous _ contiguous + ((spec at: 2) - (spec at: 1)) + 1.
		].
	].
	^ BytecodeTable! !

!BytecodeV3WriteStream class methodsFor: 'initialize' stamp: 'ajh 1/15/2003 17:32'!
initialize

	self initializeBytecodeTable.
	self initializeSpecialSelectors.
	self initializeSpecialConstants.
	self initializeSpecialReturnConstants.
! !

!BytecodeV3WriteStream class methodsFor: 'initialize' stamp: 'ajh 1/15/2003 17:58'!
initializeBytecodeTable
	"BytecodeWriteStream initialize"
	"Defines all the bytecode instructions for the Compiler and the Interpreter.
	The following bytecode tuple format is:

	#(Bytecode BytecodeSelector)

	BytecodeSelector is the method in the Interpreter that gets executed for the given bytecode.
	BytecodeDecoder willSend expects all send bytecodes and only send bytecodes to start with 'send'.
	Common Send selector position within the specialSelectorsArray is hard code in the Interpreter, see senders of Interpreter specialSelector:."

	^ self bytecodeTableFrom:
	#(
		(  0  15 pushReceiverVariableBytecode)
		( 16  31 pushTemporaryVariableBytecode)
		( 32  63 pushLiteralConstantBytecode)
		( 64  95 pushLiteralVariableBytecode)
		( 96 103 storeAndPopReceiverVariableBytecode)
		(104 111 storeAndPopTemporaryVariableBytecode)
		(112 pushReceiverBytecode)
		(113 pushConstantTrueBytecode)
		(114 pushConstantFalseBytecode)
		(115 pushConstantNilBytecode)
		(116 pushConstantMinusOneBytecode)
		(117 pushConstantZeroBytecode)
		(118 pushConstantOneBytecode)
		(119 pushConstantTwoBytecode)
		(120 returnReceiver)
		(121 returnTrue)
		(122 returnFalse)
		(123 returnNil)
		(124 returnTopFromMethod)
		(125 returnTopFromBlock)
		(126 unknownBytecode)
		(127 unknownBytecode)
		(128 extendedPushBytecode)
		(129 extendedStoreBytecode)
		(130 extendedStoreAndPopBytecode)
		(131 singleExtendedSendBytecode)
		(132 doubleExtendedDoAnythingBytecode)
		(133 singleExtendedSuperBytecode)
		(134 secondExtendedSendBytecode)
		(135 popStackBytecode)
		(136 duplicateTopBytecode)
		(137 pushActiveContextBytecode)
		(138 143 experimentalBytecode)
		(144 151 shortUnconditionalJump)
		(152 159 shortConditionalJump)
		(160 167 longUnconditionalJump)
		(168 171 longJumpIfTrue)
		(172 175 longJumpIfFalse)

		"176-191 were sendArithmeticSelectorBytecode"
		(176 bytecodePrimAdd)
		(177 bytecodePrimSubtract)
		(178 bytecodePrimLessThan)
		(179 bytecodePrimGreaterThan)
		(180 bytecodePrimLessOrEqual)
		(181 bytecodePrimGreaterOrEqual)
		(182 bytecodePrimEqual)
		(183 bytecodePrimNotEqual)
		(184 bytecodePrimMultiply)
		(185 bytecodePrimDivide)
		(186 bytecodePrimMod)
		(187 bytecodePrimMakePoint)
		(188 bytecodePrimBitShift)
		(189 bytecodePrimDiv)
		(190 bytecodePrimBitAnd)
		(191 bytecodePrimBitOr)	

		"192-207 were sendCommonSelectorBytecode"
		(192 bytecodePrimAt)
		(193 bytecodePrimAtPut)
		(194 bytecodePrimSize)
		(195 bytecodePrimNext)
		(196 bytecodePrimNextPut)
		(197 bytecodePrimAtEnd)
		(198 bytecodePrimEquivalent)
		(199 bytecodePrimClass)
		(200 bytecodePrimBlockCopy)
		(201 bytecodePrimValue)
		(202 bytecodePrimValueWithArg)
		(203 bytecodePrimDo)
		(204 bytecodePrimNew)
		(205 bytecodePrimNewWithArg)
		(206 bytecodePrimPointX)
		(207 bytecodePrimPointY)

		(208 255 sendLiteralSelectorBytecode)
	)
! !

!BytecodeV3WriteStream class methodsFor: 'initialize' stamp: 'ajh 1/15/2003 17:32'!
initializeSpecialConstants

	SpecialConstants _ {true. false. nil. -1. 0. 1. 2}! !

!BytecodeV3WriteStream class methodsFor: 'initialize' stamp: 'ajh 1/15/2003 17:32'!
initializeSpecialReturnConstants

	SpecialReturnConstants _ {true. false. nil}! !

!BytecodeV3WriteStream class methodsFor: 'initialize' stamp: 'ajh 1/15/2003 17:32'!
initializeSpecialSelectors
	"Create a map from specialSelector -> bytecode offset from sendAdd (the first one)"

	| array |
	SpecialSelectors _ IdentityDictionary new.
	array _ self specialSelectorsArray.  "Smalltalk specialObjectsArray at: 24"
	1 to: array size by: 2 "skip numArgs" do: [:i |
		SpecialSelectors at: (array at: i) put: i - 1 / 2].
! !

!BytecodeV3WriteStream class methodsFor: 'initialize' stamp: 'ajh 1/15/2003 17:32'!
reverseDictTable: array
	"Build a reverse table from the array, with repeats represented as Intervals"

	| lastValue lastIndex reverseDict |
	reverseDict _ IdentityDictionary new: array size.
	lastValue _ #dummy.
	lastIndex _ nil.
	array withIndexDo: [:value :i |
		value = lastValue ifTrue: [
			lastIndex isInteger 
				ifTrue: [lastIndex _ lastIndex to: i - 1]
				ifFalse: [lastIndex setFrom: lastIndex first to: i - 1 by: 1]
		] ifFalse: [
			reverseDict at: lastValue put: lastIndex.
			lastValue _ value.
			lastIndex _ i - 1]].
	reverseDict at: lastValue put: lastIndex.
	reverseDict removeKey: #dummy.
	^ reverseDict! !


!ClosureInspector methodsFor: 'accessing' stamp: 'ajh 1/18/2002 12:29'!
doItContext
	"Use my closure as the context to get closure variables"

	^ object! !

!ClosureInspector methodsFor: 'accessing' stamp: 'ajh 9/14/2002 09:51'!
doItReceiver
	"My doItContext will take precedence"

	^ nil! !

!ClosureInspector methodsFor: 'accessing' stamp: 'ajh 2/4/2003 13:21'!
fieldList 
	"Refer to the comment in Inspector|fieldList."

	object == nil ifTrue: [^ #('thisClosure')].
	^ #('thisClosure' 'block code' 'all free vars'), object closureVarNames! !

!ClosureInspector methodsFor: 'selecting' stamp: 'ajh 2/9/2003 16:19'!
replaceSelectionValue: anObject 
	"Refer to the comment in Inspector|replaceSelectionValue:."

	selectionIndex = 1 ifTrue: [^ object].
	^ object closureAt: selectionIndex - 3 put: anObject! !

!ClosureInspector methodsFor: 'selecting' stamp: 'ajh 2/4/2003 13:20'!
selection 
	"Refer to the comment in Inspector|selection."

	selectionIndex = 0 ifTrue:[^ ''].
	selectionIndex = 1 ifTrue: [^ object].
	selectionIndex = 2 ifTrue: [^ object method blockNode decompileString].
	selectionIndex = 3 ifTrue: [^ object closureVarsAndValues].
	^ object closureAt: selectionIndex - 3! !


!CompiledMethod methodsFor: 'initialize-release' stamp: 'ajh 1/22/2003 16:26'!
isClosureCompiled: bool
	"Use the sign bit in the header to mark methods that have been compiled using the new closure compiler (Parser2)."

	bool
		ifTrue: [self objectAt: 1 put: (self header bitOr: 1 << 30) as31BitSmallInt]
		ifFalse: [self error: 'Don''t send if not closure compiled']! !

!CompiledMethod methodsFor: 'literals' stamp: 'ajh 2/6/2003 13:25'!
hasLiteralSuchThat: testBlock
	"Answer true if testBlock returns true for any literal in this method, even if imbedded."

	| lit |
	2 to: self numLiterals + 1 do: 
		[:index | 
		(testBlock value: (lit _ self objectAt: index)) ifTrue: [^ true].
		(lit hasLiteralSuchThat: testBlock) ifTrue: [^ true]].
	^ false! !

!CompiledMethod methodsFor: 'literals' stamp: 'ajh 1/31/2003 12:05'!
hasLiteralThorough: literal
	"Answer true if any literal in this method is literal,
	even if embedded in array structure."

	| lit |
	2 to: self numLiterals + 1 do: 
		[:index | 
		(lit _ self objectAt: index) == literal ifTrue: [^ true].
		(lit hasLiteralThorough: literal) ifTrue: [^ true]].
	^ false! !

!CompiledMethod methodsFor: 'scanning' stamp: 'ajh 2/6/2003 13:27'!
remoteReturns
	"For closure methods only"

	^ self messages includes: #remoteReturnTo:! !

!CompiledMethod methodsFor: 'decompiling' stamp: 'ajh 7/14/2001 12:34'!
primitiveNode

	| primNode n |
	primNode _ PrimitiveNode new num: (n _ self primitive).
	(n = 117 or: [n = 120]) ifTrue: [
		primNode spec: (self literalAt: 1)].
	^ primNode! !

!CompiledMethod methodsFor: 'decompiling' stamp: 'ajh 2/6/2003 12:48'!
searchImageForHomeMethod

	Smalltalk allObjectsDo: [:obj |
		obj class == CompiledMethod ifTrue: [
			(obj pointsTo: self) ifTrue: [^ obj searchImageForHomeMethod]
		] ifFalse: [obj class == BlockClosure ifTrue: [
			(obj method == self and: [obj size = 0])
				ifTrue: [^ obj searchImageForHomeMethod]
		]]
	].
	^ self  "must be a loner block method"! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:29'!
createBlock
	"Generate by the compiler to create a block closure with self as its method and no captured temps"

	^ BlockClosure new method: self! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:23'!
createBlockt: t1
	"Generate by the compiler to create a block closure with self as its method, and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 1) method: self;
		at: 1 put: t1;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:22'!
createBlockt: t1 t: t2
	"Generate by the compiler to create a block closure with self as its method, and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 2) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:22'!
createBlockt: t1 t: t2 t: t3
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 3) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:23'!
createBlockt: t1 t: t2 t: t3 t: t4
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 4) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:23'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 5) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:24'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 6) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:24'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 7) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:24'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 8) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:25'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 9) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:25'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 10) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:25'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10 t: t11
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 11) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		at: 11 put: t11;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:26'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10 t: t11 t: t12
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 12) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		at: 11 put: t11;
		at: 12 put: t12;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:26'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10 t: t11 t: t12 t: t13
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 13) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		at: 11 put: t11;
		at: 12 put: t12;
		at: 13 put: t13;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:26'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10 t: t11 t: t12 t: t13 t: t14
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 14) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		at: 11 put: t11;
		at: 12 put: t12;
		at: 13 put: t13;
		at: 14 put: t14;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:27'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10 t: t11 t: t12 t: t13 t: t14 t: t15
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 15) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		at: 11 put: t11;
		at: 12 put: t12;
		at: 13 put: t13;
		at: 14 put: t14;
		at: 15 put: t15;
		yourself! !


!Compiler methodsFor: 'public access' stamp: 'ajh 2/8/2003 23:03'!
compileDoIt: textOrStream in: aContext to: receiver notifying: aRequestor ifFail: failBlock
	"Compiles the sourceStream into a parse tree, then generates code into a method.  This method is then wrapped in a block with the receiver or context as the sole free variable which the method refers to.  If requestor is not nil, then it will receive a notify:at: message if there is a compile error, followed by the failBlock being executed."

	| scope parser blockNode method |
	scope _ aContext ifNil: [receiver class].
	parser _ self parserClass new.
	blockNode _ parser
			parse: textOrStream readStream
			class: scope
			noPattern: true
			notifying: aRequestor
			ifFail: [^ failBlock value].
	method _ blockNode compiledMethod.
	"cacheDoItNode ifTrue: [
		method cacheDoItBlockNode: blockNode.
		method cacheHomeMethod: nil.
	]."
	^ MethodClosure new
		method: method;
		receiver: (aContext ifNil: [receiver]);
		yourself! !

!Compiler methodsFor: 'public access' stamp: 'ajh 9/14/2002 22:00'!
evaluate2: textOrStream in: aContext to: receiver notifying: aRequestor ifFail: failBlock

	^ (self compileDoIt: textOrStream in: aContext to: receiver notifying: aRequestor ifFail: [^ failBlock value]) value! !

!Compiler methodsFor: 'public access' stamp: 'ajh 1/27/2003 20:35'!
evaluate: textOrStream in: aContext to: receiver notifying: aRequestor ifFail: failBlock
	"Compiles the sourceStream into a parse tree, then generates code into a 
	method. This method is then installed in the receiver's class so that it 
	can be invoked. In other words, if receiver is not nil, then the text can 
	refer to instance variables of that receiver (the Inspector uses this). If 
	aContext is not nil, the text can refer to temporaries in that context (the 
	Debugger uses this). If aRequestor is not nil, then it will receive a 
	notify:at: message before the attempt to evaluate is aborted. Finally, the 
	compiled method is invoked from here as DoIt or (in the case of 
	evaluation in aContext) DoItIn:. The method is subsequently removed 
	from the class, but this will not get done if the invocation causes an 
	error which is terminated. Such garbage can be removed by executing: 
	Smalltalk allBehaviorsDo: [:cl | cl removeSelector: #DoIt; removeSelector: 
	#DoItIn:]."

	| methodNode method value |
	Preferences compileBlocksAsClosures ifTrue: [
		cacheDoItNode _ true.
		^ self evaluate2: textOrStream in: aContext to: receiver notifying: aRequestor ifFail: failBlock].
	class _ (aContext == nil ifTrue: [receiver] ifFalse: [aContext receiver]) class.
	self from: textOrStream class: class context: aContext notifying: aRequestor.
	methodNode _ self translate: sourceStream noPattern: true ifFail:
		[^failBlock value].
	method _ methodNode generate: #(0 0 0 0).
	self interactive ifTrue:
		[method _ method copyWithTempNames: methodNode tempNames].
	context == nil
		ifTrue: [class addSelector: #DoIt withMethod: method.
				value _ receiver DoIt.
				InMidstOfFileinNotification signal ifFalse: [
					class removeSelectorSimply: #DoIt.
				].
				^value]
		ifFalse: [class addSelector: #DoItIn: withMethod: method.
				value _ receiver DoItIn: context.
				InMidstOfFileinNotification signal ifFalse: [
					class removeSelectorSimply: #DoItIn:.
				].
				^value]! !


!Compiler class methodsFor: 'accessing' stamp: 'ajh 1/15/2003 17:46'!
bytecodeWriteStream
	"Return the correct bytecode version for the current image/VM format"

	^ BytecodeV3WriteStream! !

!Compiler class methodsFor: 'accessing' stamp: 'ajh 1/20/2002 18:57'!
parserClass
	"Return a parser class to use for parsing method headers."

	^ Preferences compileBlocksAsClosures
		ifTrue: [Parser2]
		ifFalse: [Parser]! !


!ContextPart methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:39'!
temp: index
	"Answer the value of the temporary variable whose index is the 
	argument, index.  The first argument is temp 1"

	^ self subclassResponsibility! !

!ContextPart methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:40'!
temp: index put: value 
	"Store the argument, value, as the temporary variable whose index is the 
	argument, index.  The first arg of self is temp 1"

	^ self subclassResponsibility! !

!ContextPart methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:45'!
tempAt: index
	"Get the temp value at index.  Make SharedTempHolders transparent."

	| var |
	var _ self temp: index.
	^ (var isMemberOf: SharedTempHolder)
		ifTrue: [var value]
		ifFalse: [var]! !

!ContextPart methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:45'!
tempAt: index put: val
	"Set the temp at index.  Make SharedTempHolders transparent."

	| var |
	var _ self temp: index.
	^ (var isMemberOf: SharedTempHolder)
		ifTrue: [var value: val]
		ifFalse: [self temp: index put: val]! !

!ContextPart methodsFor: 'private' stamp: 'ajh 1/31/2003 15:07'!
doPrimitive: primitiveIndex method: meth receiver: receiver args: arguments 
	"Simulate a primitive method whose index is primitiveIndex.  The
	simulated receiver and arguments are given as arguments to this message."

	| value |
	<primitive: 19> "Simulation guard"
	"If successful, push result and return resuming context,
		else ^ PrimitiveFailToken"
	(primitiveIndex = 19) ifTrue:[
		Debugger 
			openContext: self
			label:'Code simulation error'
			contents: nil].

	(primitiveIndex = 80 and: [receiver isKindOf: ContextPart])
		ifTrue: [^self push: ((BlockContext newForMethod: receiver home method)
						home: receiver home
						startpc: pc + 2
						nargs: (arguments at: 1))].
	(primitiveIndex = 81 and: [receiver isMemberOf: BlockContext])
		ifTrue: [^receiver pushArgs: arguments from: self].
	primitiveIndex = 83 "afr 9/11/1998 19:50"
		ifTrue: [^ self send: arguments first to: receiver
					with: arguments allButFirst
					super: false].
	primitiveIndex = 84 "afr 9/11/1998 19:50"
		ifTrue: [^ self send: arguments first to: receiver
					with: (arguments at: 2)
					super: false].
	primitiveIndex = 186 ifTrue: [ "closure value"
		| m |
		m _ receiver method.
		arguments size = m numArgs ifFalse: [^ PrimitiveFailToken].
		^ self activateMethod: m
			withArgs: arguments
			receiver: receiver
			class: receiver class].
	primitiveIndex = 187 ifTrue: [ "closure valueWithArguments:"
		| m args |
		m _ receiver method.
		args _ arguments first.
		args size = m numArgs ifFalse: [^ PrimitiveFailToken].
		^ self activateMethod: m
			withArgs: args
			receiver: receiver
			class: receiver class].
	arguments size > 6 ifTrue: [^ PrimitiveFailToken].
	primitiveIndex = 117 
		ifTrue:[value _ self tryNamedPrimitiveIn: meth for: receiver withArgs: arguments]
		ifFalse:[value _ receiver tryPrimitive: primitiveIndex withArgs: arguments].
	value == PrimitiveFailToken
		ifTrue: [^ PrimitiveFailToken]
		ifFalse: [^ self push: value]! !

!ContextPart methodsFor: 'query' stamp: 'ajh 1/31/2003 21:53'!
isClosureContext

	^ false! !

!ContextPart methodsFor: 'query' stamp: 'ajh 1/31/2003 21:50'!
parseEnv

	^ DoItContextEnv new context: self! !


!BlockContext methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:40'!
temp: index 
	"Refer to the comment in ContextPart|temp:."

	^ home at: index! !

!BlockContext methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:40'!
temp: index put: value 
	"Refer to the comment in ContextPart|temp:put:."

	^ home at: index put: value! !


!CreateBlockInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/17/2003 23:22'!
addTo: sequence

	super addTo: sequence.
	(Compiler bytecodeWriteStream canUnderstand: #createBlock:) ifTrue: [^ self].

	"Create selector for normal send"
	closureSize > 15 ifTrue: [
		self error: 'Block closure can''t reference more than 15 outer temps'].
	selector _ WriteStream on: ''.
	selector nextPutAll: #createBlock.
	closureSize timesRepeat: [selector nextPutAll: 't:'].
	selector _ selector contents asSymbol.
	isSpecial _ Compiler bytecodeWriteStream specialSelectors includesKey: selector.
	isSpecial ifFalse: [sequence method addLiteral: selector].
! !

!CreateBlockInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/17/2003 23:21'!
bytecodesOn: stream

	selector ifNil: [^ stream createBlock: closureSize].

	isSpecial
		ifTrue: [stream sendSpecial: selector]
		ifFalse: [stream
			sendLiteral: (owningSequence method literalSlotIndexOf: selector)
			numArgs: closureSize]
! !

!CreateBlockInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/5/2002 01:59'!
closureSize: n

	closureSize _ n! !

!CreateBlockInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/5/2002 01:59'!
printOn: stream

	stream nextPutAll: 'createBlock '; print: closureSize.
! !

!CreateBlockInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/5/2002 01:59'!
stackEffect: parseStack

	parseStack pop: closureSize! !


!Decompiler2 methodsFor: 'initialize' stamp: 'ajh 2/9/2003 01:14'!
asBlock

	constructor asBlockNode: #()! !

!Decompiler2 methodsFor: 'initialize' stamp: 'ajh 2/4/2003 15:25'!
asBlock: capturedVars

	constructor asBlockNode: capturedVars! !

!Decompiler2 methodsFor: 'initialize' stamp: 'ajh 1/28/2003 18:22'!
initialize

	super initialize.
	keepNodesBasic _ false.
! !

!Decompiler2 methodsFor: 'initialize' stamp: 'ajh 8/6/2002 13:13'!
keepBasicConstructs
	"Don't convert while loops to to:do: loops, nor brace arrays to ArrayNodes, nor ifs to cases, nor indirect literals to pool variables, so we don't lose the increment statement at the end of the loop, nor the initial push Array literal, nor each case test instr, nor push global literal (respectively) with their bytecode mapping that may be needed"
	"protecing against ArrayNodes and CaseMessageNodes has not been implemented yet"

	keepNodesBasic _ true! !

!Decompiler2 methodsFor: 'initialize' stamp: 'ajh 2/3/2003 18:27'!
outer: decompiler
	"Set outer method decompiler"

	outer _ decompiler! !

!Decompiler2 methodsFor: 'var scope' stamp: 'ajh 10/4/2002 21:45'!
scopeNumberOfVar: varNode

	^ constructor scopeNumberOfVar: varNode! !

!Decompiler2 methodsFor: 'var scope' stamp: 'ajh 10/4/2002 21:45'!
scopeNumberOfVar: varNode is: n

	^ constructor scopeNumberOfVar: varNode is: n! !

!Decompiler2 methodsFor: 'control' stamp: 'ajh 2/9/2003 22:22'!
checkForBlock: receiver
	"We just saw a blockCopy: message. Check for a following block."

	| savePc jump args argPos block blockCopyStart startPc |
	receiver = constructor codeThisContext ifFalse: [^false].
	blockCopyStart _ lastPc.
	savePc _ pc.
	(jump _ self interpretJump) notNil
		ifFalse:
			[pc _ savePc.  ^nil].
	"Definitely a block"
	startPc _ pc.
	jump _ jump + pc.
	argPos _ statements size.
	[self willStorePop]
		whileTrue:
			[stack addLast: ArgumentFlag.  "Flag for doStore:"
			self interpretNextInstructionFor: self].
	args _ Array new: statements size - argPos.
	1 to: args size do:  "Retrieve args"
		[:i | args at: i put: statements removeLast.
		(args at: i) scopeIn: self set: -1.  "flag args as block temps"].
	block _ self blockTo: jump.
	stack addLast: (constructor codeArguments: args block: block range: (blockCopyStart - (method initialPC - 1) to: startPc - 1 - (method initialPC - 1))).
	^true! !

!Decompiler2 methodsFor: 'control' stamp: 'ajh 1/31/2003 13:47'!
convertToDoLoop
	"If statements contains the pattern
		var _ startExpr.
		[var <= limit] whileTrue: [...statements... var _ var + incConst]
	then replace this by
		startExpr to: limit by: incConst do: [:var | ...statements...]"
	| initStmt toDoStmt limitStmt |
	statements size < 2 ifTrue: [^ self].
	initStmt _ statements at: statements size-1.
	(toDoStmt _ statements last toDoFromWhileWithInit: initStmt)
		== nil ifTrue: [^ self].
	initStmt variable scopeIn: self set: -1.  "Flag arg as block temp"
	statements removeLast; removeLast; addLast: toDoStmt.

	"Attempt further conversion of the pattern
		limitVar _ limitExpr.
		startExpr to: limitVar by: incConst do: [:var | ...statements...]
	to
		startExpr to: limitExpr by: incConst do: [:var | ...statements...]"
	statements size < 2 ifTrue: [^ self].
	limitStmt _ statements at: statements size-1.
	((constructor isAssignmentNode: limitStmt)
		and: [limitStmt variable isTemp
		and: [limitStmt variable referenceNode = toDoStmt arguments first
		and: [self methodRefersOnlyOnceToTemp: limitStmt variable fieldOffset]]])
		ifFalse: [^ self].
	toDoStmt arguments at: 1 put: limitStmt value.
	limitStmt variable scopeIn: self set: -2.  "Flag limit var so it won't print"
	statements removeLast; removeLast; addLast: toDoStmt.

! !

!Decompiler2 methodsFor: 'control' stamp: 'ajh 2/4/2003 14:39'!
jump: dist if: condition

	| savePc elseDist sign elsePc elseStart end cond ifExpr thenBlock elseBlock thenJump elseJump condHasValue b isIfNil saveStack prevRange |
	stack last == CascadeFlag ifTrue: [^ self case: dist].
	prevRange _ self lastBytecodeRange.
	elsePc _ lastPc.
	elseStart _ pc + dist.
	end _ limit.
	"Check for bfp-jmp to invert condition.
	Don't be fooled by a loop with a null body."
	sign _ condition.
	savePc _ pc.
	((elseDist _ self interpretJump) notNil and: [elseDist >= 0 and: [elseStart = pc]])
		ifTrue: [sign _ sign not.  elseStart _ pc + elseDist].
	pc _ savePc.
	ifExpr _ stack removeLast.
	(stack size > 0 and: [stack last == IfNilFlag])
		ifTrue: [stack removeLast.  isIfNil _ true]
		ifFalse: [isIfNil _ false].
	saveStack _ stack.
	stack _ OrderedCollection new.
	thenBlock _ self blockTo: elseStart.
	condHasValue _ hasValue or: [isIfNil].
	"ensure jump is within block (in case thenExpr returns)"
	thenJump _ exit <= end ifTrue: [exit] ifFalse: [elseStart].
	"if jump goes back, then it's a loop"
	thenJump < elseStart ifTrue: [
		"Must be a while loop...
		thenJump will jump to the beginning of the while expr.  In the case of
		while's with a block in the condition, the while expr
		should include more than just the last expression: find all the
		statements needed by re-decompiling."
		stack _ saveStack.
		condHasValue ifTrue: [	"to-do loop"
			stack push: thenBlock statements last var referenceNode].
		pc _ thenJump.
		b _ self statementsTo: elsePc.
		condHasValue ifTrue: [  "to-do loop"
			statements add: ifExpr receiverNode.
			b _ b asOrderedCollection.
			b add: stack removeLast.
		] ifFalse: [
			"discard unwanted statements from block"
			b size - 1 timesRepeat: [statements removeLast]
		].
		statements addLast: (constructor
			codeMessage: (constructor codeBlock: b returns: false)
			selector: (constructor codeSelector: (sign ifTrue: [#whileFalse:] ifFalse: [#whileTrue:]) code: #macro)
			arguments: {thenBlock}
			range: prevRange).
		pc _ elseStart.
		self convertToDoLoop.
	] ifFalse: [
		"Must be a conditional..."
		elseBlock _ self blockTo: thenJump.
		elseJump _ exit.
		"if elseJump is backwards, it is not part of the elseExpr"
		elseJump < elsePc ifTrue: [pc _ lastPc].
		isIfNil ifTrue: [
			cond _ constructor
				codeMessage: ifExpr ifNilReceiver
				selector: (sign
					ifTrue: [constructor codeSelector: #ifNotNil: code: #macro]
					ifFalse: [constructor codeSelector: #ifNil: code: #macro])
				arguments: {thenBlock}
				range: prevRange.
		] ifFalse: [
			cond _ constructor
				codeMessage: ifExpr
				selector: (constructor codeSelector: #ifTrue:ifFalse: code: #macro)
				arguments: (sign ifTrue: [{elseBlock. thenBlock}] ifFalse: [{thenBlock. elseBlock}])
				range: prevRange.
		].
		stack _ saveStack.
		condHasValue
			ifTrue: [stack addLast: cond]
			ifFalse: [statements addLast: cond]
	].
! !

!Decompiler2 methodsFor: 'private' stamp: 'ajh 2/9/2003 17:04'!
lastTempVarIndex

	^ constructor lastTempVarIndex! !

!Decompiler2 methodsFor: 'private' stamp: 'ajh 2/4/2003 14:57'!
method

	^ method! !

!Decompiler2 methodsFor: 'private' stamp: 'ajh 2/4/2003 13:28'!
methodNode

	^ constructor methodNode! !

!Decompiler2 methodsFor: 'private' stamp: 'ajh 2/3/2003 18:35'!
outer

	^ outer! !

!Decompiler2 methodsFor: 'private' stamp: 'ajh 2/3/2003 23:24'!
outermost

	^ outer ifNil: [self] ifNotNil: [outer outermost]! !

!Decompiler2 methodsFor: 'private' stamp: 'ajh 1/28/2003 18:11'!
privNewConstructor

	^ DecompilerConstructor2 new decompiler: self! !

!Decompiler2 methodsFor: 'instruction decoding' stamp: 'ajh 2/4/2003 00:13'!
doStore: stackOrBlock
	"Only called internally, not from InstructionStream. StackOrBlock is stack
	for store, statements for storePop."

	| var expr |
	var _ stack removeLast.
	expr _ stack removeLast.
	(expr isMemberOf: NoopNode) ifTrue: [^ self].  "ignore"
	stackOrBlock addLast: (expr == ArgumentFlag
		ifTrue: [var]
		ifFalse: [constructor codeAssignTo: var value: expr])! !

!Decompiler2 methodsFor: 'instruction decoding' stamp: 'ajh 8/6/2002 16:57'!
popIntoLiteralVariable: assoc

	keepNodesBasic ifFalse: [^ super popIntoLiteralVariable: assoc].

	stack addLast: (constructor codeAnyLiteral: assoc).
	self send: #setInPool: super: false numArgs: 1.
	self doPop.
! !

!Decompiler2 methodsFor: 'instruction decoding' stamp: 'ajh 8/6/2002 13:15'!
pushLiteralVariable: assoc

	keepNodesBasic ifFalse: [^ super pushLiteralVariable: assoc].

	stack addLast: (constructor codeAnyLiteral: assoc).
	self send: #poolValue super: false numArgs: 0.
! !

!Decompiler2 methodsFor: 'instruction decoding' stamp: 'ajh 8/6/2002 16:57'!
storeIntoLiteralVariable: assoc

	keepNodesBasic ifFalse: [^ super storeIntoLiteralVariable: assoc].

	stack addLast: (constructor codeAnyLiteral: assoc).
	self send: #setInPool: super: false numArgs: 1.
! !


!Decompiler2 class methodsFor: 'as yet unclassified' stamp: 'ajh 8/6/2002 13:30'!
new

	^ super new initialize! !


!DecompilerConstructor2 methodsFor: 'initialize-release' stamp: 'ajh 2/9/2003 01:15'!
asBlockNode: outerTempVars

	capturedVars _ outerTempVars asOrderedCollection! !

!DecompilerConstructor2 methodsFor: 'initialize-release' stamp: 'ajh 8/12/2002 14:48'!
decompiler: aDecompiler

	decompiler _ aDecompiler! !

!DecompilerConstructor2 methodsFor: 'initialize-release' stamp: 'ajh 2/8/2003 18:22'!
initialize

	methodNode _ BlockNode2 new.
! !

!DecompilerConstructor2 methodsFor: 'initialize-release' stamp: 'ajh 2/8/2003 18:54'!
makeThisContextIn: statements reference: block

	statements do: [:s | 
		s allParseNodesDo: [:n |
			((n isMemberOf: VarReferenceNode) and: [n var isTemp and: [n var owningBlockNode notNil and: [n var isThisContextVariable]]]) ifTrue: [
				"Change from context of method to context of this block node"
				n var: block thisContextTempVar].
		] doChildrenIf: [:n | n isBlockNode not].
	].
! !

!DecompilerConstructor2 methodsFor: 'initialize-release' stamp: 'ajh 2/9/2003 00:24'!
method: aMethod class: aClassOrNil literals: literals

	classEnv _ aClassOrNil parseEnv.
	classEnv receiverVar: methodNode receiverTempVar.
	tempVarScopeDict _ IdentityDictionary new.
	tempNames ifNil: [tempNames _ #()].
	1 to: aMethod numArgs do: [:i |
		(methodNode addTempVarNamed: (i <= tempNames size
				ifTrue: [tempNames at: i]
				ifFalse: ['t', i printString])
			asArg: true) scopeIn: self set: 0
	].
	aMethod numArgs + 1 to: aMethod numTemps do: [:i |
		(methodNode addTempVarNamed: (i <= tempNames size
				ifTrue: [tempNames at: i]
				ifFalse: ['t', i printString])
			asArg: false) scopeIn: self set: 0
	].
! !

!DecompilerConstructor2 methodsFor: 'initialize-release' stamp: 'ajh 2/3/2003 22:35'!
tempNames: strings

	tempNames _ strings! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 10/4/2002 12:08'!
at: index
	"Protocol for codeConstants.  Answer one of the following nodes at index:
	{self, true, false, nil, -1, 0, 1, 2}"

	| node |
	node _ index caseOf: {
		[1] -> [methodNode receiverTempVar referenceNode].
		[2] -> [ConstantNode new object: true].
		[3] -> [ConstantNode new object: false].
		[4] -> [ConstantNode new object: nil]}
		otherwise: [ConstantNode new object: index - 6].
	node bytecodeRange: decompiler lastBytecodeRange.
	^ node! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 10/4/2002 12:08'!
codeAnyLitInd: association

	| node |
	node _ (PoolVarNode new poolVariable: association) referenceNode.
	node bytecodeRange: decompiler lastBytecodeRange.
	^ node
! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 2/4/2003 00:23'!
codeAnyLiteral: value

	| node |
	node _ (value isKindOf: BlockClosure) ifTrue: [
		node _ (Decompiler2 new
			outer: decompiler;
			asBlock;
			decompile: value method)
			startTempNamesAfter: self lastTempVarIndex
	] ifFalse: [
		node _ ConstantNode new object: value
	].
	node bytecodeRange: decompiler lastBytecodeRange.
	^ node
! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 7/28/2002 21:58'!
codeAnySelector: selector

	^ selector! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 10/4/2002 12:37'!
codeArguments: argRefs block: block range: bytecodeRange

	block arguments: (argRefs collect: [:varRef | varRef var]).
	block bytecodeRange: bytecodeRange.
	^ block
! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 10/4/2002 12:39'!
codeAssignTo: varReference value: expression

	| node |
	node _ VarAssignmentNode new variable: varReference var value: expression.
	node bytecodeRange: decompiler lastBytecodeRange.
	^ node
! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 2/4/2003 14:49'!
codeBlock: statements returns: returns

	| node |
	returns ifTrue: [
		(node _ statements last) isReturn ifFalse: [
			statements atLast: 1 put: (methodNode returnNode: node)].
	].
	node _ BlockNode2 new statements: statements asOrderedCollection.
	self makeThisContextIn: statements reference: node.
	^ node
! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 10/4/2002 12:39'!
codeBrace: elements

	| node |
	node _ ArrayNode new elements: elements.
	node bytecodeRange: decompiler lastBytecodeRange.
	^ node
! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 10/4/2002 12:39'!
codeCascade: receiver messages: messages

	| node |
	node _ (ArrayNode new matchBraceStreamReceiver: receiver messages: messages)
		ifNil: [CascadeMessageNode new receiver: receiver messages: messages].
	node bytecodeRange: decompiler lastBytecodeRange.
	^ node! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 10/4/2002 12:46'!
codeCascadedMessage: selector arguments: arguments

	^self
		codeMessage: nil
		selector: selector
		arguments: arguments
		range: decompiler lastBytecodeRange! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 7/29/2002 01:29'!
codeConstants
	"Answer with an array of the objects representing self, true, false, nil,
	-1, 0, 1, 2.
	 Actually, answer self that will respond to at: with the appropriate node"

	^ self! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 2/9/2003 12:20'!
codeInst: offset
	"Return instVar reference at offset.  If instVar doesn't exist (because class was unknown/UndefinedObject) then create it"

	| node i |
	node _ capturedVars ifNil: [
		InstVarNode new
			name: (classEnv instVarNameAt: offset + 1);
			instVar: offset + 1;
			selfVar: methodNode receiverTempVar
	] ifNotNil: [
		capturedVars
			at: (i _ offset + 1 - BlockClosure varOffset)
			ifAbsentAdd: [TempVarNode new name: 'free', i printString]
	].
	node _ node referenceNode.
	node bytecodeRange: decompiler lastBytecodeRange.
	^ node! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 2/3/2003 21:58'!
codeMessage: receiver selector: selector arguments: arguments range: bytecodeRange

	| node |
	node _ self
			matchReceiver: receiver
			specialSelector: selector
			arguments: arguments.
	node ifNil: [node _ ArrayNode new
			matchBraceWithReceiver: receiver
			selector: selector
			arguments: arguments].
	node ifNil: [node _ self decodeIfNilWithReceiver: receiver
			selector: selector
			arguments: arguments].
	node ifNil: [node _ MessageNode2 new
			receiver: receiver
			selector: selector
			arguments: arguments].
	node simplifyIfPossible.
	node bytecodeRange: bytecodeRange.
	^ node
! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 2/9/2003 16:04'!
codeMethod: selector block: block primitive: primitive class: class

	| primNode |
	self makeThisContextIn: block statements reference: methodNode.
	methodNode otherTempVars removeAllSuchThat: [:t | (t scopeIn: self) < 0].
	primNode _ PrimitiveNode new num: primitive.
	(primitive = 117 or: [primitive = 120]) ifTrue: [
		primNode spec: (decompiler method literalAt: 1)].
	selector ifNotNil: [methodNode selector: selector].
	methodNode
		statements: block statements;
		primitive: primNode;
		compiledMethod: decompiler method.
	capturedVars ifNotNil: [methodNode removeExplicitLocalReturn].
	^ methodNode! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 10/4/2002 13:07'!
codeReturnNode: parseNode

	| node |
	node _ methodNode returnNode: parseNode.
	^ node! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 7/29/2002 01:10'!
codeSelector: sel code: code

	^ sel! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 10/4/2002 13:17'!
codeSuper: selfNode

	| node |
	node _ selfNode asSuperNode: methodNode classEncoding.
	^ node! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 2/3/2003 22:28'!
codeTemp: index

	| node |
	node _ (methodNode tempVarAt: index) referenceNode.
	node bytecodeRange: decompiler lastBytecodeRange.
	^ node! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 10/4/2002 12:31'!
codeThisContext

	| node |
	node _ methodNode thisContextTempVar referenceNode.
	node bytecodeRange: decompiler lastBytecodeRange.
	^ node! !

!DecompilerConstructor2 methodsFor: 'constructor' stamp: 'ajh 8/23/2002 01:20'!
decodeIfNilWithReceiver: receiver selector: selector arguments: arguments

	selector == #ifTrue:ifFalse: ifFalse: [^ nil].
	(receiver isMessage: #== receiver: nil arguments: [:argNode | argNode = Parser2 nilNode]) ifFalse: [^ nil].
	^ MessageNode2 new
		receiver: receiver
		selector: #ifTrue:ifFalse:
		arguments: arguments! !

!DecompilerConstructor2 methodsFor: 'testing' stamp: 'ajh 7/29/2002 10:48'!
isAssignmentNode: node

	^ node isMemberOf: VarAssignmentNode! !

!DecompilerConstructor2 methodsFor: 'testing' stamp: 'ajh 8/9/2002 00:55'!
isThisContext: node

	^ node var isThisContextVariable! !

!DecompilerConstructor2 methodsFor: 'var scope' stamp: 'ajh 10/4/2002 21:44'!
scopeNumberOfVar: varNode

	^ tempVarScopeDict at: varNode! !

!DecompilerConstructor2 methodsFor: 'var scope' stamp: 'ajh 10/4/2002 21:44'!
scopeNumberOfVar: varNode is: n

	tempVarScopeDict at: varNode put: n! !

!DecompilerConstructor2 methodsFor: 'private' stamp: 'ajh 2/3/2003 23:59'!
lastTempVarIndex

	| d |
	^ methodNode argTempVars size + methodNode otherTempVars size + 
		((d _ decompiler outer) ifNil: [0] ifNotNil: [d lastTempVarIndex])! !

!DecompilerConstructor2 methodsFor: 'private' stamp: 'ajh 2/3/2003 23:53'!
matchReceiver: rcvrNode specialSelector: selector arguments: argNodes

	selector = (BlockNode2 selectorFor: argNodes size) ifTrue: [
		^ (Decompiler2 new
			outer: decompiler;
			asBlock: (argNodes collect: [:ref | ref var]);
			decompile: rcvrNode object)
			startTempNamesAfter: self lastTempVarIndex
	].
	selector = #remoteReturnTo: ifTrue: [
		^ decompiler outermost methodNode returnNode: rcvrNode
	].
	selector = #setInPool: ifTrue: [
		^ VarAssignmentNode new
			variable: (PoolVarNode new poolVariable: argNodes first object)
			value: rcvrNode
	].
	selector = #poolValue ifTrue: [
		^ (PoolVarNode new poolVariable: rcvrNode object) referenceNode
	].
	selector = #storeIn:instVar: ifTrue: [
		^ VarAssignmentNode new
			variable: (argNodes first var instVarNode: nil at: argNodes second object)
			value: rcvrNode
	].
	selector = #getInstVarAt: ifTrue: [
		^ (rcvrNode var instVarNode: nil at: argNodes first object) referenceNode
	].
	selector = #tempHolderValue ifTrue: [
		^ rcvrNode
	].
	selector = #storeInTempHolder: ifTrue: [
		^ VarAssignmentNode new
			variable: argNodes first var
			value: rcvrNode
	].
	selector = #wrapInTempHolder ifTrue: [
		^ NoopNode new
	].

	^ nil  "no match"! !

!DecompilerConstructor2 methodsFor: 'private' stamp: 'ajh 2/4/2003 13:28'!
methodNode

	^ methodNode! !


!DecompilerConstructor2 class methodsFor: 'as yet unclassified' stamp: 'ajh 1/31/2003 17:29'!
new

	^ super new initialize! !


!DupInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 16:58'!
bytecodesOn: stream

	stream dupTop
! !

!DupInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/3/2002 20:59'!
printOn: stream

	stream nextPutAll: 'dup'! !

!DupInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/27/2002 12:27'!
stackEffect: parseStack

	parseStack push! !


!FiniteAutomaton methodsFor: 'initializing' stamp: 'ajh 9/22/2002 17:22'!
actions: actionsArray transitions: transitionTable

	actions _ IdentityDictionary new.
	actionsArray withIndexDo: [:action :i |
		actions at: action put: i].

	states _ Array new: transitionTable size.
	transitions _ IdentityDictionary new.
	transitionTable withIndexDo: [:transitionPair :i |
		states at: i put: transitionPair first.
		transitions at: transitionPair first "from state" 
				 put: transitionPair second "to states indexed by action index"].
! !

!FiniteAutomaton methodsFor: 'accessing' stamp: 'ajh 9/22/2002 17:22'!
higherState: state1 or: state2

	^ (states indexOf: state1) >= (states indexOf: state2)
		ifTrue: [state1]
		ifFalse: [state2]! !

!FiniteAutomaton methodsFor: 'accessing' stamp: 'ajh 9/22/2002 17:53'!
state: state transition: action
	"Return new state after action transition from state"

	^ (transitions at: state) at: (actions at: action)! !


!InstrTempVar methodsFor: 'as yet unclassified' stamp: 'ajh 9/27/2002 08:42'!
fpOffset

	^ fpOffset! !

!InstrTempVar methodsFor: 'as yet unclassified' stamp: 'ajh 9/27/2002 08:42'!
fpOffset: int

	fpOffset _ int! !

!InstrTempVar methodsFor: 'as yet unclassified' stamp: 'ajh 9/26/2002 19:23'!
key

	^ key! !

!InstrTempVar methodsFor: 'as yet unclassified' stamp: 'ajh 9/26/2002 19:43'!
key: obj

	key _ obj! !

!InstrTempVar methodsFor: 'as yet unclassified' stamp: 'ajh 10/4/2002 15:39'!
name

	^ key name! !

!InstrTempVar methodsFor: 'as yet unclassified' stamp: 'ajh 9/26/2002 19:49'!
printOn: stream

	key printOn: stream! !


!InstructionBuilder methodsFor: 'initializing' stamp: 'ajh 1/18/2003 13:19'!
addTemp: tempKey

	self addTemps: {tempKey}! !

!InstructionBuilder methodsFor: 'initializing' stamp: 'ajh 1/18/2003 13:06'!
addTemps: tempKeys

	| fpOffset |
	fpOffset _ rargTemps size + otherTemps size - 1.
	otherTemps _ otherTemps, (tempKeys collect: [:key |
		(self newTempVarAt: key) fpOffset: (fpOffset _ fpOffset + 1)]).
! !

!InstructionBuilder methodsFor: 'initializing' stamp: 'ajh 1/21/2003 12:26'!
initialize

	tempVars _ KeyedSet new.
	literals _ LiteralList new.
	jumpAheadStacks _ IdentityDictionary new.
	jumpBackTargetStacks _ IdentityDictionary new.

	"Leave an empty sequence up front (guaranteed not to be in loop) so temp allocators can be added if necessary"
	startSequence _ InstructionSequence new method: self; orderNumber: 0.
	instrs _ InstructionSequence new method: self; orderNumber: 1.
	JumpInstr new destinationSequence: instrs; addTo: startSequence.
	nextSequenceNumber _ 1.  "pre-incremented"
	declareInstrs _ OrderedCollection new.
	contextVars _ OrderedCollection new.

	primitiveNode _ PrimitiveNode null.  "in case never set"
	otherTemps _ #().! !

!InstructionBuilder methodsFor: 'initializing' stamp: 'ajh 10/12/2002 19:15'!
primitive: aPrimitiveNode

	primitiveNode _ aPrimitiveNode! !

!InstructionBuilder methodsFor: 'initializing' stamp: 'ajh 1/18/2003 13:08'!
rargs: tempKeys
	"Receiver and args can now be referenced by push/storeTemp: tempKey"

	| fpOffset |
	fpOffset _ - 1.
	rargTemps _ tempKeys collect: [:key |
		(self newTempVarAt: key) fpOffset: (fpOffset _ fpOffset + 1)].
! !

!InstructionBuilder methodsFor: 'initializing' stamp: 'ajh 1/18/2003 13:05'!
thisContext: contextTempKey
	"Push temps with this key will be converted to pushThisContext"

	contextVars add: (self newTempVarAt: contextTempKey)! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 10/7/2002 15:23'!
createBlock: size

	CreateBlockInstr new
		closureSize: size;
		addTo: instrs! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 9/27/2002 13:56'!
declareAsTemp: tempKey
	"Current stack slot can be referenced later via push/storeTemp: tempKey"

	declareInstrs add:
		(DeclareTempInstr new
			tempVar: (self newTempVarAt: tempKey);
			addTo: instrs)! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 10/22/2002 00:06'!
doDup

	DupInstr new
		addTo: instrs! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 1/15/2003 21:28'!
doPop

	| i |
	(instrs notEmpty and:
	 [(((i _ instrs last) isMemberOf: StoreTempInstr) or: [i isMemberOf: StoreInstVarInstr]) and: [i pops not]]
	) ifTrue: [
		instrs localParseStack pop.
		i doPop.
		^ self
	].

	PopInstr new
		addTo: instrs! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 2/3/2003 23:44'!
getField: instVarIndex
	"Receiver must be on top"

	self lastWasPushReceiver ifTrue: [
		instrs last convertToPushInstVar: instVarIndex.
		^ self
	].
		
	self pushLiteral: instVarIndex.
	self send: #getInstVarAt:.
! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 9/24/2002 13:22'!
jumpAheadTarget: labelSymbol
	"Pop latest jumpAheadTo: with this labelSymbol and have it point to this new instruction sequence"

	| jumpInstr |
	self startNewSequence.
	jumpInstr _ (jumpAheadStacks at: labelSymbol ifAbsent: [self error: 'Missing jumpAheadTo: ', labelSymbol printString]) pop.
	jumpInstr destinationSequence: instrs.
! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 9/26/2002 15:37'!
jumpAheadTo: labelSymbol
	"Jump to the sequence that will be created when jumpAheadTarget: labelSymbol is sent to self.  This is and its corresponding target is only good for one use.  Other jumpAheadTo: with the same label will be put on a stack and superceed existing ones until its jumpAheadTarget: is called."

	"jumpAheadTarget: label will pop this and replace destination with its basic block"
	(jumpAheadStacks at: labelSymbol ifAbsentPut: [OrderedCollection new])
		push: (JumpInstr new addTo: instrs).
	self startNewSequence.
! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 9/26/2002 15:38'!
jumpAheadTo: labelSymbol if: boolean
	"Conditional jump to the sequence that will be created when jumpAheadTarget: labelSymbol is sent to self.  This and its corresponding target is only good for one use.  Other jumpAheadTo:... with the same label will be put on a stack and superceed existing ones until its jumpAheadTarget: is called."

	| i |
	"jumpAheadTarget: label will pop this and replace destination with its basic block"
	(jumpAheadStacks at: labelSymbol ifAbsentPut: [OrderedCollection new])
		push: (i _ JumpIfInstr new condition: boolean; addTo: instrs).
	self startNewSequence.
	i otherwiseSequence: instrs.
! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 9/24/2002 13:22'!
jumpBackTarget: labelSymbol
	"Remember this basic block for a future jumpBackTo: labelSymbol.  Stack up remembered targets with same name and remove them from stack for each jumpBackTo: called with same name."

	self startNewSequence.
	(jumpBackTargetStacks at: labelSymbol ifAbsentPut: [OrderedCollection new])
		push: instrs.
! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 9/26/2002 15:38'!
jumpBackTo: labelSymbol
	"Pop last remembered position with this label and write an unconditional jump to it"

	| instrSequence |
	instrSequence _ (jumpBackTargetStacks at: labelSymbol ifAbsent: [self error: 'Missing jumpBackTarget: ', labelSymbol printString]) pop.
	JumpInstr new destinationSequence: instrSequence; addTo: instrs.
	self startNewSequence.
! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 1/15/2003 15:34'!
localReturnTop

	(instrs notEmpty
	 and: [instrs last isPushSpecialConstant]) ifTrue: [
		instrs last convertToLocalReturnConstant.
		^ self startNewSequence].

	(instrs notEmpty
	 and: [(instrs last isMemberOf: PushTempInstr)
	 and: [instrs last tempVar == self receiverVar]]) ifTrue: [
		instrs last convertToLocalReturnSelf.
		^ self startNewSequence].
		
	LocalReturnTopInstr new addTo: instrs.
	self startNewSequence.
! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 10/4/2002 18:51'!
popTemp
	"ASSUMES top of stack is alligned with most recently delared temp.
	 Pop and remove most recently declared temp."

	self doPop.

	declareInstrs reverseDo: [:d |
		d deallocInstr isNil ifTrue: [
			d deallocInstr: instrs last.
			tempVars remove: d tempVar.
			^ self]
	].
	self error: 'no temp to pop'.
! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 1/15/2003 15:49'!
pushLiteral: obj

	PushConstantInstr new constant: obj; addTo: instrs! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 10/9/2002 13:46'!
pushTemp: tempKey

	| tempVar i |
	tempVar _ self tempVarAt: tempKey.
	(contextVars identityIncludes: tempVar)
		ifTrue: [^ self pushThisContext].

	(instrs notEmpty
	 and: [((i _ instrs last) isMemberOf: StoreTempInstr)
	 and: [i pops
	 and: [i tempVar == tempVar]]]
	) ifTrue: [
		instrs localParseStack push.
		i pops: false.
		^ self
	].
		
	PushTempInstr new
		tempVar: tempVar;
		addTo: instrs! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 10/9/2002 13:46'!
pushThisContext

	PushThisContextInstr new addTo: instrs! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 10/24/2002 17:03'!
remoteReturn
	"homeContext's contextTag is expected to be on top of stack above return result.  Return result to homeContext's sender"

	RemoteReturnTopInstr new addTo: instrs.
	self startNewSequence.
! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 1/15/2003 15:13'!
send: selector

	SendSelectorInstr new
		selector: selector;
		addTo: instrs! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 1/15/2003 15:14'!
send: selector toSuperOf: behavior

	SendSelectorSuperInstr new
		selector: selector;
		behavior: behavior;
		addTo: instrs! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 1/15/2003 21:53'!
setField: instVarIndex
	"receiver must be on top with new field value underneath"

	self lastWasPushReceiver ifTrue: [
		instrs last convertToStoreInstVar: instVarIndex.
		^ self
	].

	self pushLiteral: instVarIndex.
	self send: #storeIn:instVar:.
! !

!InstructionBuilder methodsFor: 'instructions' stamp: 'ajh 9/26/2002 19:39'!
storeTemp: tempKey

	StoreTempInstr new
		tempVar: (self tempVarAt: tempKey);
		addTo: instrs! !

!InstructionBuilder methodsFor: 'private' stamp: 'ajh 1/21/2003 12:29'!
addLastLiteral: object

	lastLiteral ifNil: [^ lastLiteral _ object].
	(lastLiteral literalEqual: object)
		ifFalse: [self error: 'there can only be one last literal'].! !

!InstructionBuilder methodsFor: 'private' stamp: 'ajh 9/24/2002 13:22'!
addLiteral: object

	literals add: object! !

!InstructionBuilder methodsFor: 'private' stamp: 'ajh 1/15/2003 21:52'!
lastWasPushReceiver

	^ (instrs last isMemberOf: PushTempInstr)
	   and: [instrs last tempVar == self receiverVar]! !

!InstructionBuilder methodsFor: 'private' stamp: 'ajh 1/16/2003 01:03'!
literalSlotIndexOf: object

	^ primitiveNode spec notNil asBit  "will occupy first literal if not nil"
	  + (literals indexOf: object ifAbsent: [self errorMissingLiteral: object])! !

!InstructionBuilder methodsFor: 'private' stamp: 'ajh 9/26/2002 19:27'!
newTempVarAt: tempKey

	^ tempVars add: (InstrTempVar new key: tempKey)! !

!InstructionBuilder methodsFor: 'private' stamp: 'ajh 9/24/2002 13:22'!
nextSequenceNumber

	^ nextSequenceNumber _ nextSequenceNumber + 1! !

!InstructionBuilder methodsFor: 'private' stamp: 'ajh 9/24/2002 13:22'!
startNewSequence
	"End current instruction sequence and start a new sequence to add instructions to.  If ending block just falls through to new block then add an explicit jump to it so they stay linked"

	| newInstrs |
	instrs isEmpty ifTrue: [^ self].  "block is still empty, continue using it"
	newInstrs _ InstructionSequence new method: self; orderNumber: self nextSequenceNumber.
	instrs last isJumpOrReturn ifFalse: [
		JumpInstr new destinationSequence: newInstrs; addTo: instrs].
	instrs _ newInstrs.
! !

!InstructionBuilder methodsFor: 'private' stamp: 'ajh 10/9/2002 13:45'!
startSequence

	^ startSequence! !

!InstructionBuilder methodsFor: 'private' stamp: 'ajh 9/26/2002 19:25'!
tempVarAt: tempKey

	^ tempVars at: tempKey! !

!InstructionBuilder methodsFor: 'accessing' stamp: 'ajh 9/24/2002 13:22'!
allSequences

	^ startSequence withAllSuccessors! !

!InstructionBuilder methodsFor: 'accessing' stamp: 'ajh 9/24/2002 13:22'!
endSequences

	^ self allSequences select: [:seq | seq last isReturnInstr]! !

!InstructionBuilder methodsFor: 'accessing' stamp: 'ajh 10/2/2002 01:54'!
isLocalTemp: tempKey
	"if not then tempKey must be a free var"

	^ tempVars includesKey: tempKey! !

!InstructionBuilder methodsFor: 'accessing' stamp: 'ajh 10/3/2002 20:34'!
longPrintOn: stream

	| seqs seqMap |
	seqs _ self allSequences.
	seqMap _ IdentityDictionary new: seqs size.
	seqs withIndexDo: [:b :i | seqMap at: b put: i].
	1 to: seqs size do: [:i |
		stream print: i; nextPut: $..
		(seqs at: i) do: [:instr |
			instr printOn: stream seqMap: seqMap.
		].
	].
! !

!InstructionBuilder methodsFor: 'accessing' stamp: 'ajh 9/28/2002 00:24'!
mapToObject

	^ mapToObject! !

!InstructionBuilder methodsFor: 'accessing' stamp: 'ajh 10/4/2002 18:59'!
numArgs

	^ rargTemps size - 1! !

!InstructionBuilder methodsFor: 'accessing' stamp: 'ajh 1/18/2003 13:14'!
numArgsAndTemps

	^ rargTemps size - 1 + otherTemps size! !

!InstructionBuilder methodsFor: 'accessing' stamp: 'ajh 10/4/2002 16:14'!
receiverArg

	^ self receiverVar key! !

!InstructionBuilder methodsFor: 'accessing' stamp: 'ajh 10/4/2002 16:14'!
receiverVar
	"first temp on stack"

	^ rargTemps first! !

!InstructionBuilder methodsFor: 'simplifying' stamp: 'ajh 9/24/2002 13:22'!
absorbConstantConditionalJumps

	startSequence absorbConstantConditionalJumps: IdentitySet new! !

!InstructionBuilder methodsFor: 'simplifying' stamp: 'ajh 9/24/2002 13:22'!
absorbJumpsToSingleInstrs

	startSequence absorbJumpToSingleInstr: IdentitySet new! !

!InstructionBuilder methodsFor: 'simplifying' stamp: 'ajh 1/18/2003 13:14'!
aggregateParseStack

	| toDo |
	startSequence parseStackIn: (ParseStack new push: self numArgsAndTemps).
	toDo _ OrderedCollection with: startSequence.
	[toDo isEmpty] whileFalse: [toDo removeFirst updateParseStackOut: toDo].
! !

!InstructionBuilder methodsFor: 'simplifying' stamp: 'ajh 9/27/2002 13:58'!
assignTempVarPositions

	declareInstrs do: [:instr | instr assignTempStackPosition]! !

!InstructionBuilder methodsFor: 'simplifying' stamp: 'ajh 1/18/2003 13:12'!
simplify

	instrs ifNil: [^ self].  "already simplified"
	(instrs isEmpty or: [instrs last isJumpOrReturn]) ifFalse: [self errorUnfinishedInstructionSequence: instrs].
	instrs _ nil.  "indicates self is done taking new instructions and is simplified"

	self aggregateParseStack.
	"self assignTempVarPositions."
	startSequence size = 1 ifTrue: [ "startSeq is just unconditional jump, forget it"
		startSequence _ startSequence last destinationSequence].
	self absorbJumpsToSingleInstrs.  "do before next to get in right form"
	self absorbConstantConditionalJumps.
	self absorbJumpsToSingleInstrs.  "do again since new opportunities may have arised after last step" 
! !

!InstructionBuilder methodsFor: 'bytecodes' stamp: 'ajh 9/24/2002 13:22'!
bytecodes
	"Translate VMInstructions into bytecodes"

	| sequences |
	sequences _ self allSequences.
	sequences do: [:seq | seq bytecodesWithoutJump].
	[ sequences inject: false into: [:changed :seq | 
		changed | (seq updateJumpBytecode: sequences)]
	] whileTrue.
	^ ByteArray streamContents: [:stream |
		sequences do: [:seq | seq bytecodesOn: stream]]! !

!InstructionBuilder methodsFor: 'bytecodes' stamp: 'ajh 1/22/2003 16:26'!
compiledMethodWith: trailer

	| method bytes i |
	self simplify.
	bytes _ self bytecodes.
	method _ CompiledMethod
		newBytes: bytes size
		trailerBytes: trailer 
		nArgs: self numArgs
		nTemps: self numArgsAndTemps
		nStack: self stackSize
		nLits: literals size + primitiveNode spec notNil asBit + lastLiteral notNil asBit
		primitive: (self quickMethodPrim ifNil: [primitiveNode num]).
	(WriteStream with: method)
		position: method initialPC - 1;
		nextPutAll: bytes.
	i _ 0.
	primitiveNode spec ifNotNil: [method literalAt: (i _ i + 1) put: primitiveNode spec].
	literals do: [:obj | method literalAt: (i _ i + 1) put: obj].
	lastLiteral ifNotNil: [method literalAt: (i _ i + 1) put: lastLiteral].
	method isClosureCompiled: true.
	^ method! !

!InstructionBuilder methodsFor: 'bytecodes' stamp: 'ajh 1/15/2003 20:50'!
quickMethodPrim

	| seq last |
	primitiveNode num = 0 ifFalse: [^ nil].
	self numArgs = 0 ifFalse: [^ nil].
	seq _ startSequence.
	last _ seq last.
	last isReturnInstr ifFalse: [^ nil].
	seq size = 1 ifTrue: [
		(last isMemberOf: LocalReturnSelfInstr) ifTrue: [^ 256].
		(last isMemberOf: LocalReturnSpecialConstantInstr) ifTrue: [
			^ 256 + (Compiler bytecodeWriteStream specialConstants identityIndexOf: last constant ifAbsent: [self halt: 'should have been found'])].
		^ nil
	].
	seq size = 2 ifFalse: [^ nil].
	(last isMemberOf: LocalReturnTopInstr) ifFalse: [^ nil].
	(seq first isMemberOf: PushInstVarInstr) ifTrue: [^ 263 + seq first slot].
	^ nil! !

!InstructionBuilder methodsFor: 'bytecodes' stamp: 'ajh 9/24/2002 13:22'!
stackSize
	"return the max amount the parse stack grows during execution of this method"

	^ self endSequences inject: 0 into: [:size :seq | size max: seq parseStackOut size]! !

!InstructionBuilder methodsFor: 'querying' stamp: 'ajh 10/12/2002 14:15'!
stackTempsBeforeBytecodeIp: ip

	| temps |
	self compiledMethod.  "map instructions to bytecodes"
	temps _ declareInstrs select: [:instr | 
		instr bytecodeRange last < ip and:
		[instr deallocInstr isNil or: [instr deallocInstr bytecodeRange first >= ip]]
	].
	^ temps collect: [:instr | instr tempVar]! !

!InstructionBuilder methodsFor: 'querying' stamp: 'ajh 1/20/2003 23:38'!
tempNames
	"Return arg and temp names in declaration order"

	^ (rargTemps allButFirst collect: [:var | var name])
	  , (otherTemps collect: [:var | var name])! !

!InstructionBuilder methodsFor: 'mapping' stamp: 'ajh 9/28/2002 00:51'!
mapTo: object
	"object will be sent the #mapInstr: message when a new instruction is added to self"

	mapToObject _ object.
	object initInstrMap.
! !

!InstructionBuilder methodsFor: 'mapping' stamp: 'ajh 9/28/2002 00:22'!
unMap
	"no object will be sent the #mapInstr: message"

	mapToObject _ nil! !


!InstructionBuilder class methodsFor: 'as yet unclassified' stamp: 'ajh 9/24/2002 13:22'!
new

	^ super new initialize! !


!InstructionSequence methodsFor: 'initializing' stamp: 'ajh 10/3/2002 10:41'!
initialize

	localParseStack _ ParseStack new.
	jumpBytecodes _ #().
! !

!InstructionSequence methodsFor: 'initializing' stamp: 'ajh 6/5/2002 18:25'!
method
	"Return my owning methodBuilder"

	^ method! !

!InstructionSequence methodsFor: 'initializing' stamp: 'ajh 6/5/2002 18:26'!
method: methodBuilder

	method _ methodBuilder! !

!InstructionSequence methodsFor: 'initializing' stamp: 'ajh 7/18/2002 01:23'!
orderNumber
	"Sequences are sorted by this number"

	^ orderNumber! !

!InstructionSequence methodsFor: 'initializing' stamp: 'ajh 7/18/2002 01:23'!
orderNumber: n
	"Sequences are sorted by this number"

	orderNumber _ n! !

!InstructionSequence methodsFor: 'comparing' stamp: 'ajh 6/18/2002 15:09'!
= other

	^ self == other! !

!InstructionSequence methodsFor: 'comparing' stamp: 'ajh 6/18/2002 15:09'!
hash

	^ self identityHash! !

!InstructionSequence methodsFor: 'successor sequences' stamp: 'ajh 6/13/2002 00:12'!
successorSequences

	^ self last successorSequences! !

!InstructionSequence methodsFor: 'successor sequences' stamp: 'ajh 9/26/2002 16:14'!
withAllSuccessors
	"Return me and all my successors sorted by sequence orderNumber"

	| list |
	list _ OrderedCollection new: 20.
	self withAllSuccessorsDo: [:seq | list add: seq].
	^ list asSortedCollection: [:x :y | x orderNumber <= y orderNumber]! !

!InstructionSequence methodsFor: 'successor sequences' stamp: 'ajh 7/18/2002 01:37'!
withAllSuccessorsDo: block
	"Iterate over me and all my successors only once"

	self withAllSuccessorsDo: block alreadySeen: IdentitySet new! !

!InstructionSequence methodsFor: 'successor sequences' stamp: 'ajh 7/18/2002 01:36'!
withAllSuccessorsDo: block alreadySeen: set
	"Iterate over me and all my successors only once"

	(set includes: self) ifTrue: [^ self].
	set add: self.
	block value: self.
	self successorSequences do: [:seq |
		seq withAllSuccessorsDo: block alreadySeen: set].
! !

!InstructionSequence methodsFor: 'simplifying' stamp: 'ajh 1/15/2003 15:46'!
absorbConstantConditionalJumps: alreadySeen
	"Collapse sequences that look like:
			[if] goto s1
			...
		 s1:	pushConst: true/false
			goto s2
		 s2:	if true/false goto s3 else s4
	into:
			[if] goto s3/s4

	These sequences are produced by and:/or: messages"

	| seq bool if |
	(alreadySeen includes: self) ifTrue: [^ self].
	alreadySeen add: self.

	[(seq _ self successorSequences) size > 0  "not return"
	  and: [(seq _ seq first "destinationSequence") size = 2
	   and: [(seq first isPushSpecialConstant and: [(bool _ seq first constant) isKindOf: Boolean])
	    and: [(seq last isMemberOf: JumpInstr)
	     and: [(if _ seq last destinationSequence first) isMemberOf: JumpIfInstr]]]]
	] whileTrue: [ "absorb"
		self last destinationSequence: (bool == if condition
			ifTrue: [if destinationSequence]
			ifFalse: [if otherwiseSequence]).
	].

	self successorSequences do: [:instrs | instrs absorbConstantConditionalJumps: alreadySeen].
! !

!InstructionSequence methodsFor: 'simplifying' stamp: 'ajh 7/18/2002 15:33'!
absorbJumpToSingleInstr: alreadySeen
	"Collapse jumps to single return instructions into caller"

	| seqs seq |
	(alreadySeen includes: self) ifTrue: [^ self].
	alreadySeen add: self.

	[ (seqs _ self successorSequences) size = 1  "unconditional jump..." 
	   and: [(seq _ seqs first) size = 1  "...to single instruction..."
	    and: [(self orderNumber + 1 ~= seq orderNumber or: [seq successorSequences size = 1])  "...but leave returns and ifs alone if they come next (so they aren't repeated)..."
	     and: [seq successorSequences size < 2]]]  "...and don't collapse conditional jumps so their otherwiseSequences can stay right after them"
	] whileTrue: [ "replace goto with single instruction"
		self removeLast.
		seq do: [:instr | instr copy addTo: self].
	].

	seqs do: [:instrs | instrs absorbJumpToSingleInstr: alreadySeen].
! !

!InstructionSequence methodsFor: 'parse stack' stamp: 'ajh 9/27/2002 14:32'!
localParseStack

	^ localParseStack! !

!InstructionSequence methodsFor: 'parse stack' stamp: 'ajh 6/6/2002 14:59'!
mergeParseStackIn: aParseStack
	"Return true if this merge changes my parseStackIn, otherwise return false"

	parseStackIn ifNil: [parseStackIn _ aParseStack.  ^ true].
	^ parseStackIn mergeStack: aParseStack! !

!InstructionSequence methodsFor: 'parse stack' stamp: 'ajh 6/6/2002 14:45'!
parseStackIn

	^ parseStackIn! !

!InstructionSequence methodsFor: 'parse stack' stamp: 'ajh 6/6/2002 14:56'!
parseStackIn: parseStack

	parseStackIn _ parseStack! !

!InstructionSequence methodsFor: 'parse stack' stamp: 'ajh 6/17/2002 16:00'!
parseStackOut

	^ parseStackOut! !

!InstructionSequence methodsFor: 'parse stack' stamp: 'ajh 6/17/2002 15:59'!
updateParseStackOut: toDo
	"addStack: will raise an error if underflow occurs.
	 mergeStackIn: will raise an error if stack positions don't match"

	parseStackOut _ parseStackIn copy addStack: self localParseStack.
	self successorSequences do: [:seq |
		(seq mergeParseStackIn: parseStackOut) ifTrue: [toDo addIfAbsent: seq]].
! !

!InstructionSequence methodsFor: 'bytecodes' stamp: 'ajh 9/27/2002 21:15'!
bytecodesOn: stream

	startBytecodePosition _ stream position.
	stream nextPutAll: self bytecodesWithoutJump.
	stream nextPutAll: jumpBytecodes.
! !

!InstructionSequence methodsFor: 'bytecodes' stamp: 'ajh 1/26/2003 15:02'!
bytecodesSize

	^ (self bytecodesWithoutJump size + jumpBytecodes size) max: self notEmpty asBit  "in case of single goto whose jump size hasn't been calculated yet.  1 is needed since 0 breaks BytecodeV3WriteStream>>#jump:if:"! !

!InstructionSequence methodsFor: 'bytecodes' stamp: 'ajh 1/15/2003 17:39'!
bytecodesWithoutJump
	"MacroJumpInstr adds nothing in bytecodesOn:"

	^ bytecodesWithoutJump ifNil: [
		| stream |
		stream _ Compiler bytecodeWriteStream on: (ByteArray new: 50).
		self do: [:instr | instr mapBytecodesOn: stream].
		bytecodesWithoutJump _ stream contents.
	  ]! !

!InstructionSequence methodsFor: 'bytecodes' stamp: 'ajh 9/27/2002 21:36'!
startBytecodePosition

	^ startBytecodePosition! !

!InstructionSequence methodsFor: 'bytecodes' stamp: 'ajh 9/27/2002 21:29'!
updateJumpBytecode: allSequences
	"Recalculate final jump bytecodes.  Return true if jump bytecodes SIZE has changed, otherwise return false"

	| newBytecodes sizeChanged |
	self last isReturnInstr ifTrue: [^ false].

	newBytecodes _ self last jumpBytecodes: allSequences.
	sizeChanged _ newBytecodes size ~= jumpBytecodes size.
	jumpBytecodes _ newBytecodes.
	^ sizeChanged! !

!InstructionSequence methodsFor: 'printing' stamp: 'ajh 9/27/2002 16:07'!
longPrintOn: stream

	self class printOn: stream.
	stream space.
	stream nextPut: $(.
	stream nextPutAll: self psuedoPrintId.
	stream nextPut: $).
	stream cr.
	self do: [:instr |
		stream space.
		instr printOn: stream.
		stream cr].
! !

!InstructionSequence methodsFor: 'printing' stamp: 'ajh 9/27/2002 16:09'!
printOn: stream

	stream nextPutAll: 'an '.
	self class printOn: stream.
	stream space.
	stream nextPut: $(.
	stream nextPutAll: self psuedoPrintId.
	stream nextPut: $).
! !

!InstructionSequence methodsFor: 'printing' stamp: 'ajh 7/18/2002 01:41'!
psuedoPrintId

	^ 's', orderNumber printString! !

!InstructionSequence methodsFor: 'querying' stamp: 'ajh 9/27/2002 20:58'!
bytecodeRange

	^ self first bytecodeRange first to: self last bytecodeRange last! !

!InstructionSequence methodsFor: 'querying' stamp: 'ajh 8/11/2002 20:57'!
containsBytecodeIp: ip

	^ self bytecodeRange includes: ip! !


!InstructionSequence class methodsFor: 'as yet unclassified' stamp: 'ajh 6/7/2002 14:09'!
new

	^ super new initialize! !


!Interpreter methodsFor: 'contexts' stamp: 'ajh 1/28/2003 12:58'!
primitiveTerminateTo
	"Primitive. Terminate up the context stack from the receiver up to but not including the argument, if previousContext is on my Context stack. Make previousContext my sender. This prim has to shadow the code in ContextPart>terminateTo: to be correct"
	| thisCntx currentCntx aContext nextCntx nilOop |
	aContext _ self popStack.
	thisCntx _ self popStack.

	"make sure that aContext is in my chain"
	(self context: thisCntx hasSender: aContext) ifTrue:[
		nilOop _ nilObj.
		currentCntx _ self fetchPointer: SenderIndex ofObject: thisCntx.
		[currentCntx = aContext] whileFalse: [
			nextCntx _ self fetchPointer: SenderIndex ofObject: currentCntx.
			self storePointer: SenderIndex ofObject: currentCntx withValue: nilOop.
			self storePointer: InstructionPointerIndex ofObject: currentCntx withValue: nilOop.
			currentCntx _ nextCntx]].

	self storePointer: SenderIndex ofObject: thisCntx withValue: aContext.
	^self push: thisCntx! !

!Interpreter methodsFor: 'control primitives' stamp: 'ajh 1/28/2003 13:01'!
primitiveClosureValue

	| blockClosure |
	blockClosure _ self stackValue: argumentCount.
	newMethod _ self fetchPointer: BlockMethodIndex ofObject: blockClosure.
	primitiveIndex _ self primitiveIndexOf: newMethod.
	self success: (argumentCount = (self argumentCountOf: newMethod)).
	successFlag ifTrue: [self executeNewMethod].
! !

!Interpreter methodsFor: 'control primitives' stamp: 'ajh 1/28/2003 12:51'!
primitiveClosureValueWithArgs

	| blockClosure |
	blockClosure _ self stackValue: 1.
	self push: (self fetchPointer: BlockMethodIndex ofObject: blockClosure).
	self primitiveExecuteMethod.
	successFlag ifFalse: [self pop: 1].
! !

!Interpreter methodsFor: 'control primitives' stamp: 'ajh 1/28/2003 12:39'!
primitiveExecuteMethod
	"receiver, argsArray, then method are on top of stack.  Execute method against receiver and args"

	| argCnt argumentArray |
	newMethod _ self popStack.
	primitiveIndex _ self primitiveIndexOf: newMethod.
	argCnt _ self argumentCountOf: newMethod.
	argumentArray _ self popStack.
	self assertClassOf: argumentArray is: (self splObj: ClassArray).
	successFlag ifTrue: [
		self success: (argCnt = (self fetchWordLengthOf: argumentArray))].
	successFlag ifTrue: [
		self transfer: argCnt
			from: argumentArray + BaseHeaderSize
			to: stackPointer + 4.
		self unPop: argCnt.
		argumentCount _ argCnt.
		self executeNewMethod.
	] ifFalse: [self unPop: 2].
! !


!Interpreter class methodsFor: 'initialization' stamp: 'ajh 1/28/2003 12:08'!
initializeContextIndices
	"Class MethodContext"
	SenderIndex _ 0.
	InstructionPointerIndex _ 1.
	StackPointerIndex _ 2.
	MethodIndex _ 3.
	ReceiverIndex _ 5.
	TempFrameStart _ 6.
	"Class BlockContext"
	CallerIndex _ 0.
	BlockArgumentCountIndex _ 3.
	InitialIPIndex _ 4.
	HomeIndex _ 5.

	"Class BlockClosure"
	BlockMethodIndex _ 0.
! !

!Interpreter class methodsFor: 'initialization' stamp: 'ajh 1/28/2003 12:22'!
initializePrimitiveTable 
	"This table generates a C switch statement for primitive dispatching."

	"NOTE: The real limit here is 2047, but our C compiler currently barfs over 700"
	MaxPrimitiveIndex _ 700.
	PrimitiveTable _ Array new: MaxPrimitiveIndex + 1.
	self table: PrimitiveTable from: 
	#(	"Integer Primitives (0-19)"
		(0 primitiveFail)
		(1 primitiveAdd)
		(2 primitiveSubtract)
		(3 primitiveLessThan)
		(4 primitiveGreaterThan)
		(5 primitiveLessOrEqual)
		(6 primitiveGreaterOrEqual)
		(7 primitiveEqual)
		(8 primitiveNotEqual)
		(9 primitiveMultiply)
		(10 primitiveDivide)
		(11 primitiveMod)
		(12 primitiveDiv)
		(13 primitiveQuo)
		(14 primitiveBitAnd)
		(15 primitiveBitOr)
		(16 primitiveBitXor)
		(17 primitiveBitShift)
		(18 primitiveMakePoint)
		(19 primitiveFail)					"Guard primitive for simulation -- *must* fail"

		"LargeInteger Primitives (20-39)"
		"32-bit logic is aliased to Integer prims above"
		(20 39 primitiveFail)

		"Float Primitives (40-59)"
		(40 primitiveAsFloat)
		(41 primitiveFloatAdd)
		(42 primitiveFloatSubtract)
		(43 primitiveFloatLessThan)
		(44 primitiveFloatGreaterThan)
		(45 primitiveFloatLessOrEqual)
		(46 primitiveFloatGreaterOrEqual)
		(47 primitiveFloatEqual)
		(48 primitiveFloatNotEqual)
		(49 primitiveFloatMultiply)
		(50 primitiveFloatDivide)
		(51 primitiveTruncated)
		(52 primitiveFractionalPart)
		(53 primitiveExponent)
		(54 primitiveTimesTwoPower)
		(55 primitiveSquareRoot)
		(56 primitiveSine)
		(57 primitiveArctan)
		(58 primitiveLogN)
		(59 primitiveExp)

		"Subscript and Stream Primitives (60-67)"
		(60 primitiveAt)
		(61 primitiveAtPut)
		(62 primitiveSize)
		(63 primitiveStringAt)
		(64 primitiveStringAtPut)
		(65 primitiveNext)
		(66 primitiveNextPut)
		(67 primitiveAtEnd)

		"StorageManagement Primitives (68-79)"
		(68 primitiveObjectAt)
		(69 primitiveObjectAtPut)
		(70 primitiveNew)
		(71 primitiveNewWithArg)
		(72 primitiveArrayBecomeOneWay)	"Blue Book: primitiveBecome"
		(73 primitiveInstVarAt)
		(74 primitiveInstVarAtPut)
		(75 primitiveAsOop)
		(76 primitiveStoreStackp)					"Blue Book: primitiveAsObject"
		(77 primitiveSomeInstance)
		(78 primitiveNextInstance)
		(79 primitiveNewMethod)

		"Control Primitives (80-89)"
		(80 primitiveBlockCopy)
		(81 primitiveValue)
		(82 primitiveValueWithArgs)
		(83 primitivePerform)
		(84 primitivePerformWithArgs)
		(85 primitiveSignal)
		(86 primitiveWait)
		(87 primitiveResume)
		(88 primitiveSuspend)
		(89 primitiveFlushCache)

		"Input/Output Primitives (90-109)"
		(90 primitiveMousePoint)
		(91 primitiveTestDisplayDepth)			"Blue Book: primitiveCursorLocPut"
		(92 primitiveSetDisplayMode)				"Blue Book: primitiveCursorLink"
		(93 primitiveInputSemaphore)
		(94 primitiveGetNextEvent)				"Blue Book: primitiveSampleInterval"
		(95 primitiveInputWord)
		(96 primitiveObsoleteIndexedPrimitive)	"primitiveCopyBits"
		(97 primitiveSnapshot)
		(98 primitiveStoreImageSegment)
		(99 primitiveLoadImageSegment)
		(100 primitivePerformInSuperclass)		"Blue Book: primitiveSignalAtTick"
		(101 primitiveBeCursor)
		(102 primitiveBeDisplay)
		(103 primitiveScanCharacters)
		(104 primitiveObsoleteIndexedPrimitive)	"primitiveDrawLoop"
		(105 primitiveStringReplace)
		(106 primitiveScreenSize)
		(107 primitiveMouseButtons)
		(108 primitiveKbdNext)
		(109 primitiveKbdPeek)

		"System Primitives (110-119)"
		(110 primitiveEquivalent)
		(111 primitiveClass)
		(112 primitiveBytesLeft)
		(113 primitiveQuit)
		(114 primitiveExitToDebugger)
		(115 primitiveChangeClass)					"Blue Book: primitiveOopsLeft"
		(116 primitiveFlushCacheByMethod)
		(117 primitiveExternalCall)
		(118 primitiveDoPrimitiveWithArgs)
		(119 primitiveFlushCacheSelective)
			"Squeak 2.2 and earlier use 119.  Squeak 2.3 and later use 116.
			Both are supported for backward compatibility."

		"Miscellaneous Primitives (120-127)"
		(120 primitiveCalloutToFFI)
		(121 primitiveImageName)
		(122 primitiveNoop)					"Blue Book: primitiveImageVolume"
		(123 primitiveValueUninterruptably)	"@@@: Remove this when all VMs have support"
		(124 primitiveLowSpaceSemaphore)
		(125 primitiveSignalAtBytesLeft)

		"Squeak Primitives Start Here"

		"Squeak Miscellaneous Primitives (128-149)"
		(126 primitiveDeferDisplayUpdates)
		(127 primitiveShowDisplayRect)
		(128 primitiveArrayBecome)
		(129 primitiveSpecialObjectsOop)
		(130 primitiveFullGC)
		(131 primitiveIncrementalGC)
		(132 primitiveObjectPointsTo)
		(133 primitiveSetInterruptKey)
		(134 primitiveInterruptSemaphore)
		(135 primitiveMillisecondClock)
		(136 primitiveSignalAtMilliseconds)
		(137 primitiveSecondsClock)
		(138 primitiveSomeObject)
		(139 primitiveNextObject)
		(140 primitiveBeep)
		(141 primitiveClipboardText)
		(142 primitiveVMPath)
		(143 primitiveShortAt)
		(144 primitiveShortAtPut)
		(145 primitiveConstantFill)
		"NOTE: When removing the obsolete indexed primitives,
		the following two should go become #primitiveIntegerAt / atPut"
		(146 primitiveObsoleteIndexedPrimitive)	"primitiveReadJoystick"
		(147 primitiveObsoleteIndexedPrimitive)	"primitiveWarpBits"
		(148 primitiveClone)
		(149 primitiveGetAttribute)

		"File Primitives (150-169) - NO LONGER INDEXED"
		(150 164 primitiveObsoleteIndexedPrimitive)
		(165 primitiveIntegerAt)		"hacked in here for now"
		(166 primitiveIntegerAtPut)
		(167 168 primitiveFail)
		(169 primitiveObsoleteIndexedPrimitive)

		"Sound Primitives (170-199) - NO LONGER INDEXED"
		(170 185 primitiveObsoleteIndexedPrimitive)
		"(186 188 primitiveFail)"

		"Closure primitives"
		(186 primitiveClosureValue)
		(187 primitiveClosureValueWithArgs)
		(188 primitiveExecuteMethod)

		"Soun Primitives (continued) - NO LONGER INDEXED"
		(189 194 primitiveObsoleteIndexedPrimitive)

		"Unwind primitives"
		(195 primitiveFindNextUnwindContext)
		(196 primitiveTerminateTo)
		(197 primitiveFindHandlerContext)
		(198 primitiveMarkUnwindMethod)
		(199 primitiveMarkHandlerMethod)

		"Networking Primitives (200-229) - NO LONGER INDEXED"
		(200 225 primitiveObsoleteIndexedPrimitive)
		(226 229 primitiveFail)

		"Other Primitives (230-249)"
		(230 primitiveRelinquishProcessor)
		(231 primitiveForceDisplayUpdate)
		(232 primitiveFormPrint)
		(233 primitiveSetFullScreen)
		(234 primitiveObsoleteIndexedPrimitive) "primBitmapdecompressfromByteArrayat"
		(235 primitiveObsoleteIndexedPrimitive) "primStringcomparewithcollated"
		(236 primitiveObsoleteIndexedPrimitive) "primSampledSoundconvert8bitSignedFromto16Bit"
		(237 primitiveObsoleteIndexedPrimitive) "primBitmapcompresstoByteArray"
		(238 241 primitiveObsoleteIndexedPrimitive) "serial port primitives"
		(242 primitiveFail)
		(243 primitiveObsoleteIndexedPrimitive) "primStringtranslatefromtotable"
		(244 primitiveObsoleteIndexedPrimitive) "primStringfindFirstInStringinSetstartingAt"
		(245 primitiveObsoleteIndexedPrimitive) "primStringindexOfAsciiinStringstartingAt"
		(246 primitiveObsoleteIndexedPrimitive) "primStringfindSubstringinstartingAtmatchTable"
		(247 primitiveSnapshotEmbedded)
		(248 249 primitiveFail)

		"VM Implementor Primitives (250-255)"
		(250 clearProfile)
		(251 dumpProfile)
		(252 startProfiling)
		(253 stopProfiling)
		(254 primitiveVMParameter)
		(255 primitiveInstVarsPutFromStack) "Never used except in Disney tests.  Remove after 2.3 release."

		"Quick Push Const Methods"
		(256 primitivePushSelf)
		(257 primitivePushTrue)
		(258 primitivePushFalse)
		(259 primitivePushNil)
		(260 primitivePushMinusOne)
		(261 primitivePushZero)
		(262 primitivePushOne)
		(263 primitivePushTwo)

		"Quick Push Const Methods"
		(264 519 primitiveLoadInstVar)

		"MIDI Primitives (520-539) - NO LONGER INDEXED"
		(520 529 primitiveObsoleteIndexedPrimitive)
		(530 539 primitiveFail)  "reserved for extended MIDI primitives"

		"Experimental Asynchrous File Primitives - NO LONGER INDEXED"
		(540 545 primitiveObsoleteIndexedPrimitive)
		(546 547 primitiveFail)

		"Pen Tablet Primitives - NO LONGER INDEXED"
		(548 primitiveObsoleteIndexedPrimitive)
		(549 primitiveObsoleteIndexedPrimitive)

		"Sound Codec Primitives - NO LONGER INDEXED"
		(550 553 primitiveObsoleteIndexedPrimitive)
		(554 569 primitiveFail)

		"External primitive support primitives"
		(570 primitiveFlushExternalPrimitives)
		(571 primitiveUnloadModule)
		(572 primitiveListBuiltinModule)
		(573 primitiveListExternalModule)
		(574 primitiveFail) "reserved for addl. external support prims"

		"Unassigned Primitives"
		(575 700 primitiveFail)).
! !


!JumpInstr methodsFor: 'as yet unclassified' stamp: 'ajh 8/6/2002 20:08'!
bytecodesOn: stream
	"do nothing here.  see jumpBytecodes:"! !

!JumpInstr methodsFor: 'as yet unclassified' stamp: 'ajh 7/16/2002 20:58'!
destinationSequence

	^ destinationSequence! !

!JumpInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/5/2002 18:13'!
destinationSequence: instructionSequence

	destinationSequence _ instructionSequence! !

!JumpInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/13/2002 00:04'!
isJumpOrReturn

	^ true! !

!JumpInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 17:40'!
jumpBytecodes: sequenceOrder

	| from to distance stream bytes |
	from _ sequenceOrder identityIndexOf: owningSequence.
	to _ sequenceOrder identityIndexOf: destinationSequence.
	from + 1 = to ifTrue: [
		"fall through, no jump needed"
		localBytecodeRange _ owningSequence bytecodesWithoutJump size + 1 extend: 0.
		^ #()
	].
	stream _ Compiler bytecodeWriteStream on: (ByteArray new: 4).
	from < to ifTrue: [ "jump forward"
		distance _ (from + 1 to: to - 1) inject: 0 into: [:size :i | 
				size + (sequenceOrder at: i) bytecodesSize].
		stream jumpForward: distance.
	] ifFalse: [ "jump backward"
		distance _ ((to to: from - 1) inject: 0 into: [:size :i |
				size + (sequenceOrder at: i) bytecodesSize])
			+ owningSequence bytecodesWithoutJump size.
		stream jumpBackward: distance.
	].
	bytes _ stream contents.
	localBytecodeRange _ owningSequence bytecodesWithoutJump size + 1 extend: bytes size.
	^ bytes! !

!JumpInstr methodsFor: 'as yet unclassified' stamp: 'ajh 7/16/2002 20:41'!
printOn: stream

	stream nextPutAll: 'goto '.
	stream nextPutAll: destinationSequence psuedoPrintId.
! !

!JumpInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/3/2002 20:34'!
printOn: stream seqMap: seqToIndexMap

	| myNum destNum |
	myNum _ seqToIndexMap at: owningSequence.
	destNum _ seqToIndexMap at: destinationSequence.
	myNum + 1 = destNum ifTrue: [^ self].  "fall through don't print anything"
	stream tab.
	stream nextPutAll: 'goto '.
	destNum printOn: stream.
	stream cr.
! !

!JumpInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/5/2002 18:50'!
stackEffect: parseStack

	"no effect"! !

!JumpInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/13/2002 00:14'!
successorSequences

	^ {destinationSequence}! !


!JumpIfInstr methodsFor: 'as yet unclassified' stamp: 'ajh 7/16/2002 21:01'!
condition

	^ jumpCondition! !

!JumpIfInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/5/2002 18:32'!
condition: boolean

	jumpCondition _ boolean! !

!JumpIfInstr methodsFor: 'as yet unclassified' stamp: 'ajh 2/5/2003 12:08'!
jumpBytecodes: sequenceOrder

	| from to distance continue stream bytes |
	from _ sequenceOrder identityIndexOf: owningSequence.
	to _ sequenceOrder identityIndexOf: destinationSequence.
	from < to ifFalse: [self errorConditionalJumpBackwards].
	continue _ sequenceOrder identityIndexOf: otherwiseSequence.
	from + 1 = continue ifFalse: [self errorFallThroughSequenceNotNext].
	stream _ Compiler bytecodeWriteStream on: (ByteArray new: 4).
	to = continue ifTrue: [
		"jumps to fall-through branch, condition irrelevant.  Eg. x ifTrue: []."
		stream popTop.
	] ifFalse: [
		distance _ (from + 1 to: to - 1) inject: 0 into: [:size :i | 
			size + (sequenceOrder at: i) bytecodesSize].
		stream jump: distance if: jumpCondition.
	].
	bytes _ stream contents.
	localBytecodeRange _ owningSequence bytecodesWithoutJump size + 1 extend: bytes size.
	^ bytes! !

!JumpIfInstr methodsFor: 'as yet unclassified' stamp: 'ajh 7/16/2002 21:03'!
otherwiseSequence

	^ otherwiseSequence! !

!JumpIfInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/5/2002 18:32'!
otherwiseSequence: instructionSequence

	otherwiseSequence _ instructionSequence! !

!JumpIfInstr methodsFor: 'as yet unclassified' stamp: 'ajh 8/7/2002 12:02'!
printOn: stream

	stream nextPutAll: 'if '.
	jumpCondition printOn: stream.
	stream nextPutAll: ' goto '.
	stream nextPutAll: destinationSequence psuedoPrintId.
	stream nextPutAll: ' else '.
	stream nextPutAll: otherwiseSequence psuedoPrintId.
! !

!JumpIfInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/3/2002 20:35'!
printOn: stream seqMap: seqToIndexMap

	| myNum othNum |
	stream tab.
	stream nextPutAll: 'if '.
	jumpCondition printOn: stream.
	stream nextPutAll: ' goto '.
	(seqToIndexMap at: destinationSequence) printOn: stream.
	myNum _ seqToIndexMap at: owningSequence.
	othNum _ seqToIndexMap at: otherwiseSequence.
	myNum + 1 = othNum ifTrue: [^ stream cr].  "falls through, don't print else"
	stream nextPutAll: ' else '.
	othNum printOn: stream.
	stream cr.
! !

!JumpIfInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/6/2002 13:37'!
stackEffect: parseStack

	parseStack pop! !

!JumpIfInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/13/2002 00:14'!
successorSequences

	^ {destinationSequence. otherwiseSequence}! !


!LiteralList methodsFor: 'adding' stamp: 'ajh 1/21/2003 12:24'!
addLast: lit
	"If the end is out of optimal range for lit and there is still some slots left for its kind in its optimal range the slide it down to the last slot in optimal range"

	(equalitySet includes: lit) ifTrue: [^ lit].
	equalitySet add: lit.
	super addLast: lit.
	^ lit
! !

!LiteralList methodsFor: 'private' stamp: 'ajh 1/21/2003 12:21'!
setCollection: anArray

	super setCollection: anArray.
	equalitySet _ LiteralSet new: anArray size.
! !

!LiteralList methodsFor: 'accessing' stamp: 'ajh 1/21/2003 12:21'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock

	start to: self size do:
		[:index | ((self at: index) literalEqual: anElement) ifTrue: [^ index]].
	^ exceptionBlock value! !


!LiteralSet methodsFor: 'as yet unclassified' stamp: 'ajh 12/9/2001 16:03'!
add: newObject
	"Include newObject as one of the receiver's elements.  If equivalent is already present don't add and return equivalent object"

	| index |
	newObject ifNil: [self error: 'Sets cannot meaningfully contain nil as an element'].
	index _ self findElementOrNil: newObject.
	^ (array at: index) 
		ifNil: [self atNewIndex: index put: newObject. newObject]
		ifNotNil: [array at: index]! !

!LiteralSet methodsFor: 'as yet unclassified' stamp: 'ajh 2/2/2002 19:16'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."
	| element start finish |
	start _ (anObject hash \\ array size) + 1.
	finish _ array size.

	"Search from (hash mod size) to the end."
	start to: finish do:
		[:index | ((element _ array at: index) == nil
					or: [element literalEqual: anObject])
					ifTrue: [^ index ]].

	"Search from 1 to where we started."
	1 to: start-1 do:
		[:index | ((element _ array at: index) == nil
					or: [element literalEqual: anObject])
					ifTrue: [^ index ]].

	^ 0  "No match AND no empty slot"! !


!LocalReturnInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 10/8/2002 16:32'!
stackEffect: parseStack

	"no effect (does not matter since returning)"! !

!LocalReturnInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 6/13/2002 00:13'!
successorSequences

	^ #()! !

!LocalReturnInstruction methodsFor: 'testing' stamp: 'ajh 6/13/2002 00:03'!
isJumpOrReturn

	^ true! !

!LocalReturnInstruction methodsFor: 'testing' stamp: 'ajh 6/13/2002 10:28'!
isReturnInstr

	^ true! !


!LocalReturnSelfInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/6/2002 15:34'!
bytecodesOn: stream

	stream localReturnSelf: self spBefore! !

!LocalReturnSelfInstr methodsFor: 'as yet unclassified' stamp: 'ajh 8/7/2002 11:52'!
printOn: stream

	stream nextPutAll: 'localReturnSelf'.
! !


!LocalReturnSpecialConstantInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 17:40'!
bytecodesOn: stream

	(Compiler bytecodeWriteStream specialReturnConstants identityIncludes: constant) ifTrue: [
		stream localReturn: self spBefore const: constant.
	] ifFalse: [
		stream pushConst: constant.
		stream localReturnTop: self spBefore + 1.
	].
! !

!LocalReturnSpecialConstantInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:57'!
constant
	"true, false, or nil"

	^ constant! !

!LocalReturnSpecialConstantInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:56'!
printOn: stream

	stream nextPutAll: 'localReturn '.
	constant printOn: stream.
! !


!LocalReturnTopInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/6/2002 15:43'!
bytecodesOn: stream

	stream localReturnTop: self spBefore! !

!LocalReturnTopInstr methodsFor: 'as yet unclassified' stamp: 'ajh 8/7/2002 11:51'!
printOn: stream

	stream nextPutAll: 'localReturnTop'.
! !


!LookupKey methodsFor: 'accessing' stamp: 'ajh 9/12/2002 12:04'!
canAssign

	^ true! !

!LookupKey methodsFor: 'accessing' stamp: 'ajh 9/12/2002 02:27'!
poolName

	^ self key! !

!LookupKey methodsFor: 'accessing' stamp: 'ajh 7/21/2002 13:15'!
poolValue

	^ self value! !

!LookupKey methodsFor: 'accessing' stamp: 'ajh 7/21/2002 13:16'!
poolValue: obj

	self value: obj.
	^ obj! !


!Association methodsFor: 'accessing' stamp: 'ajh 9/19/2002 23:58'!
poolValue

	^ value! !


!MessageTally methodsFor: 'initialize-release' stamp: 'ajh 2/9/2003 20:21'!
spyEvery: millisecs on: aBlock 
	"Create a spy and spy on the given block at the specified rate."

	| myDelay value startTime time0 |
	(aBlock isMemberOf: BlockContext)
		ifTrue: [self class: aBlock receiver class method: aBlock method]
		ifFalse: [self class: aBlock class method: aBlock method].
	"set up the probe"
	ObservedProcess _ Processor activeProcess.
	myDelay := Delay forMilliseconds: millisecs.
	time0 := Time millisecondClockValue.
	gcStats _ Smalltalk getVMParameters.
	Timer :=
		[[true] whileTrue: 
			[startTime := Time millisecondClockValue.
			myDelay wait.
			self tally: ObservedProcess suspendedContext
				"tally can be > 1 if ran a long primitive"
				by: (Time millisecondClockValue - startTime) // millisecs].
		nil] newProcess.
	Timer priority: Processor userInterruptPriority.
		"activate the probe and evaluate the block"
	Timer resume.
	value := aBlock value.
	"Collect gc statistics"
	Smalltalk getVMParameters keysAndValuesDo:
		[:idx :gcVal| gcStats at: idx put: (gcVal - gcStats at: idx)].
		"cancel the probe and return the value"
	Timer terminate.
	time := Time millisecondClockValue - time0.
	^value! !


!Metaclass methodsFor: 'compiling' stamp: 'ajh 6/18/2002 14:48'!
holder

	^ nil -> self! !

!Metaclass methodsFor: 'compiling' stamp: 'ajh 9/7/2001 16:24'!
scopeHasPoolVariable: name ifTrue: assocBlock  

	^thisClass scopeHasPoolVariable: name ifTrue: assocBlock! !


!MethodClosure methodsFor: 'as yet unclassified' stamp: 'ajh 2/8/2003 23:01'!
method: compiledMethod

	method _ compiledMethod! !

!MethodClosure methodsFor: 'as yet unclassified' stamp: 'ajh 2/8/2003 23:01'!
receiver: obj

	receiver _ obj! !

!MethodClosure methodsFor: 'as yet unclassified' stamp: 'ajh 2/8/2003 22:53'!
value

	^ receiver withArgs: #() executeMethod: method! !


!MethodContext methodsFor: 'accessing' stamp: 'ajh 1/31/2003 21:53'!
isClosureContext

	^ self isExecutingBlock! !

!MethodContext methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:39'!
temp: index 
	"Refer to the comment in ContextPart|temp:."

	^ self at: index! !

!MethodContext methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:40'!
temp: index put: value 
	"Refer to the comment in ContextPart|temp:put:."

	^ self at: index put: value! !


!OrderedLiterals methodsFor: 'adding' stamp: 'ajh 2/4/2002 00:31'!
addLast: lit
	"If the end is out of optimal range for lit and there is still some slots left for its kind in its optimal range the slide it down to the last slot in optimal range"

	| category lastOptimal |
	(equalitySet includes: lit) ifTrue: [^ lit].
	equalitySet add: lit.
	super addLast: lit.
	category _ self categoryFor: lit.
	category ifNil: [^ lit].
	lastOptimal _ self lastOptimalFor: category.
	lastOptimal >= self size ifTrue: [^ lit].
	(self makeRoomIn: category) ifFalse: [^ lit].
	(self categoriesAfter: category) do: [:cat |
		"Slide down one since they will all slide up one after this loop"
		self slideThoseIn: cat].
	self slide: {self size} down: self size - lastOptimal.
	^ lit
! !

!OrderedLiterals methodsFor: 'private' stamp: 'ajh 2/2/2002 18:35'!
categoriesAfter: catIndex

	^ catIndex + 1 to: self categorySizes size! !

!OrderedLiterals methodsFor: 'private' stamp: 'ajh 2/2/2002 16:49'!
categoryForConstant: lit

	^ nil! !

!OrderedLiterals methodsFor: 'private' stamp: 'ajh 2/3/2002 15:14'!
categoryForSelector: selector

	^ selector numArgs caseOf: {
		[0] -> [3 "#(send1Range send0Range)"].
		[1] -> [3 "#(send1Range send0Range)"].
		[2] -> [2 "#send2Range"].
		[3] -> [1 "#send3Range"]
	  } otherwise: [nil]! !

!OrderedLiterals methodsFor: 'private' stamp: 'ajh 2/4/2002 00:32'!
categoryNames
	"#categorySizes must match these ranges"

	^ #(send3Range send2Range (send1Range send0Range))! !

!OrderedLiterals methodsFor: 'private' stamp: 'ajh 2/2/2002 18:36'!
lastOptimalFor: catIndex

	^ self categorySizes at: catIndex! !

!OrderedLiterals methodsFor: 'private' stamp: 'ajh 2/2/2002 18:58'!
makeRoomIn: catIndex
	"Slide elements in this category and in range down by one so a new element can be slid into the last position.  If the range is full with elements of the category already do nothing and return false."

	| lastOptimal lit |
	lastOptimal _ self lastOptimalFor: catIndex.
	lastOptimal >= self size ifTrue: [^ false].
	lastOptimal 
		to: (catIndex = 1 ifTrue: [1] ifFalse: [(self lastOptimalFor: catIndex-1) + 1])
		by: -1 
		do: [:i |
			lit _ self at: i.
			(self categoryFor: lit) = catIndex ifFalse: [
				self slide: {i} up: lastOptimal - i.
				^ true]
		].
	^ false! !

!OrderedLiterals methodsFor: 'private' stamp: 'ajh 2/2/2002 19:30'!
setCollection: anArray

	super setCollection: anArray.
	equalitySet _ LiteralSet new: anArray size.
! !

!OrderedLiterals methodsFor: 'private' stamp: 'ajh 2/4/2002 00:30'!
slideThoseIn: catIndex
	"Slide elements in this category and in range down by one so they can all slide up when a new element is added before it leaving the elements in range."

	| lastOptimal lit |
	lastOptimal _ self lastOptimalFor: catIndex.
	lastOptimal >= self size ifTrue: [^ self].
	lastOptimal 
		to: (catIndex = 1 ifTrue: [1] ifFalse: [self lastOptimalFor: catIndex-1])
		by: -1 
		do: [:i |
			lit _ self at: i.
			(self categoryFor: lit) = catIndex ifFalse: [
				^ self slide: {i} up: lastOptimal - i]
		].
! !

!OrderedLiterals methodsFor: 'configuration' stamp: 'ajh 10/15/2002 14:42'!
categoryFor: lit
	"Return category index in categorySizes that lit belongs in, or nil if none."

	(lit isMemberOf: Symbol) ifTrue: [
		^ lit numArgs caseOf: {
			[0] -> [2].  "sendLiteral0Range"
			[1] -> [2].  "sendLiteral1Range"
			[2] -> [1]  "sendLiteral2Range"
	  		} otherwise: [nil]
	].
	^ 3! !

!OrderedLiterals methodsFor: 'configuration' stamp: 'ajh 10/15/2002 16:22'!
categorySizes

	^ #(
		2  "sendLiteralRange2"
		12  "sendLiteralRange1 sendLiteralRange0"
		18  "pushLiteralRange"
	  )! !

!OrderedLiterals methodsFor: 'accessing' stamp: 'ajh 9/10/2002 23:28'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock

	start to: self size do:
		[:index | ((self at: index) literalEqual: anElement) ifTrue: [^ index]].
	^ exceptionBlock value! !


!ParseEnv methodsFor: 'initializing' stamp: 'ajh 10/15/2002 04:10'!
parseEnv

	^ self! !

!ParseEnv methodsFor: 'initializing' stamp: 'ajh 10/20/2002 22:48'!
receiverVar: var
	"Subclasses may want to reference this and/or rename it.  Default is do nothing"
! !

!ParseEnv methodsFor: 'parsing' stamp: 'ajh 9/14/2002 17:51'!
classEncoding
	"subclass responsibility, default is nil"

	^ nil! !

!ParseEnv methodsFor: 'parsing' stamp: 'ajh 9/14/2002 17:50'!
lookupVar: name
	"subclass responsibility, default is to look up global vars"

	ProtoObject scopeHas: name asSymbol ifTrue:
		[:assoc | ^ PoolVarNode new poolVariable: assoc].
	^ nil
! !

!ParseEnv methodsFor: 'parsing' stamp: 'ajh 9/14/2002 17:49'!
possibleVarsFor: name continued: listOrNil
	"subclass responsibility, default is to look up global vars"

	name first isUppercase ifFalse: [^ listOrNil].
	^ ProtoObject possibleVariablesFor: name continuedFrom: listOrNil! !

!ParseEnv methodsFor: 'decompiling' stamp: 'ajh 2/9/2003 12:20'!
instVarNameAt: index

	^ nil! !

!ParseEnv methodsFor: 'old protocol' stamp: 'ajh 9/22/2002 23:31'!
requestor: editor
	"Old protocol, do nothing"
! !


!ClassEnv methodsFor: 'initializing' stamp: 'ajh 9/14/2002 17:53'!
class: aBehavior

	class _ aBehavior! !

!ClassEnv methodsFor: 'initializing' stamp: 'ajh 9/19/2002 02:29'!
printOn: stream

	class printOn: stream.
	stream nextPutAll: ' parseScope'.
! !

!ClassEnv methodsFor: 'initializing' stamp: 'ajh 10/20/2002 12:53'!
receiverVar: var

	receiverVar _ var.
	receiverVar name: 'self'! !

!ClassEnv methodsFor: 'parsing' stamp: 'ajh 9/14/2002 15:39'!
classEncoding

	^ class! !

!ClassEnv methodsFor: 'parsing' stamp: 'ajh 9/14/2002 17:53'!
lookupVar: name
	"Find my inst or pool var with name.  Return nil if none found."

	name = receiverVar name ifTrue: [^ receiverVar].  "self"
	^ (self lookupInstVar: name) ifNil: [self lookupPoolVar: name]! !

!ClassEnv methodsFor: 'parsing' stamp: 'ajh 9/14/2002 15:56'!
possibleVarsFor: name continued: listOrNil
	"Return my inst and pool var names that are close to name"

	| list |
	list _ self possibleInstVarsFor: name continued: listOrNil.
	^ self possiblePoolVarsFor: name continued: list! !

!ClassEnv methodsFor: 'decompiling' stamp: 'ajh 2/9/2003 12:21'!
instVarNameAt: index

	^ class allInstVarNames at: index! !

!ClassEnv methodsFor: 'private' stamp: 'ajh 9/14/2002 17:53'!
lookupInstVar: name
	"Return a VarNode for my instance var at name.  Return nil if none found"

	| index |
	index _ class allInstVarNames indexOf: name.
	index = 0 ifTrue: [^ nil].
	^ receiverVar instVarNode: name at: index! !

!ClassEnv methodsFor: 'private' stamp: 'ajh 9/14/2002 17:54'!
lookupPoolVar: name
	"Return a VarNode for my pool var with this name.  Return nil if none found"

	class scopeHas: name asSymbol ifTrue:
		[:assoc | ^ PoolVarNode new poolVariable: assoc].
	^ nil
! !

!ClassEnv methodsFor: 'private' stamp: 'ajh 9/14/2002 17:54'!
possibleInstVarsFor: name continued: listOrNil

	^ name
		correctAgainst: class allInstVarNames, {receiverVar name}
		continuedFrom: listOrNil
! !

!ClassEnv methodsFor: 'private' stamp: 'ajh 9/14/2002 17:54'!
possiblePoolVarsFor: name continued: listOrNil

	name first isUppercase ifFalse: [^ listOrNil].
	^ class possibleVariablesFor: name continuedFrom: listOrNil! !


!DoItClosureEnv methodsFor: 'parsing' stamp: 'ajh 10/20/2002 12:59'!
classEncoding

	^ classEnv classEncoding! !

!DoItClosureEnv methodsFor: 'parsing' stamp: 'ajh 10/20/2002 12:59'!
lookupVar: name
	"Return a VarNode that will reference the named closure var from receiverVar, which represent self"

	| index |
	name = receiverVar name ifTrue: [^ receiverVar].  "thisClosure"
	index _ closure closureVarNames indexOf: name.
	index = 0 ifTrue: [^ classEnv lookupVar: name].
	^ self closureVarNode: name at: index! !

!DoItClosureEnv methodsFor: 'parsing' stamp: 'ajh 10/20/2002 12:59'!
possibleVarsFor: name continued: listOrNil
	"Return my var names that are close in spelling to name"

	| list |
	list _ name
		correctAgainst: closure closureVarNames, {receiverVar name}
		continuedFrom: listOrNil.
	^ classEnv possibleVarsFor: name continued: list! !

!DoItClosureEnv methodsFor: 'private' stamp: 'ajh 9/14/2002 16:25'!
closureVarNode: name at: index

	^ MessageVarNode new
		name: name;
		getMessageNode: (MessageNode2 new
			receiver: receiverVar referenceNode
			selector: #closureAt:
			arguments: {ConstantNode new object: index});
		setMessageNode: (MessageNode2 new
			receiver: receiverVar referenceNode
			selector: #closureAt:put:
			arguments: {ConstantNode new object: index. TempVarNode dummy})! !

!DoItClosureEnv methodsFor: 'initializing' stamp: 'ajh 9/14/2002 16:23'!
closure: aBlockClosure

	closure _ aBlockClosure! !

!DoItClosureEnv methodsFor: 'initializing' stamp: 'ajh 10/20/2002 13:00'!
receiverVar: var

	| i |
	receiverVar _ var.
	receiverVar name: 'thisClosure'.
	i _ closure closureVarNames indexOf: 'self'.
	classEnv _ i = 0
		ifTrue: [nil parseEnv]
		ifFalse: [(closure at: i) class parseEnv
					receiverVar: (self closureVarNode: 'self' at: i)]! !


!DoItContextEnv methodsFor: 'parsing' stamp: 'ajh 10/20/2002 13:02'!
classEncoding

	^ receiverEnv classEncoding! !

!DoItContextEnv methodsFor: 'parsing' stamp: 'ajh 10/20/2002 13:02'!
lookupVar: name
	"Return a VarNode that will reference the named closure var from receiverVar, which represent self"

	| index |
	name = receiverVar name ifTrue: [^ receiverVar].  "thisContext"
	index _ context tempNames indexOf: name.
	index = 0 ifTrue: [^ receiverEnv lookupVar: name].
	^ self tempVarNode: name at: index! !

!DoItContextEnv methodsFor: 'parsing' stamp: 'ajh 10/20/2002 13:02'!
possibleVarsFor: name continued: listOrNil
	"Return my var names that are close in spelling to name"

	| list |
	list _ name
		correctAgainst: context tempNames, {receiverVar name}
		continuedFrom: listOrNil.
	^ receiverEnv possibleVarsFor: name continued: list! !

!DoItContextEnv methodsFor: 'private' stamp: 'ajh 9/14/2002 16:51'!
tempVarNode: name at: index

	^ MessageVarNode new
		name: name;
		getMessageNode: (MessageNode2 new
			receiver: receiverVar referenceNode
			selector: #tempAt:
			arguments: {ConstantNode new object: index});
		setMessageNode: (MessageNode2 new
			receiver: receiverVar referenceNode
			selector: #tempAt:put:
			arguments: {ConstantNode new object: index. TempVarNode dummy})! !

!DoItContextEnv methodsFor: 'initializing' stamp: 'ajh 9/14/2002 16:34'!
context: aMethodContext

	context _ aMethodContext! !

!DoItContextEnv methodsFor: 'initializing' stamp: 'ajh 1/31/2003 22:08'!
receiverVar: var

	| rcvrVar |
	receiverVar _ var.
	receiverVar name: 'thisContext'.
	rcvrVar _ MessageVarNode new
		getMessageNode: (MessageNode2 new
			receiver: receiverVar referenceNode
			selector: #receiver
			arguments: #()).
	receiverEnv _ context isClosureContext
		ifTrue: [context receiver parseEnv receiverVar: rcvrVar]
		ifFalse: [context receiver class parseEnv receiverVar: rcvrVar].
! !


!ParseNode2 methodsFor: 'initializing' stamp: 'ajh 2/9/2003 13:44'!
comment: string

	comment _ string! !

!ParseNode2 methodsFor: 'enumerating' stamp: 'ajh 8/6/2002 14:01'!
allParseNodesDo: block
	"Evaluate block against all children ParseNodes then self"

	self childrenParseNodesDo: [:node | node allParseNodesDo: block].
	block value: self.
! !

!ParseNode2 methodsFor: 'enumerating' stamp: 'ajh 8/9/2002 12:53'!
allParseNodesDo: doBlock doChildrenIf: testBlock
	"Evaluate doBlock against self then apply again to children if self satisfies testBlock"

	doBlock value: self.
	(testBlock value: self) ifTrue: [
		self childrenParseNodesDo: [:node | 
			node allParseNodesDo: doBlock doChildrenIf: testBlock].
	].
! !

!ParseNode2 methodsFor: 'enumerating' stamp: 'ajh 8/2/2002 13:07'!
childrenParseNodes

	| list |
	list _ OrderedCollection new.
	self childrenParseNodesDo: [:node | list add: node].
	^ list! !

!ParseNode2 methodsFor: 'enumerating' stamp: 'ajh 6/12/2002 17:05'!
childrenParseNodesDo: block
	"Evaluate block against my immediate children ParseNodes.  Do not include VarNodes"
	"self subclassResponsibility.  Default is no children"
! !

!ParseNode2 methodsFor: 'enumerating' stamp: 'ajh 9/25/2002 15:00'!
containsNode: parseNode

	self allParseNodesDo: [:node |
		node == parseNode ifTrue: [^ true].
	].
	^ false! !

!ParseNode2 methodsFor: 'enumerating' stamp: 'ajh 8/16/2002 22:15'!
findANode: nodeClass

	self allParseNodesDo: [:node |
		(node isMemberOf: nodeClass) ifTrue: [^ node].
	].
	^ nil! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 6/10/2002 15:03'!
canAssign

	^ false! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 6/18/2002 13:06'!
canCascade

	^ false! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 6/4/2002 23:13'!
isArrayNode

	^ false! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 6/4/2002 23:03'!
isBlockNode

	^ false! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 6/12/2002 11:40'!
isConstantNode: value

	^ false! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 9/14/2002 11:14'!
isDummyTemp

	^ false! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 6/18/2002 14:18'!
isLeafNode

	^ false! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 7/28/2002 22:51'!
isMessage: selSymbol receiver: rcvrPred arguments: argsPred
	"See comment in MessageNode"

	^ false! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 6/4/2002 23:14'!
isMessageNode

	^ false! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 6/10/2002 13:10'!
isReturn

	^ false! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 6/10/2002 15:12'!
isSelfPseudoVariable

	^ false! !

!ParseNode2 methodsFor: 'testing' stamp: 'ajh 6/18/2002 12:10'!
isSuperNode

	^ false! !

!ParseNode2 methodsFor: 'var usage' stamp: 'ajh 10/1/2002 20:13'!
calcVarUsage: usage
	"traverse my tree and note var usage in usage.
	 self subclassResponsibilty.  Default is no var usage"! !

!ParseNode2 methodsFor: 'code generation' stamp: 'ajh 7/16/2002 10:28'!
emit: method forValue: bool
	"Add instruction to method that will produce the value or effect that self represents.  bool indicates value or effect."

	self subclassResponsibility! !

!ParseNode2 methodsFor: 'code generation' stamp: 'ajh 7/16/2002 10:11'!
emitEffect: method
	"Add instruction to method that will produce the effect that self represents"

	self emit: method forValue: false! !

!ParseNode2 methodsFor: 'code generation' stamp: 'ajh 7/16/2002 10:12'!
emitValue: method
	"Add instruction to method that will produce the value that self represents"

	self emit: method forValue: true! !

!ParseNode2 methodsFor: 'evaluating' stamp: 'ajh 6/4/2002 16:34'!
evaluate
	"Evaluate my expression looking up receiver and tempVar values in env.  (env at: #returnBlock) is a block that will return it arg out to the top.  See MethodNode evaluateAgainst:args:"

	^ self subclassResponsibility! !

!ParseNode2 methodsFor: 'error handling' stamp: 'ajh 6/9/2002 21:01'!
notify: messageString

	ParseError signal: messageString! !

!ParseNode2 methodsFor: 'converting' stamp: 'ajh 6/29/2002 16:10'!
referenceNode

	^ self! !

!ParseNode2 methodsFor: 'printing' stamp: 'ajh 2/9/2003 13:46'!
comment

	^ comment! !

!ParseNode2 methodsFor: 'printing' stamp: 'ajh 2/9/2003 13:51'!
nextWordFrom: aStream setCharacter: aBlock
	| outStream char |
	outStream _ WriteStream on: (String new: 16).
	[(aStream peekFor: Character space) 
		or: [aStream peekFor: Character tab]] whileTrue.
	[aStream atEnd
		or:
			[char _ aStream next.
			char = Character cr or: [char = Character space]]]
		whileFalse: [outStream nextPut: char].
	aBlock value: char.
	^ outStream contents! !

!ParseNode2 methodsFor: 'printing' stamp: 'ajh 2/9/2003 13:53'!
printCommentOn: aStream indent: indent

	| thisComment |
	comment == nil ifTrue: [^ self].
	aStream withStyleFor: #comment
		do: [1 to: comment size do: 
				[:index | 
				index > 1 ifTrue: [aStream crtab: indent].
				aStream nextPut: $".
				thisComment _ comment at: index.
				self printSingleComment: thisComment
					on: aStream
					indent: indent.
				aStream nextPut: $"]].
! !

!ParseNode2 methodsFor: 'printing' stamp: 'ajh 6/10/2002 15:45'!
printOn: aStream 
	"Refer to the comment in Object|printOn:."

	aStream nextPutAll: '{'.
	aStream nextPutAll: ((DialectStream dialect: #ST80
								contents: [:strm | self printOn: strm indent: 0])
							asString).
	aStream nextPutAll: '}'! !

!ParseNode2 methodsFor: 'printing' stamp: 'ajh 6/10/2002 15:46'!
printOn: aStream indent: anInteger 
	"If control gets here, avoid recursion loop."

	super printOn: aStream! !

!ParseNode2 methodsFor: 'printing' stamp: 'ajh 6/10/2002 15:59'!
printOn: aStream indent: level precedence: p

	self printOn: aStream indent: level! !

!ParseNode2 methodsFor: 'printing' stamp: 'ajh 2/9/2003 13:46'!
printSingleComment: aString on: aStream indent: indent 
	"Print the comment string, assuming it has been indented indent tabs.
	Break the string at word breaks, given the widths in the default
	font, at 450 points."

	| readStream word position lineBreak font wordWidth tabWidth spaceWidth lastChar |
	readStream _ ReadStream on: aString.
	font _ TextStyle default defaultFont.
	tabWidth _ TextConstants at: #DefaultTab.
	spaceWidth _ font widthOf: Character space.
	position _ indent * tabWidth.
	lineBreak _ 450.
	[readStream atEnd]
		whileFalse: 
			[word _ self nextWordFrom: readStream setCharacter: [:lc | lastChar _ lc].
			wordWidth _ word inject: 0 into: [:width :char | width + (font widthOf: char)].
			position _ position + wordWidth.
			position > lineBreak
				ifTrue: 
					[aStream crtab: indent.
					position _ indent * tabWidth + wordWidth + spaceWidth.
					lastChar = Character cr
						ifTrue: [[readStream peekFor: Character tab] whileTrue].
					word isEmpty ifFalse: [aStream nextPutAll: word; space]]
				ifFalse: 
					[aStream nextPutAll: word.
					readStream atEnd
						ifFalse: 
							[position _ position + spaceWidth.
							aStream space].
					lastChar = Character cr
						ifTrue: 
							[aStream crtab: indent.
							position _ indent * tabWidth.
							[readStream peekFor: Character tab] whileTrue]]]! !

!ParseNode2 methodsFor: 'decompiling' stamp: 'ajh 10/1/2002 13:01'!
bytecodeRange: interval

	bytecodeRange _ interval! !

!ParseNode2 methodsFor: 'decompiling' stamp: 'ajh 9/9/2002 15:43'!
removeJumpBacksTo: destinationIp

	^ false! !

!ParseNode2 methodsFor: 'decompiling' stamp: 'ajh 8/2/2002 20:44'!
simplifyIfPossible! !

!ParseNode2 methodsFor: 'decompiling' stamp: 'ajh 8/2/2002 00:55'!
toDoIncrement: ignored
	"Only meant for Messages or Assignments - else return nil"

	^ nil! !

!ParseNode2 methodsFor: 'bytecode map' stamp: 'ajh 9/28/2002 01:39'!
bytecodeRange

	bytecodeRange class == Interval ifTrue: [^ bytecodeRange].
	bytecodeRange ifNil: [^ nil].

	"bytecodeRange is an array of StackInstructions (see #mapInstr:).  Convert it to their first to last bytecode index"
	bytecodeRange _ bytecodeRange first bytecodeRange first
			to: bytecodeRange last bytecodeRange last.
	^ bytecodeRange! !

!ParseNode2 methodsFor: 'bytecode map' stamp: 'ajh 9/30/2002 21:24'!
bytecodeRangeFirst

	^ (self bytecodeRange ifNil: [^ nil]) first! !

!ParseNode2 methodsFor: 'bytecode map' stamp: 'ajh 10/1/2002 16:09'!
firstBytecodeIndex
	"Return the first bytecode index of me and all my children nodes"

	| first |
	first _ SmallInteger maxVal.
	self allParseNodesDo: [:node |
		first _ first min: (node bytecodeRangeFirst ifNil: [first])
	].
	first = SmallInteger maxVal ifTrue: [^ nil].
	^ first! !

!ParseNode2 methodsFor: 'bytecode map' stamp: 'ajh 9/28/2002 00:52'!
initInstrMap

	bytecodeRange _ nil! !

!ParseNode2 methodsFor: 'bytecode map' stamp: 'ajh 9/28/2002 00:49'!
mapInstr: instr
	"Remember instruction generated.  Converted to bytecode range later"

	bytecodeRange _ bytecodeRange
		ifNil: [{instr}]
		ifNotNil: [bytecodeRange copyWith: instr]
! !

!ParseNode2 methodsFor: 'bytecode map' stamp: 'ajh 9/30/2002 21:26'!
nodeAtBytecode: ip

	self allParseNodesDo: [:n |
		n bytecodeRangeFirst = ip ifTrue: [^ n]].
	^ self errorNoNodeAtBytecode: ip! !

!ParseNode2 methodsFor: 'bytecode map' stamp: 'ajh 2/8/2003 18:27'!
oldBlockNodeAtBytecode: ip

	self allParseNodesDo: [:n |
		n isMethod ifFalse: [
			n oldStartPc = ip ifTrue: [^ n]]].
	^ self errorNoBlockAtBytecode: ip! !

!ParseNode2 methodsFor: 'bytecode map' stamp: 'ajh 10/4/2002 13:48'!
oldStartPc
	"bytecodeRange defined in Decompiler checkForBlock:"

	^ (self bytecodeRange ifNil: [^ nil]) last + 1! !

!ParseNode2 methodsFor: 'source code' stamp: 'ajh 1/31/2003 19:38'!
endSourceRange: sourcePosition

	sourceRange last = 0 ifFalse: [^ self].
	sourceRange _ sourceRange first to: sourcePosition.
! !

!ParseNode2 methodsFor: 'source code' stamp: 'ajh 8/29/2002 23:47'!
sourceRange

	^ sourceRange! !

!ParseNode2 methodsFor: 'source code' stamp: 'ajh 8/22/2002 15:52'!
sourceRange: interval

	sourceRange _ interval! !

!ParseNode2 methodsFor: 'source code' stamp: 'ajh 1/31/2003 19:38'!
startSourceRange: sourcePosition

	sourceRange ifNotNil: [^ self].
	sourceRange _ sourcePosition + 1 to: 0.
! !

!ParseNode2 methodsFor: 'tiles' stamp: 'ajh 2/9/2003 13:45'!
addCommentToMorph: aMorph

	| row |
	(comment isNil or: [comment isEmpty]) ifTrue: [^ self].
	row _ aMorph addTextRow:
		(String streamContents: [:strm | self printCommentOn: strm indent: 1]).
	row firstSubmorph color: (SyntaxMorph translateColor: #comment).
	row parseNode: (self as: CommentNode).
! !


!ArrayNode methodsFor: 'initializing' stamp: 'ajh 6/18/2002 14:46'!
elements: nodes

	elementNodes _ nodes! !

!ArrayNode methodsFor: 'accessing' stamp: 'ajh 6/12/2002 17:05'!
childrenParseNodesDo: block

	elementNodes do: block! !

!ArrayNode methodsFor: 'accessing' stamp: 'ajh 6/4/2002 23:14'!
elementNodes

	^ elementNodes! !

!ArrayNode methodsFor: 'testing' stamp: 'ajh 6/4/2002 23:13'!
isArrayNode

	^ true! !

!ArrayNode methodsFor: 'var usage' stamp: 'ajh 10/1/2002 20:11'!
calcVarUsage: usage

	elementNodes do: [:node | node calcVarUsage: usage]! !

!ArrayNode methodsFor: 'code generation' stamp: 'ajh 9/28/2002 00:31'!
emit: method forValue: bool

	elementNodes size <= 4 ifTrue: [
		"Short form: Array braceWith: a with: b ..."
		(PoolVarNode new poolVariable: Array holder) referenceNode emitValue: method.
		elementNodes do: [:node | node emitValue: method].
		method mapTo: self.
		method send: self selectorForShortForm.
		method unMap.
	] ifFalse: [
		"Long form: (Array braceStream: N) nextPut: a; nextPut: b; ...; braceArray"
		(PoolVarNode new poolVariable: Array holder) referenceNode emitValue: method.
		(ConstantNode new object: elementNodes size) emitValue: method.
		method send: #braceStream:.
		elementNodes do: [:node |
			method doDup.
			node emitValue: method.
			method send: #nextPut:.
			method doPop.
		].
		method mapTo: self.
		method send: #braceArray.
		method unMap.
	].
	bool ifFalse: [method doPop].
! !

!ArrayNode methodsFor: 'code generation' stamp: 'ajh 6/5/2002 06:37'!
selectorForShortForm

	^ #(braceWithNone braceWith: braceWith:with: braceWith:with:with: braceWith:with:with:with:) at: elementNodes size + 1! !

!ArrayNode methodsFor: 'evaluating' stamp: 'ajh 6/4/2002 12:58'!
evaluate

	^ elementNodes collect: [:node | node evaluate]! !

!ArrayNode methodsFor: 'printing' stamp: 'ajh 1/31/2003 19:35'!
printOn: aStream indent: level

	self startSourceRange: aStream position.
	aStream nextPut: ${.
	1 to: elementNodes size do: 
		[:i | (elementNodes at: i) printOn: aStream indent: level.
		i < elementNodes size ifTrue: [aStream nextPutAll: '. ']].
	aStream nextPut: $}.
	self endSourceRange: aStream position.! !

!ArrayNode methodsFor: 'decompiling' stamp: 'ajh 8/2/2002 00:53'!
matchBraceStreamReceiver: receiver messages: messages

	((receiver isMessage: #braceStream: receiver: nil arguments: [:arg | arg isConstantNode: [:o | o isNumber]])
		and: [messages last isMessage: #braceArray receiver: nil arguments: nil])
		ifFalse: [^ nil "no match"].

	"Appears to be a long form brace construct"
	self elements: (messages allButLast collect:
		[:msg | (msg isMessage: #nextPut: receiver: nil arguments: nil)
					ifFalse: [^ nil "not a brace element"].
		msg arguments first]).
! !

!ArrayNode methodsFor: 'decompiling' stamp: 'ajh 7/29/2002 01:56'!
matchBraceWithReceiver: receiver selector: selector arguments: arguments

	selector = (self selectorForShortForm: arguments size)
		ifFalse: [^ nil "no match"].

	"Appears to be a short form brace construct"
	self elements: arguments! !

!ArrayNode methodsFor: 'decompiling' stamp: 'ajh 7/29/2002 01:56'!
selectorForShortForm: nElements

	nElements > 4 ifTrue: [^ nil].
	^ #(braceWithNone braceWith: braceWith:with:
			braceWith:with:with: braceWith:with:with:with:) at: nElements + 1! !


!BlockNode2 methodsFor: 'initializing' stamp: 'ajh 6/10/2002 11:53'!
addTempVarNamed: name asArg: bool

	^ (bool ifTrue: [argTempVars] ifFalse: [otherTempVars])
		add: (TempVarNode new name: name; owningBlockNode: self)! !

!BlockNode2 methodsFor: 'initializing' stamp: 'ajh 2/8/2003 17:51'!
initialize

	receiverTempVar _ TempVarNode new name: 'thisContext receiver'; owningBlockNode: self.
	thisContextTempVar _ TempVarNode new name: 'thisContext'; owningBlockNode: self.
	argTempVars _ OrderedCollection new.
	otherTempVars _ OrderedCollection new.
	statementNodes _ OrderedCollection new.
	primitive _ PrimitiveNode null.
! !

!BlockNode2 methodsFor: 'initializing' stamp: 'ajh 2/8/2003 17:52'!
primitive: aPrimitiveNode

	primitive _ aPrimitiveNode! !

!BlockNode2 methodsFor: 'initializing' stamp: 'ajh 10/4/2002 12:56'!
returnLast

	self returns ifFalse: [
		statementNodes
			at: statementNodes size
			put: (self returnNode: statementNodes last)
	]! !

!BlockNode2 methodsFor: 'initializing' stamp: 'ajh 10/4/2002 12:56'!
returnNode: parseNode

	^ ReturnNode2 new valueNode: parseNode fromContext: thisContextTempVar! !

!BlockNode2 methodsFor: 'initializing' stamp: 'ajh 9/12/2002 03:16'!
returnSelfIfNoOther

	self returns ifFalse: [
		statementNodes last isSelfPseudoVariable ifFalse: [statementNodes add: receiverTempVar referenceNode].
		self returnLast.
	]! !

!BlockNode2 methodsFor: 'initializing' stamp: 'ajh 2/8/2003 17:52'!
selector: aSelector
	"If not nil then this is a method, otherwise this is a block"

	selector _ aSelector! !

!BlockNode2 methodsFor: 'initializing' stamp: 'ajh 1/31/2003 13:17'!
sourceText: stringOrText

	sourceText _ stringOrText! !

!BlockNode2 methodsFor: 'initializing' stamp: 'ajh 8/2/2002 13:17'!
statements: parseNodes

	statementNodes _ parseNodes! !

!BlockNode2 methodsFor: 'accessing' stamp: 'ajh 6/12/2002 17:06'!
childrenParseNodesDo: block

	statementNodes do: block! !

!BlockNode2 methodsFor: 'accessing' stamp: 'ajh 8/23/2002 01:20'!
lastStatement

	^ statementNodes isEmpty
		ifTrue: [Parser2 nilNode]
		ifFalse: [statementNodes last]! !

!BlockNode2 methodsFor: 'accessing' stamp: 'ajh 2/8/2003 17:53'!
primitive

	^ primitive! !

!BlockNode2 methodsFor: 'accessing' stamp: 'ajh 2/8/2003 17:54'!
selector
	"If nil self is a block, otherwise a method"

	^ selector! !

!BlockNode2 methodsFor: 'accessing' stamp: 'ajh 1/31/2003 19:58'!
sourceText

	| text |
	text _ sourceText ifNil: [
		Text streamContents: [:s | self printOn: s indent: 0]].
	text _ text asText.
	sourceRange ifNotNil: [
		text addAttribute: TextColor gray from: 1 to: sourceRange first - 1.
		text addAttribute: TextColor gray from: sourceRange last + 1 to: text size.
	].
	^ text
! !

!BlockNode2 methodsFor: 'accessing' stamp: 'ajh 6/26/2002 16:10'!
statements

	^ statementNodes! !

!BlockNode2 methodsFor: 'accessing' stamp: 'ajh 6/4/2002 12:55'!
statementsButLastDo: block

	1 to: statementNodes size - 1 do: [:i |
		block value: (statementNodes at: i)]! !

!BlockNode2 methodsFor: 'vars' stamp: 'ajh 6/6/2002 12:12'!
argTempVars

	^ argTempVars! !

!BlockNode2 methodsFor: 'vars' stamp: 'ajh 10/17/2002 17:56'!
indexOfClosureVar: freeVar

	^ (self freeVars identityIndexOf: freeVar ifAbsent: [self error: 'free var not found'])
	  + BlockClosure varOffset! !

!BlockNode2 methodsFor: 'vars' stamp: 'ajh 6/17/2002 12:31'!
isRarg: tempVar

	receiverTempVar == tempVar ifTrue: [^ true].
	^ argTempVars identityIncludes: tempVar! !

!BlockNode2 methodsFor: 'vars' stamp: 'ajh 6/17/2002 16:42'!
numArgs

	^ argTempVars size! !

!BlockNode2 methodsFor: 'vars' stamp: 'ajh 6/28/2002 14:05'!
otherTempVars

	^ otherTempVars! !

!BlockNode2 methodsFor: 'vars' stamp: 'ajh 6/17/2002 12:40'!
rargs
	"Receiver and args temp vars"

	^ {receiverTempVar}, argTempVars! !

!BlockNode2 methodsFor: 'vars' stamp: 'ajh 6/7/2002 10:53'!
receiverTempVar

	^ receiverTempVar! !

!BlockNode2 methodsFor: 'vars' stamp: 'ajh 2/3/2003 22:28'!
tempVarAt: index

	^ index <= argTempVars size
		ifTrue: [argTempVars at: index]
		ifFalse: [otherTempVars at: index - argTempVars size]! !

!BlockNode2 methodsFor: 'vars' stamp: 'ajh 9/13/2002 14:39'!
tempVars

	^ (OrderedCollection new: 20)
		addAll: argTempVars;
		addAll: otherTempVars;
		yourself! !

!BlockNode2 methodsFor: 'vars' stamp: 'ajh 8/9/2002 00:19'!
thisContextTempVar

	^ thisContextTempVar! !

!BlockNode2 methodsFor: 'vars' stamp: 'ajh 10/2/2002 00:36'!
unusedTemps

	^ otherTempVars select: [:v | v varUse isUnused]! !

!BlockNode2 methodsFor: 'testing' stamp: 'ajh 6/18/2002 13:58'!
containsAssignmentTo: var

	self allParseNodesDo: [:node | 
		((node isKindOf: VarAssignmentNode) and: [node var == var]) ifTrue: [^ true]].
	^ false! !

!BlockNode2 methodsFor: 'testing' stamp: 'ajh 6/4/2002 23:03'!
isBlockNode

	^ true! !

!BlockNode2 methodsFor: 'testing' stamp: 'ajh 2/8/2003 18:23'!
isMethod

	^ selector notNil! !

!BlockNode2 methodsFor: 'testing' stamp: 'ajh 6/10/2002 13:07'!
returns

	statementNodes isEmpty ifTrue: [^ false].
	^ statementNodes last isReturn! !

!BlockNode2 methodsFor: 'var usage' stamp: 'ajh 2/1/2003 00:39'!
calcCapturedVarUsage

	^ [self calcVarUsage freeVarUsage asCapturedVarUsage]
		on: BlockNodeSignal
		do: [:ex |
			self == ex blockNode
				ifTrue: [ex resume: true]
				ifFalse: [ex pass]
		]! !

!BlockNode2 methodsFor: 'var usage' stamp: 'ajh 10/3/2002 12:30'!
calcVarUsage

	varUsage _ VarUsageSet new blockNode: self.
	statementNodes do: [:node |
		varUsage currentBlockStatement: node.
		node calcVarUsage: varUsage.
	].
	varUsage localFinished.
	^ varUsage! !

!BlockNode2 methodsFor: 'var usage' stamp: 'ajh 2/1/2003 00:38'!
calcVarUsage: usage

	[usage addClosureUsage: self calcVarUsage freeVarUsage]
		on: BlockNodeSignal
		do: [:ex | ex resume: self == ex blockNode]! !

!BlockNode2 methodsFor: 'var usage' stamp: 'ajh 2/9/2003 20:31'!
clearVarUsage

	varUsage _ nil! !

!BlockNode2 methodsFor: 'var usage' stamp: 'ajh 10/4/2002 00:08'!
freeVars
	"Outside temp vars used by the code, ordered by first use"

	^ self varUsage freeVars! !

!BlockNode2 methodsFor: 'var usage' stamp: 'ajh 2/1/2003 00:38'!
varUsage

	^ varUsage ifNil: [
		[self calcVarUsage]
			on: BlockNodeSignal
			do: [:ex | ex resume: self == ex blockNode]]! !

!BlockNode2 methodsFor: 'code generation' stamp: 'ajh 2/9/2003 20:30'!
emit: methodBuilder forValue: bool

	| blockMethod closureVars |
	bool ifFalse: [^ self].  "no effect"

	compiledMethod _ nil.
	blockMethod _ self compiledMethod.
	methodBuilder mapTo: self.
	closureVars _ self freeVars.
	closureVars size = 0 ifTrue: [
		"create block at compile time"
		methodBuilder pushLiteral: blockMethod createBlock.
	] ifFalse: [
		"create block at run time"
		methodBuilder pushLiteral: blockMethod.
		closureVars do: [:var |
			var emitDirectValue: methodBuilder reference: self].
		methodBuilder createBlock: closureVars size.
	].
	methodBuilder unMap.
! !

!BlockNode2 methodsFor: 'code generation' stamp: 'ajh 10/2/2002 12:56'!
emitStatements: method forValue: bool

	^ self emitStatements: method forValue: bool inlined: true! !

!BlockNode2 methodsFor: 'code generation' stamp: 'ajh 1/22/2003 15:49'!
emitStatements: method forValue: bool inlined: isInlined
	"Emit statements along with temp allocations/deallocations"

	statementNodes isEmpty ifTrue: [
		(ConstantNode new object: nil) emit: method forValue: bool.
		(isInlined not and: [bool]) ifTrue: [method localReturnTop].
		^ self
	].

	"Make method recognize thisContext temp vars"
	method thisContext: thisContextTempVar.

	"Wrap args in temp holders if indirect"
	isInlined ifFalse: [
		argTempVars do: [:var |
			var isIndirect ifTrue: [
				method pushTemp: var; send: #wrapInTempHolder; storeTemp: var; doPop]
	]].

	"Allocate temps, wrap any that are indirect"
	method addTemps: otherTempVars.
	otherTempVars do: [:var |
		var isIndirect ifTrue: [
			method pushLiteral: nil; send: #wrapInTempHolder; storeTemp: var; doPop]
	].

	"Emit statement instructions"
	statementNodes allButLastDo: [:node | node emitEffect: method].
	statementNodes last emit: method forValue: bool.

	"Return if none exists and not inlined"
	isInlined ifFalse: [
		bool ifTrue: [self returns ifFalse: [method localReturnTop]].
	].
! !

!BlockNode2 methodsFor: 'code generation' stamp: 'ajh 7/16/2002 10:19'!
emitStatementsEffect: method

	self emitStatements: method forValue: false! !

!BlockNode2 methodsFor: 'code generation' stamp: 'ajh 7/16/2002 10:19'!
emitStatementsValue: method

	self emitStatements: method forValue: true! !

!BlockNode2 methodsFor: 'code generation' stamp: 'ajh 2/9/2003 22:29'!
generate: trailerBytes

	self clearVarUsage.
	compiledMethod _ nil.
	^ self compiledMethodWith: trailerBytes! !

!BlockNode2 methodsFor: 'code generation' stamp: 'ajh 2/9/2003 22:40'!
generateIR

	compiledMethod _ nil.
	self varUsage.
	ir _ InstructionBuilder new
		primitive: self primitive;
		rargs: (argTempVars copyWithFirst: receiverTempVar);
		yourself.
	self emitStatements: ir forValue: self isMethod not inlined: false.
	^ ir
! !

!BlockNode2 methodsFor: 'code generation' stamp: 'ajh 2/9/2003 22:40'!
ir

	^ ir ifNil: [self generateIR]! !

!BlockNode2 methodsFor: 'evaluating' stamp: 'ajh 9/1/2002 13:25'!
evaluate
	"Return a block that will execute my statements upon evaluation.  Make sure the correct env is in place when executing statements"

	self toDo.

"	| env |
	env _ thisContext env.
	^ argTempVars size caseOf: {
	[0] -> [[[self evaluateStatements] doIn: env]].
	[1] -> [[:a1 | 
		[[self evaluateStatements] doIn: (Env new
			at: argTempVars first put: a1).
		] doIn: env]].
	[2] -> [[:a1 :a2 | 
		[[self evaluateStatements] doIn: (Env new
			at: argTempVars first put: a1;
			at: argTempVars second put: a2).
		] doIn: env]].
	[2] -> [[:a1 :a2 :a3 | 
		[[self evaluateStatements] doIn: (Env new
			at: argTempVars first put: a1;
			at: argTempVars second put: a2;
			at: argTempVars third put: a3).
		] doIn: env]].
	[2] -> [[:a1 :a2 :a3 :a4 | 
		[[self evaluateStatements] doIn: (Env new
			at: argTempVars first put: a1;
			at: argTempVars second put: a2;
			at: argTempVars third put: a3;
			at: argTempVars fourth put: a4).
		] doIn: env]].
	}
	otherwise: [self halt: 'need to finish cases above']
"! !

!BlockNode2 methodsFor: 'evaluating' stamp: 'ajh 6/4/2002 14:57'!
evaluateStatements

	self statementsButLastDo: [:node | node evaluate].
	^ self lastStatement evaluate! !

!BlockNode2 methodsFor: 'printing' stamp: 'ajh 2/9/2003 17:18'!
asAltSyntaxText 
	"Answer a string description of the parse tree whose root is the receiver, using the alternative syntax"

	^ DialectStream
		dialect: #SQ00
		contents: [:strm | self printOn: strm indent: 0]! !

!BlockNode2 methodsFor: 'printing' stamp: 'ajh 2/9/2003 17:16'!
asColorizedSmalltalk80Text
	"Answer a colorized Smalltalk-80-syntax string description of the parse tree whose root is the receiver."

	^ DialectStream
		dialect: #ST80
		contents: [:strm | self printOn: strm indent: 0]! !

!BlockNode2 methodsFor: 'printing' stamp: 'ajh 9/10/2002 00:20'!
decompileString 
	"Answer a string description of the parse tree whose root is the receiver."

	^ (DialectStream dialect: #ST80 contents: [:strm | self printOn: strm indent: 0])
		asString
! !

!BlockNode2 methodsFor: 'printing' stamp: 'ajh 2/9/2003 17:18'!
decompileText 
	"Answer a string description of the parse tree whose root is the receiver."

	^ DialectStream
		dialect: (Preferences printAlternateSyntax ifTrue: [#SQ00] ifFalse: [#ST80])
		contents: [:strm | self printOn: strm indent: 0]! !

!BlockNode2 methodsFor: 'printing' stamp: 'ajh 7/19/2002 20:18'!
printArgumentsOn: aStream indent: level
	argTempVars size = 0
		ifTrue: [^ self].
	aStream dialect = #SQ00
		ifTrue: [aStream
				withStyleFor: #setOrReturn
				do: [aStream nextPutAll: 'With'].
			argTempVars
				do: [:arg | 
					aStream space.
					aStream
						withStyleFor: #blockArgument
						do: [arg printOn: aStream]].
			aStream nextPutAll: '. ']
		ifFalse: [argTempVars
				do: [:arg | aStream
						withStyleFor: #blockArgument
						do: [aStream nextPutAll: ':'.
							arg printOn: aStream.
							aStream space]].
			aStream nextPutAll: '| '].
	"If >0 args and >1 statement, put all statements on separate lines"
	statementNodes size > 1
		ifTrue: [aStream crtab: level]! !

!BlockNode2 methodsFor: 'printing' stamp: 'ajh 2/9/2003 00:39'!
printOn: aStream indent: level

	self startSourceRange: aStream position.
	self isMethod ifFalse: [  "is block"
		aStream nextPut: $[.
		self printArgumentsOn: aStream indent: level.
		self printTemporariesOn: aStream indent: level.
		self printStatementsOn: aStream indent: level.
		aStream nextPut: $].
		self endSourceRange: aStream position.
		^ self
	].

	"is method"
	selector precedence = 1
		ifTrue: 
			[aStream nextPutAll: selector]
		ifFalse: 
			[selector keywords with: argTempVars do: 
				[:kwd :arg | 
				aStream dialect = #SQ00
					ifTrue: [(kwd endsWith: ':')
							ifTrue: [aStream withStyleFor: #methodSelector
									do: [aStream nextPutAll: kwd allButLast].
									aStream nextPutAll: ' (']
							ifFalse: [aStream withStyleFor: #methodSelector
									do: [aStream nextPutAll: kwd].
									aStream space]]
					ifFalse: [aStream nextPutAll: kwd; space].
				aStream withStyleFor: #methodArgument
					do: [arg printOn: aStream].
				(aStream dialect = #SQ00 and: [kwd endsWith: ':'])
					ifTrue: [aStream nextPutAll: ') ']
					ifFalse: [aStream space]]].
	self comment == nil ifFalse: 
			[aStream crtab: 1.
			self printCommentOn: aStream indent: 1].
	otherTempVars size > 0 ifTrue: 
			[aStream crtab: 1.
			aStream dialect = #SQ00
				ifTrue: [aStream withStyleFor: #setOrReturn do: [aStream nextPutAll: 'Use']]
				ifFalse: [aStream nextPutAll: '|'].
			aStream withStyleFor: #temporaryVariable
				do: [otherTempVars do: 
						[:temp | aStream space.  temp printOn: aStream]].
			aStream dialect = #SQ00
				ifTrue: [aStream nextPutAll: '.']
				ifFalse: [aStream nextPutAll: ' |']].
	primitive num > 0 ifTrue:
			[(primitive num between: 255 and: 519) ifFalse:  " Dont decompile <prim> for, eg, ^ self "
				[aStream crtab: 1.
				self printPrimitiveOn: aStream]].
	aStream crtab: 1.
	self printStatementsOn: aStream indent: 0.
	self endSourceRange: aStream position.
! !

!BlockNode2 methodsFor: 'printing' stamp: 'ajh 2/8/2003 17:57'!
printPrimitiveOn: aStream 
	"Print the primitive on aStream"

	| primIndex primDecl |
	primIndex _ primitive num.
	primIndex = 0
		ifTrue: [^ self].
	primIndex = 120
		ifTrue: ["External call spec"
			^ aStream print: primitive spec].
	aStream nextPutAll: '<primitive: '.
	primIndex = 117
		ifTrue: [primDecl _ primitive spec.
			aStream nextPut: $';
				
				nextPutAll: (primDecl at: 2);
				 nextPut: $'.
			(primDecl at: 1) notNil
				ifTrue: [aStream nextPutAll: ' module:';
						 nextPut: $';
						
						nextPutAll: (primDecl at: 1);
						 nextPut: $']]
		ifFalse: [aStream print: primIndex].
	aStream nextPut: $>.
	Smalltalk at: #Interpreter ifPresent:[:cls |
		aStream nextPutAll: ' "'
				, ((cls classPool at: #PrimitiveTable)
						at: primIndex + 1) , '" '].! !

!BlockNode2 methodsFor: 'printing' stamp: 'ajh 2/9/2003 13:55'!
printStatementsOn: aStream indent: levelOrZero

	| len shown thisStatement level |
	level _ 1 max: levelOrZero.
	"self comment == nil
		ifFalse: 
			[self printCommentOn: aStream indent: level.
			aStream crtab: level]."
	len _ shown _ statementNodes size.
	len = 0 ifTrue: [^ self].
	(levelOrZero = 0 "top level" and: [(statementNodes last isKindOf: ReturnNode2) and: [statementNodes last valueNode isSelfPseudoVariable]])
		ifTrue: [shown _ 1 max: shown - 1]
		ifFalse: [(len = 1 and: [(statementNodes first isConstantNode: [:obj | obj isNil]) and: [argTempVars size = 0]])
					ifTrue: [shown _ shown - 1]].
	1 to: shown do: 
		[:i | 
		thisStatement _ statementNodes at: i.
		thisStatement printOn: aStream indent: level.
		i < shown ifTrue: [aStream nextPut: $.; crtab: level].
		(thisStatement comment ~~ nil and: [thisStatement comment size > 0])
			ifTrue: 
				[i = shown ifTrue: [aStream crtab: level].
				thisStatement printCommentOn: aStream indent: level.
				i < shown ifTrue: [aStream crtab: level]]]! !

!BlockNode2 methodsFor: 'printing' stamp: 'ajh 8/3/2002 00:17'!
printTemporariesOn: aStream indent: level

	(otherTempVars == nil or: [otherTempVars size = 0])
		ifFalse: 
			[aStream nextPut: $|.
			otherTempVars do: 
				[:arg | 
				aStream
					space;
					withStyleFor: #temporaryVariable
						do: [arg printOn: aStream]].
			aStream nextPutAll: ' | '.
			"If >0 args and >1 statement, put all statements on separate lines"
			statementNodes size > 1 ifTrue: [aStream crtab: level]]! !

!BlockNode2 methodsFor: 'decompiling' stamp: 'ajh 7/29/2002 12:31'!
arguments: vars

	argTempVars _ vars.
	argTempVars do: [:var | var owningBlockNode: self].
! !

!BlockNode2 methodsFor: 'decompiling' stamp: 'ajh 9/10/2002 16:54'!
newArgVar
	"Add new arg temp"

	^ argTempVars add: (TempVarNode new
		name: 'arg', (argTempVars size + 1) printString;
		owningBlockNode: self)
! !

!BlockNode2 methodsFor: 'decompiling' stamp: 'ajh 9/10/2002 16:54'!
newTempVar
	"Add new other temp"

	^ otherTempVars add: (TempVarNode new
		name: 'temp', (otherTempVars size + 1) printString;
		owningBlockNode: self)
! !

!BlockNode2 methodsFor: 'decompiling' stamp: 'ajh 2/9/2003 16:08'!
removeExplicitLocalReturn
	"Since self is a block, remove explicit local return at the end, local return last statement will be assumed"

	^ self removeExplicitLocalReturnTo: thisContextTempVar! !

!BlockNode2 methodsFor: 'decompiling' stamp: 'ajh 2/9/2003 16:08'!
removeExplicitLocalReturnTo: contextVar
	"Since self is a block, remove explicit local return to contextVar at the end, local return last statement will be assumed.  Return true if explicit return was removed.  Return false if no return was found and therefore no modification performed."

	| s |
	(((s _ statementNodes last) isKindOf: ReturnNode2) and: [s contextVar = contextVar])
		ifTrue: [statementNodes atLast: 1 put: s valueNode]
		ifFalse: [^ false].

	"if an if statement exist that local returns then convert it to an if-else with the statements following it in the else"
	statementNodes size to: 1 by: -1 do: [:i |
		s _ statementNodes at: i.
		(s isMessageNode and: [s isInlineIf and: [s argumentNodes size = 1 and: [s argumentNodes first returns]]]) ifTrue: [
			(s argumentNodes first removeExplicitLocalReturnTo: contextVar)
				ifTrue: [s addElseBlock: (BlockNode2 new statements: (statementNodes popTo: i) asOrderedCollection)].
		].
	].
	^ true! !

!BlockNode2 methodsFor: 'decompiling' stamp: 'ajh 10/20/2002 21:18'!
removeJumpBacksTo: destinationIp
	"Last statement is expected to be a jumpBack, but there me other jumpBacks to the same spot due to goto optimizations done by the MethodBuilder optimizer (#simplify).  If the optimizer see a jump at the end of branch to a jumpBack then it converts the end-of-branch jump directly to the jumpBack.  So here we have to find those optimizations and revert them back to end-of-jump branches (ie. get rid of them).  So, we need to convert blocks that look like:
	[condition ifTrue: [statements1. a DummyJumpBack].
	 statements2.
	 aDummyJumpBack]
to:
	[condition ifTrue: [statements1] ifFalse: [statements2]]"

	| node ifs j |
	node _ statementNodes last.
	((node class == DummyJumpBack) and: [node destinationIp = destinationIp]) ifFalse: [^ false].
	statementNodes removeLast.

	ifs _ OrderedCollection new.
	1 to: statementNodes size do: [:i |
		((statementNodes at: i) removeJumpBacksTo: destinationIp)
			ifTrue: [ifs push: i].
	].
	"collapse statements between here and jumpBack ifs if any"
	[ifs isEmpty] whileFalse: [
		j _ ifs pop.
		(statementNodes at: j) addElseBlock:
			(BlockNode2 new statements: (statementNodes popTo: j)).
	].
	^ true! !

!BlockNode2 methodsFor: 'decompiling' stamp: 'ajh 2/3/2003 18:24'!
startTempNamesAfter: index

	| i |
	i _ index.
	argTempVars do: [:v | v name: 't', (i _ i + 1) printString].
	otherTempVars do: [:v | v name: 't', (i _ i + 1) printString].
! !

!BlockNode2 methodsFor: 'decompiling' stamp: 'ajh 7/29/2002 12:31'!
temporaries: vars

	otherTempVars _ vars.
	otherTempVars do: [:var | var owningBlockNode: self].
! !

!BlockNode2 methodsFor: 'bytecodes' stamp: 'ajh 2/9/2003 21:48'!
compiledMethod
	"Generate compiled method"

	^ compiledMethod ifNil: [self compiledMethodWith: #(0)]! !

!BlockNode2 methodsFor: 'bytecodes' stamp: 'ajh 8/6/2002 12:19'!
compiledMethod: aCompiledMethod
	"The block method generated from this block"

	compiledMethod _ aCompiledMethod! !

!BlockNode2 methodsFor: 'bytecodes' stamp: 'ajh 2/9/2003 20:34'!
compiledMethodWith: trailer
	"Generate compiled method"

	^ (compiledMethod isNil or: [compiledMethod trailer ~= trailer asByteArray])
		ifTrue: [compiledMethod _ self generateIR compiledMethodWith: trailer]
		ifFalse: [compiledMethod]! !

!BlockNode2 methodsFor: 'bytecodes' stamp: 'ajh 9/9/2002 23:52'!
privCompiledMethod
	"The block method generated from this block"

	^ compiledMethod! !

!BlockNode2 methodsFor: 'queries' stamp: 'ajh 1/31/2003 15:16'!
blockNodeFor: method

	self compiledMethod.
	self allParseNodesDo: [:node |
		(node isBlockNode and: [node privCompiledMethod = method])
			ifTrue: [^ node sourceText: self sourceText]
	].
	self errorNodeNotFound! !

!BlockNode2 methodsFor: 'queries' stamp: 'ajh 2/8/2003 17:41'!
closureVarNames

	^ self freeVars collect: [:v | v name]! !

!BlockNode2 methodsFor: 'queries' stamp: 'ajh 1/31/2003 23:11'!
sourceMap
	"Answer with a sorted collection of associations: bytecode index -> source code range.  Do not include embedded block closures."

	| list start |
	"Besides getting start position, make sure bytecodeRanges are filled in"
	start _ self compiledMethod initialPC - 1.  

	list _ OrderedCollection new: 40.
	self allParseNodesDo: [:node |  | b |
		(node ~~ self and: [(b _ node bytecodeRange) notNil])
			ifTrue: [list add: start + b first -> node sourceRange]
	] doChildrenIf: [:node |  "if not block closure"
		node == self or: 
		[node isBlockNode not or: 
		[node privCompiledMethod isNil]]
	].
	list _ list asSortedCollection: [:ass1 :ass2 | ass1 key <= ass2 key].
	"Fix up nil source ranges"
	1 to: list size do: [:i |  | ass lastSourcePos |
		ass _ list at: i.
		ass value ifNil: [
			lastSourcePos _ i = 1 ifTrue: [1] ifFalse: [(list at: i - 1) value last + 1].
			ass value: (lastSourcePos to: lastSourcePos - 1).
		].
	].
	^ list! !

!BlockNode2 methodsFor: 'queries' stamp: 'ajh 10/9/2002 12:08'!
tempNames
	"In declaration order, this order is used by decompiler"

	^ self ir tempNames! !

!BlockNode2 methodsFor: 'C translation' stamp: 'ajh 9/11/2002 02:46'!
allEmbeddedTempNames
	"include all inner block temps with my otherTempVars, and remove duplicate names"

	| temps |
	temps _ Set new.
	self allParseNodesDo: [:node |
		node isBlockNode ifTrue: [
			node == self ifFalse: 
				[node argTempVars do: [:v | temps add: v name]].
			node otherTempVars do: [:v | temps add: v name].
		]
	].
	^ temps asSortedCollection! !

!BlockNode2 methodsFor: 'C translation' stamp: 'ajh 2/9/2003 13:42'!
asTranslationMethodOfClass: aClass
 
	^ aClass new
		setSelector2: selector
		args: (argTempVars collect: [:arg | arg name])
		locals: self allEmbeddedTempNames
		block: (self asTranslatorNode setArguments: #())
		primitive: primitive num;
		comment: comment
! !

!BlockNode2 methodsFor: 'C translation' stamp: 'ajh 2/9/2003 13:41'!
asTranslatorNode

	| statementList newS |
	statementList _ OrderedCollection new.
	statementNodes do: [ :s |
		newS _ s asTranslatorNode.
		newS isStmtList ifTrue: [
			"inline the statement list returned when a CascadeNode is translated"
			statementList addAll: newS statements.
		] ifFalse: [
			statementList add: newS.
		].
	].
	^TStmtListNode new
		setArguments: (argTempVars asArray collect: [:arg | arg name])
		statements: statementList;
		comment: comment! !

!BlockNode2 methodsFor: 'scope' stamp: 'ajh 10/3/2002 18:38'!
lookupVar: name
	"Find my temp var with name.  Return nil if none found."

	argTempVars detect: [:v | v name = name] ifFound: [:v | ^ v].
	otherTempVars detect: [:v | v name = name] ifFound: [:v | ^ v].
	^ nil! !

!BlockNode2 methodsFor: 'scope' stamp: 'ajh 10/3/2002 17:50'!
possibleVarsFor: name continued: listOrNil

	^ name
		correctAgainst: (self tempVars collect: [:v | v name])
		continuedFrom: listOrNil! !

!BlockNode2 methodsFor: 'old' stamp: 'ajh 2/8/2003 18:15'!
classEncoding
	"Old protocol, do nothing"

	^ nil! !

!BlockNode2 methodsFor: 'old' stamp: 'ajh 2/8/2003 18:15'!
encoder
	"Old protocol, do nothing"

	^ self! !

!BlockNode2 methodsFor: 'old' stamp: 'ajh 2/8/2003 18:14'!
requestor: editor
	"Old protocol, do nothing"
! !


!BlockNode2 class methodsFor: 'as yet unclassified' stamp: 'ajh 6/10/2002 08:28'!
new

	^ super new initialize! !

!BlockNode2 class methodsFor: 'decompiling' stamp: 'ajh 2/8/2003 18:26'!
selectorFor: numVars

	^ (String streamContents: [:s |
		s nextPutAll: #createBlock.
		1 to: numVars do: [:i | s nextPutAll: 't:'].
	  ]) asSymbol! !


!CascadeMessageNode methodsFor: 'initializing' stamp: 'ajh 6/29/2002 15:22'!
addMessage: messageNode

	messageNode receiverNode ifNotNil: [self error: 'cascade message must have nil receiver'].
	messageNodes add: messageNode.
! !

!CascadeMessageNode methodsFor: 'initializing' stamp: 'ajh 6/29/2002 15:22'!
initialize

	messageNodes _ OrderedCollection new.
! !

!CascadeMessageNode methodsFor: 'initializing' stamp: 'ajh 6/18/2002 13:40'!
receiver: rNode messages: mNodes

	receiverNode _ rNode.
	messageNodes _ mNodes.
! !

!CascadeMessageNode methodsFor: 'initializing' stamp: 'ajh 6/29/2002 15:21'!
receiverNode: rNode

	receiverNode _ rNode! !

!CascadeMessageNode methodsFor: 'accessing' stamp: 'ajh 6/12/2002 17:07'!
childrenParseNodesDo: block

	block value: receiverNode.
	messageNodes do: block.
! !

!CascadeMessageNode methodsFor: 'accessing' stamp: 'ajh 6/4/2002 13:01'!
lastMessage

	^ messageNodes last! !

!CascadeMessageNode methodsFor: 'accessing' stamp: 'ajh 6/4/2002 13:01'!
messagesButLastDo: block

	1 to: messageNodes size - 1 do: [:i |
		block value: (messageNodes at: i)]! !

!CascadeMessageNode methodsFor: 'var usage' stamp: 'ajh 10/1/2002 20:12'!
calcVarUsage: usage

	receiverNode calcVarUsage: usage.
	messageNodes do: [:node | node calcVarUsage: usage].
! !

!CascadeMessageNode methodsFor: 'code generation' stamp: 'ajh 7/16/2002 13:02'!
emit: method forValue: bool

	receiverNode emitValue: method.
	self messagesButLastDo: [:node |
		method doDup.
		node emitEffect: method.
	].
	self lastMessage emitValue: method.

	bool ifFalse: [method doPop].
! !

!CascadeMessageNode methodsFor: 'evaluating' stamp: 'ajh 6/4/2002 13:00'!
evaluate

	| receiver |
	receiver _ receiverNode evaluate.
	self messagesButLastDo: [:m | m evaluateAgainst: receiver].
	^ self lastMessage evaluateAgainst: receiver! !

!CascadeMessageNode methodsFor: 'printing' stamp: 'ajh 6/10/2002 15:58'!
printOn: aStream indent: level
	self printOn: aStream indent: level precedence: 0! !

!CascadeMessageNode methodsFor: 'printing' stamp: 'ajh 1/31/2003 19:35'!
printOn: aStream indent: level precedence: p

	p > 0 ifTrue: [aStream nextPut: $(].
	messageNodes first printReceiver: receiverNode on: aStream indent: level.
	self startSourceRange: aStream position + 1.
	1 to: messageNodes size do: 
		[:i | (messageNodes at: i) printOn: aStream indent: level.
		i < messageNodes size ifTrue: 
				[aStream nextPut: $;.
				messageNodes first precedence >= 2 ifTrue: [aStream crtab: level + 1]]].
	self endSourceRange: aStream position.
	p > 0 ifTrue: [aStream nextPut: $)].
! !

!CascadeMessageNode methodsFor: 'C translation' stamp: 'ajh 2/9/2003 13:41'!
asTranslatorNode

	^ TStmtListNode new
		setArguments: #()
		statements: (messageNodes collect: [:msg | msg asTranslatorNode receiver: receiverNode asTranslatorNode]);
		comment: comment! !


!CascadeMessageNode class methodsFor: 'as yet unclassified' stamp: 'ajh 6/29/2002 15:22'!
new

	^ super new initialize! !


!ConstantNode methodsFor: 'initializing' stamp: 'ajh 6/10/2002 12:32'!
object: obj

	object _ obj! !

!ConstantNode methodsFor: 'accessing' stamp: 'ajh 6/10/2002 12:33'!
object

	^ object! !

!ConstantNode methodsFor: 'testing' stamp: 'ajh 8/2/2002 00:52'!
isConstantNode: valueTestBlock

	^ valueTestBlock value: object! !

!ConstantNode methodsFor: 'testing' stamp: 'ajh 6/18/2002 14:19'!
isLeafNode

	^ true! !

!ConstantNode methodsFor: 'code generation' stamp: 'ajh 9/28/2002 00:37'!
emit: method forValue: bool

	method mapTo: self.
	bool ifTrue: [method pushLiteral: object].
	method unMap.
! !

!ConstantNode methodsFor: 'evaluating' stamp: 'ajh 6/4/2002 13:04'!
evaluate

	^ object! !

!ConstantNode methodsFor: 'printing' stamp: 'ajh 2/1/2003 13:02'!
printOn: aStream indent: level

	self startSourceRange: aStream position.
	(object isKindOf: LookupKey)
		ifTrue:
			[object key isNil
				ifTrue:
					[aStream nextPutAll: '###';
					 	nextPutAll: object value soleInstance name]
				ifFalse:
					[aStream nextPutAll: '##';
						nextPutAll: object key]]
		ifFalse:
			[aStream withStyleFor: #literal
					do: [object storeOn: aStream]].
	self endSourceRange: aStream position.
! !

!ConstantNode methodsFor: 'comparing' stamp: 'ajh 8/2/2002 18:57'!
= other

	^ other class == self class and: [other object = object]! !

!ConstantNode methodsFor: 'comparing' stamp: 'ajh 7/29/2002 10:32'!
hash

	^ object hash! !

!ConstantNode methodsFor: 'C translation' stamp: 'ajh 8/18/2002 22:04'!
asTranslatorNode

	^ TConstantNode new setValue: object! !


!DialectMethodNode2 methodsFor: 'as yet unclassified' stamp: 'ajh 1/20/2002 18:47'!
setDialect: dialectSymbol

	dialect _ dialectSymbol! !


!DummyJumpBack methodsFor: 'as yet unclassified' stamp: 'ajh 7/18/2002 00:47'!
destinationIp

	^ destinationIp! !

!DummyJumpBack methodsFor: 'as yet unclassified' stamp: 'ajh 7/18/2002 00:46'!
destinationIp: bytecodeIndex

	destinationIp _ bytecodeIndex! !


!DummyJumpForward methodsFor: 'as yet unclassified' stamp: 'ajh 7/20/2002 11:22'!
destinationIp

	^ destinationIp! !

!DummyJumpForward methodsFor: 'as yet unclassified' stamp: 'ajh 7/20/2002 11:22'!
destinationIp: bytecodeIndex

	destinationIp _ bytecodeIndex! !


!MessageNode2 methodsFor: 'initializing' stamp: 'ajh 6/18/2002 13:08'!
cascadeReceiver
	"Nil out rcvr (to indicate cascade) and return what it had been."

	| rcvr |
	rcvr _ receiverNode.
	receiverNode _ nil.
	^ rcvr! !

!MessageNode2 methodsFor: 'initializing' stamp: 'ajh 10/1/2002 23:34'!
receiver: rcvrNode selector: aSelector arguments: argNodes

	receiverNode _ rcvrNode.
	selector _ aSelector.
	argumentNodes _ argNodes.
! !

!MessageNode2 methodsFor: 'initializing' stamp: 'ajh 7/17/2002 21:55'!
receiverNode: parseNode

	receiverNode _ parseNode! !

!MessageNode2 methodsFor: 'accessing' stamp: 'ajh 10/1/2002 16:51'!
argNode

	argumentNodes size = 1 ifFalse: [^ self error: 'expected only one arg'].
	^ argumentNodes first! !

!MessageNode2 methodsFor: 'accessing' stamp: 'ajh 6/4/2002 23:16'!
argumentNodes

	^ argumentNodes! !

!MessageNode2 methodsFor: 'accessing' stamp: 'ajh 6/12/2002 17:07'!
childrenParseNodesDo: block

	receiverNode ifNotNil: [block value: receiverNode].
	argumentNodes do: block.
! !

!MessageNode2 methodsFor: 'accessing' stamp: 'ajh 9/14/2002 10:58'!
postCopy

	argumentNodes _ argumentNodes copy.
! !

!MessageNode2 methodsFor: 'accessing' stamp: 'ajh 6/4/2002 23:16'!
receiverNode

	^ receiverNode! !

!MessageNode2 methodsFor: 'accessing' stamp: 'ajh 6/19/2002 15:48'!
selector

	^ selector! !

!MessageNode2 methodsFor: 'testing' stamp: 'ajh 10/1/2002 23:33'!
canCascade

	self isInlineIf ifTrue: [^ false].
	self isInlineIfNil ifTrue: [^ false].
	self isInlineAndOr ifTrue: [^ false].
	self isInlineWhile ifTrue: [^ false].
	self isInlineToDo ifTrue: [^ false].
	self isInlineCase ifTrue: [^ false].
	^ receiverNode isSuperNode not! !

!MessageNode2 methodsFor: 'testing' stamp: 'ajh 6/14/2002 17:55'!
isInlineAndOr

	selector isAndOrSelector ifFalse: [^ false].
	argumentNodes first isBlockNode ifFalse: [^ false].
	argumentNodes first argTempVars isEmpty ifFalse: [self notify: 'and: (or:) takes zero-arg block'. ^ false].
	^ true! !

!MessageNode2 methodsFor: 'testing' stamp: 'ajh 8/22/2002 21:49'!
isInlineCase

	selector isCaseSelector ifFalse: [^ false].
	argumentNodes size = 2 ifTrue: [  "otherwise block"
		argumentNodes last isBlockNode ifFalse: [^ false]].
	argumentNodes first isArrayNode ifFalse: [^ false].
	argumentNodes first elementNodes do: [:assoc |
		(assoc isMessageNode and: [assoc selector = #->]) ifFalse: [^ false].
		assoc receiverNode isBlockNode ifFalse: [^ false].
		assoc receiverNode argTempVars isEmpty ifFalse: [self notify: 'caseOf: takes zero-arg blocks'. ^ false].
		assoc argumentNodes first isBlockNode ifFalse: [^ false].
		assoc argumentNodes first argTempVars isEmpty ifFalse: [self notify: 'caseOf: takes zero-arg blocks'].
	].
	^ true! !

!MessageNode2 methodsFor: 'testing' stamp: 'ajh 6/14/2002 17:56'!
isInlineIf

	selector isIfSelector ifFalse: [^ false].
	argumentNodes do: [:node | node isBlockNode ifFalse: [^ false]].
	argumentNodes do: [:block |
		block argTempVars isEmpty ifFalse: [
			self notify: 'ifTrue:ifFalse: takes zero-arg blocks'.
			^ false
		]
	].
	^ true! !

!MessageNode2 methodsFor: 'testing' stamp: 'ajh 8/22/2002 21:50'!
isInlineIfNil

	| assertNone assertOneOrNone |
	selector isIfNilSelector ifFalse: [^ false].
	argumentNodes do: [:node | node isBlockNode ifFalse: [^ false]].
	assertNone _ [:block | 
		block argTempVars isEmpty ifFalse: [self notify: 'ifNil: takes zero-arg block'. ^ false]
	].
	assertOneOrNone _ [:block | 
		block argTempVars size > 1 ifTrue: [self notify: 'ifNotNil: takes zero- or one-arg block'. ^ false]
	].
	selector caseOf: {
		[#ifNil:] 		 	-> [assertNone value: argumentNodes first].
		[#ifNil:ifNotNil:] 	-> [assertNone value: argumentNodes first. 
							   assertOneOrNone value: argumentNodes last].
		[#ifNotNil:] 		  	-> [assertOneOrNone value: argumentNodes first].
		[#ifNotNil:ifNil:] 	-> [assertOneOrNone value: argumentNodes first. 
							   assertNone value: argumentNodes last]
	}.
	^ true! !

!MessageNode2 methodsFor: 'testing' stamp: 'ajh 1/17/2003 23:04'!
isInlineToDo

	| block step |
	selector isToDoSelector ifFalse: [^ false].
	block _ argumentNodes last.
	block isBlockNode ifFalse: [^ false].
	block argTempVars size = 1 ifFalse: [
		self notify: 'to:do: block must take one arg'. ^ false].
	(block containsAssignmentTo: block argTempVars first) ifTrue: [^ false].
	argumentNodes size = 3 "to:by:do:" ifTrue: [
		step _ argumentNodes second.
		((step isMemberOf: ConstantNode) and: [step object = 0]) ifTrue: [
			self notify: 'by: step can''t be zero'. ^ false].
	].
	^ true! !

!MessageNode2 methodsFor: 'testing' stamp: 'ajh 6/14/2002 18:03'!
isInlineWhile

	selector isWhileSelector ifFalse: [^ false].
	receiverNode isBlockNode ifFalse: [^ false].
	receiverNode argTempVars isEmpty ifFalse: [self notify: 'while receiver block must have no arguments'. ^ false].
	argumentNodes isEmpty ifFalse: [
		argumentNodes first isBlockNode ifFalse: [^ false].
		argumentNodes first argTempVars isEmpty ifFalse: [self notify: 'while takes a zero-arg block as its argument'. ^ false].
	].
	^ true! !

!MessageNode2 methodsFor: 'testing' stamp: 'ajh 7/28/2002 22:51'!
isMessage: selSymbol receiver: rcvrPred arguments: argsPred
	"Answer whether selector is selSymbol, and the predicates rcvrPred and argsPred evaluate to true with respect to receiver and the list of arguments.  If selSymbol or either predicate is nil, it means 'don't care'.  Note that argsPred takes numArgs arguments.  All block arguments are ParseNodes."

	^(selSymbol isNil or: [selSymbol==selector]) and:
		[(rcvrPred isNil or: [rcvrPred value: receiverNode]) and:
			[(argsPred isNil or: [argsPred valueWithArguments: argumentNodes])]]! !

!MessageNode2 methodsFor: 'testing' stamp: 'ajh 6/4/2002 23:14'!
isMessageNode

	^ true! !

!MessageNode2 methodsFor: 'testing' stamp: 'ajh 10/1/2002 23:26'!
isReturn

	self isInlineIf ifTrue: [^ self isReturnIf].
	self isInlineIfNil ifTrue: [^ self isReturnIfNil].
	self isInlineCase ifTrue: [^ self isReturnCase].
	^ super isReturn! !

!MessageNode2 methodsFor: 'var usage' stamp: 'ajh 10/1/2002 23:02'!
calcVarUsage: usage

	self isInlineIf ifTrue: [^ self calcVarUsageIf: usage].
	self isInlineIfNil ifTrue: [^ self calcVarUsageIfNil: usage].
	self isInlineAndOr ifTrue: [^ self calcVarUsageAndOr: usage].
	self isInlineWhile ifTrue: [^ self calcVarUsageWhile: usage].
	self isInlineToDo ifTrue: [^ self calcVarUsageToDo: usage].
	self isInlineCase ifTrue: [^ self calcVarUsageCase: usage].
	^ self calcVarUsageMessage: usage! !

!MessageNode2 methodsFor: 'var usage' stamp: 'ajh 10/1/2002 23:00'!
calcVarUsageMessage: usage

	receiverNode ifNotNil: [receiverNode calcVarUsage: usage].
	argumentNodes do: [:node | node calcVarUsage: usage].
! !

!MessageNode2 methodsFor: 'code generation' stamp: 'ajh 10/1/2002 23:19'!
emit: method forValue: bool

	self isInlineIf ifTrue: [^ self emitIf: method forValue: bool].
	self isInlineIfNil ifTrue: [^ self emitNil: method forValue: bool].
	self isInlineAndOr ifTrue: [^ self emitAndOr: method forValue: bool].
	self isInlineWhile ifTrue: [^ self emitWhile: method forValue: bool].
	self isInlineToDo ifTrue: [^ self emitToDo: method forValue: bool].
	self isInlineCase ifTrue: [^ self emitCase: method forValue: bool].
	^ self emitMessage: method forValue: bool! !

!MessageNode2 methodsFor: 'code generation' stamp: 'ajh 10/1/2002 23:15'!
emitMessage: method forValue: bool

	receiverNode ifNotNil: [ "nil means Cascade message"
		receiverNode emitValue: method].
	argumentNodes do: [:node | node emitValue: method].
	method mapTo: self.
	(receiverNode notNil and: [receiverNode isSuperNode])
		ifTrue: [method send: selector toSuperOf: receiverNode varClass]
		ifFalse: [method send: selector].
	method unMap.

	bool ifFalse: [method doPop].
! !

!MessageNode2 methodsFor: 'evaluating' stamp: 'ajh 6/4/2002 13:04'!
evaluate

	^ self evaluateAgainst: receiverNode evaluate! !

!MessageNode2 methodsFor: 'evaluating' stamp: 'ajh 6/4/2002 13:03'!
evaluateAgainst: receiver

	^ receiver 
		perform: selector 
		withArguments: (argumentNodes collect: [:arg | arg evaluate])! !

!MessageNode2 methodsFor: 'printing' stamp: 'ajh 6/29/2002 16:04'!
precedence

	^ selector precedence! !

!MessageNode2 methodsFor: 'printing' stamp: 'ajh 6/10/2002 16:02'!
printKeywords: key arguments: args on: aStream indent: level

	^ self printKeywords: key arguments: args on: aStream indent: level prefix: false
! !

!MessageNode2 methodsFor: 'printing' stamp: 'ajh 9/11/2002 14:05'!
printKeywords: key arguments: args on: aStream indent: level prefix: isPrefix
	| keywords indent noColons arg kwd hasBrackets doCrTab |
	args size = 0 ifTrue: [aStream space; nextPutAll: key. ^ self].
	keywords _ key keywords.
	noColons _ aStream dialect = #SQ00 and: [keywords first endsWith: ':'].
	doCrTab _ args size > 2 or:
		[{receiverNode} , args
			inject: false
			into: [:was :thisArg |
				was or: [(thisArg isKindOf: BlockNode2)  "thisArg(receiver) is nil when inside a CascadeNode"
					or: [(thisArg isKindOf: MessageNode2) and: [thisArg precedence >= 3]]]]].
	1 to: (args size min: keywords size) do:
		[:i | arg _ args at: i.  kwd _ keywords at: i.
		doCrTab
			ifTrue: [aStream crtab: level+1. indent _ 1] "newline after big args"
			ifFalse: [aStream space. indent _ 0].
		noColons
			ifTrue: [aStream withStyleFor: (isPrefix ifTrue: [#prefixKeyword] ifFalse: [#keyword])
						do: [aStream nextPutAll: kwd allButLast; space].
					hasBrackets _ arg isBlockNode or: [arg isKindOf: ArrayNode].
					hasBrackets ifFalse: [aStream nextPutAll: '(']]
			ifFalse: [aStream nextPutAll: kwd; space].
		arg printOn: aStream indent: level + 1 + indent
			 	precedence: (selector precedence = 2 ifTrue: [1] ifFalse: [selector precedence]).
		noColons
			ifTrue: [hasBrackets ifFalse: [aStream nextPutAll: ')']]]! !

!MessageNode2 methodsFor: 'printing' stamp: 'ajh 1/31/2003 19:36'!
printOn: aStream indent: level

	| symbol leadingKeyword |
	symbol _ selector.
	symbol first = $:
		ifTrue: [leadingKeyword _ symbol keywords first.
				aStream nextPutAll: leadingKeyword; space.
				self printReceiver: receiverNode on: aStream indent: level.
				self startSourceRange: aStream position + 1.
				self printKeywords: (symbol allButFirst: leadingKeyword size + 1) arguments: argumentNodes
					on: aStream indent: level]
		ifFalse: [(aStream dialect = #SQ00 and: [symbol == #do:])
				ifTrue: ["Add prefix keyword"
						aStream withStyleFor: #prefixKeyword do: [aStream nextPutAll: 'Repeat '].
						self printParenReceiver: receiverNode on: aStream indent: level + 1.
						self startSourceRange: aStream position + 1.
						self printKeywords: symbol arguments: argumentNodes
							on: aStream indent: level prefix: true]
				ifFalse: [self printReceiver: receiverNode on: aStream indent: level.
						self startSourceRange: aStream position + 1.
						self printKeywords: symbol arguments: argumentNodes
							on: aStream indent: level]].
	self endSourceRange: aStream position.
! !

!MessageNode2 methodsFor: 'printing' stamp: 'ajh 6/10/2002 15:59'!
printOn: strm indent: level precedence: outerPrecedence

	| parenthesize |
	parenthesize _ selector precedence > outerPrecedence
		or: [outerPrecedence = 3 and: [selector precedence = 3 "both keywords"]].
	parenthesize
		ifTrue: [strm nextPutAll: '('.
				self printOn: strm indent: level.
				strm nextPutAll: ')']
		ifFalse: [self printOn: strm indent: level]! !

!MessageNode2 methodsFor: 'printing' stamp: 'ajh 9/11/2002 14:05'!
printParenReceiver: rcvr on: aStream indent: level
					
	rcvr isBlockNode ifTrue:
		[^ rcvr printOn: aStream indent: level].
	aStream nextPutAll: '('.
	rcvr printOn: aStream indent: level.
	aStream nextPutAll: ')'
! !

!MessageNode2 methodsFor: 'printing' stamp: 'ajh 6/10/2002 16:04'!
printReceiver: rcvr on: aStream indent: level
					
	rcvr ifNil: [^ self].

	"Force parens around keyword receiver of kwd message"
	(selector precedence = 3 and: [aStream dialect = #SQ00])
		ifTrue: [rcvr printOn: aStream indent: level precedence: selector precedence - 1]
		ifFalse: [rcvr printOn: aStream indent: level precedence: selector precedence]
! !

!MessageNode2 methodsFor: 'decompiling' stamp: 'ajh 8/2/2002 22:00'!
arguments

	^ argumentNodes! !

!MessageNode2 methodsFor: 'decompiling' stamp: 'ajh 8/12/2002 00:28'!
ifNilReceiver

	^ receiverNode! !

!MessageNode2 methodsFor: 'decompiling' stamp: 'ajh 10/1/2002 15:57'!
selector: symbol

	selector _ symbol! !

!MessageNode2 methodsFor: 'decompiling' stamp: 'ajh 8/2/2002 00:53'!
toDoIncrement: variable

	(receiverNode = variable referenceNode and: [selector = #+]) 
		ifFalse: [^ nil].
	(argumentNodes first isConstantNode: [:o | o isNumber])
		ifTrue: [^ argumentNodes first]
		ifFalse: [^ nil]! !

!MessageNode2 methodsFor: 'decompiling' stamp: 'ajh 8/2/2002 01:00'!
toDoLimit: variable

	^ (receiverNode = variable referenceNode and: [selector = #<= or: [selector = #>=]]) 
		ifTrue: [argumentNodes first]
		ifFalse: [nil]! !

!MessageNode2 methodsFor: 'decompiling' stamp: 'ajh 2/3/2003 23:20'!
var
	"Called from #matchReceiver:specialSelector:arguments: collecting captured temps"

	selector = #contextTag ifTrue: [^ receiverNode var].
	^ super var  "doesNotUnderstand"! !

!MessageNode2 methodsFor: 'C translation' stamp: 'ajh 8/18/2002 22:05'!
asTranslatorNode

	^ TSendNode new
		setSelector: selector
		receiver: ((receiverNode == nil)
				ifTrue: [nil]
				ifFalse: [receiverNode asTranslatorNode])
		arguments: (argumentNodes collect: [:arg | arg asTranslatorNode])! !

!MessageNode2 methodsFor: 'inline and/or' stamp: 'ajh 10/3/2002 13:48'!
calcVarUsageAndOr: usage

	receiverNode calcVarUsage: usage.
	usage addUsage: argumentNodes first calcCapturedVarUsage withEmptyBranch.
! !

!MessageNode2 methodsFor: 'inline and/or' stamp: 'ajh 10/1/2002 23:20'!
emitAnd: method forValue: bool

	receiverNode emitValue: method.
	method mapTo: self.
	method jumpAheadTo: #false if: false.
	method unMap.
	argumentNodes first emitStatements: method forValue: bool.
	bool ifTrue: [method jumpAheadTo: #end].
	method jumpAheadTarget: #false.
	bool ifTrue: [
		method pushLiteral: false.
		method jumpAheadTarget: #end.
	].
! !

!MessageNode2 methodsFor: 'inline and/or' stamp: 'ajh 10/1/2002 23:20'!
emitAndOr: method forValue: bool

	selector == #and: 
		ifTrue: [self emitAnd: method forValue: bool]
		ifFalse: [self emitOr: method forValue: bool]! !

!MessageNode2 methodsFor: 'inline and/or' stamp: 'ajh 10/1/2002 23:20'!
emitOr: method forValue: bool

	receiverNode emitValue: method.
	method mapTo: self.
	method jumpAheadTo: #true if: true.
	method unMap.
	argumentNodes first emitStatements: method forValue: bool.
	bool ifTrue: [method jumpAheadTo: #end].
	method jumpAheadTarget: #true.
	bool ifTrue: [
		method pushLiteral: true.
		method jumpAheadTarget: #end.
	].
! !

!MessageNode2 methodsFor: 'inline and/or' stamp: 'ajh 10/1/2002 23:24'!
makeLeftAssociative
	"Convert	a or: [b or: c]		to	(a or: b) or: c
	 or 			a and: [b and: c]	to	(a and: b) and: c"

	| rightBlock rightExpr |
	self isInlineAndOr ifFalse: [^ self].
	rightBlock _ self argNode.
	rightExpr _ rightBlock lastStatement.
	rightExpr isInlineAndOr ifFalse: [^ self].
	rightExpr selector = selector ifFalse: [^ self].

	rightBlock statements removeLast.
	rightBlock statements add: rightExpr receiverNode.
	receiverNode _ MessageNode2 new
		receiver: receiverNode selector: selector arguments: {rightBlock}.
	argumentNodes _ rightExpr argumentNodes.
! !

!MessageNode2 methodsFor: 'inline case' stamp: 'ajh 10/3/2002 14:45'!
calcVarUsageCase: usage

	| cases case mergeUsage |
	receiverNode calcVarUsage: usage.
	cases _ argumentNodes first elementNodes copy.
	case _ cases removeFirst.
	mergeUsage _ case receiverNode calcCapturedVarUsage.
	mergeUsage
		calcVarUsageForCase: case argumentNodes first
		rest: cases
		otherwise: (argumentNodes size = 2
			ifTrue: [argumentNodes last]
			ifFalse: [BlockNode2 new]).
	usage addUsage: mergeUsage.
! !

!MessageNode2 methodsFor: 'inline case' stamp: 'ajh 10/1/2002 23:21'!
emitCase: method forValue: bool

	| cases assocMessageNode |
	cases _ argumentNodes first "arrayNode" elementNodes.
	receiverNode emitValue: method.
	1 to: cases size - 1 do: [:i |
		method doDup.
		assocMessageNode _ cases at: i.
		assocMessageNode receiverNode emitStatementsValue: method.
		method send: #=.
		method jumpAheadTo: #next if: false.
		method doPop.
		assocMessageNode argumentNodes first emitStatements: method forValue: bool.
		method jumpAheadTo: #end.
		method jumpAheadTarget: #next.
	].
	argumentNodes size = 2 ifTrue: [
		"last case with otherwise"
		assocMessageNode _ cases last.
		assocMessageNode receiverNode emitStatementsValue: method.
		method send: #=.
		method jumpAheadTo: #next if: false.
		assocMessageNode argumentNodes first emitStatements: method forValue: bool.
		method jumpAheadTo: #end.
		method jumpAheadTarget: #next.
		argumentNodes last emitStatements: method forValue: bool.
	] ifFalse: [
		"last case without otherwise"
		method doDup.
		assocMessageNode _ cases last.
		assocMessageNode receiverNode emitStatementsValue: method.
		method send: #=.
		method jumpAheadTo: #next if: false.
		method doPop.
		assocMessageNode argumentNodes first emitStatements: method forValue: bool.
		method jumpAheadTo: #end.
		method jumpAheadTarget: #next.
		method send: #caseError.
		self isReturn ifTrue: [
			method returnTop: (self findANode: ReturnNode2) contextVar.
		] ifFalse: [
			bool ifFalse: [method doPop].
		].
	].

	"there has to be a one-to-one correspondence between every jump and target"	
	1 to: cases size do: [:i | method jumpAheadTarget: #end].
! !

!MessageNode2 methodsFor: 'inline case' stamp: 'ajh 10/1/2002 23:27'!
isReturnCase

	^ (argumentNodes first "arrayNode" elementNodes
		allSatisfy: [:assocMessage | assocMessage argumentNodes first "action block" returns])
	   and: [selector == #caseOf: or: [argumentNodes second "otherwise block" returns]]! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/1/2002 23:29'!
addElseBlock: blockNode

	blockNode statements isEmpty ifTrue: [^ self].
	selector _ selector == #ifTrue: ifTrue: [#ifTrue:ifFalse:]
		ifFalse: [selector == #ifFalse: ifTrue: [#ifFalse:ifTrue:]
			ifFalse: [self error: 'can only add else branch to single-branch if']].
	argumentNodes _ argumentNodes copyWith: blockNode.
! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/3/2002 13:48'!
calcVarUsageIf: usage

	| branch1 branch2 |
	receiverNode calcVarUsage: usage.
	branch1 _ argumentNodes first calcCapturedVarUsage.
	branch2 _ argumentNodes size = 2
		ifTrue: [argumentNodes second calcCapturedVarUsage]
		ifFalse: [VarUsageSet new].
	usage addUsage: (branch1 mergeUsage: branch2).
! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/1/2002 23:21'!
emitIf: method forValue: bool

	selector caseOf: {
		[#ifTrue:] 		  -> [self emitIfTrue: method forValue: bool].
		[#ifTrue:ifFalse:] -> [self emitIfTrueIfFalse: method forValue: bool].
		[#ifFalse:] 		  -> [self emitIfFalse: method forValue: bool].
		[#ifFalse:ifTrue:] -> [self emitIfFalseIfTrue: method forValue: bool]
	}
! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/1/2002 23:21'!
emitIfFalse: method forValue: bool

	receiverNode emitValue: method.
	method mapTo: self.
	method jumpAheadTo: #true if: true.
	method unMap.
	argumentNodes first emitStatements: method forValue: bool.
	bool ifTrue: [method jumpAheadTo: #end].
	method jumpAheadTarget: #true.
	bool ifTrue: [
		method pushLiteral: nil.
		method jumpAheadTarget: #end.
	].! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/1/2002 23:21'!
emitIfFalseIfTrue: method forValue: bool

	receiverNode emitValue: method.
	method mapTo: self.
	method jumpAheadTo: #true if: true.
	method unMap.
	argumentNodes first emitStatements: method forValue: bool.
	method jumpAheadTo: #end.
	method jumpAheadTarget: #true.
	argumentNodes second emitStatements: method forValue: bool.
	method jumpAheadTarget: #end.
! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/1/2002 23:21'!
emitIfTrue: method forValue: bool

	receiverNode emitValue: method.
	method mapTo: self.
	method jumpAheadTo: #false if: false.
	method unMap.
	argumentNodes first emitStatements: method forValue: bool.
	bool ifTrue: [method jumpAheadTo: #end].
	method jumpAheadTarget: #false.
	bool ifTrue: [
		method pushLiteral: nil.
		method jumpAheadTarget: #end.
	].! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/1/2002 23:22'!
emitIfTrueIfFalse: method forValue: bool

	receiverNode emitValue: method.
	method mapTo: self.
	method jumpAheadTo: #false if: false.
	method unMap.
	argumentNodes first emitStatements: method forValue: bool.
	method jumpAheadTo: #end.
	method jumpAheadTarget: #false.
	argumentNodes second emitStatements: method forValue: bool.
	method jumpAheadTarget: #end.
! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/1/2002 23:29'!
falseBlock

	^ selector caseOf: {
		[#ifTrue:] -> [nil].
		[#ifFalse:] -> [argumentNodes first].
		[#ifTrue:ifFalse:] -> [argumentNodes second].
		[#ifFalse:ifTrue:] -> [argumentNodes first]}! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/1/2002 23:27'!
isReturnIf
	"is returning if"

	^ (#(ifTrue:ifFalse: ifFalse:ifTrue:) includes: selector)
		and: [argumentNodes first returns and: [argumentNodes last returns]]! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/1/2002 23:31'!
removeJumpBacksTo: destinationIp

	self isInlineIf ifFalse: [^ super removeJumpBacksTo: destinationIp].
	argumentNodes size = 1 ifFalse: [^ false].
	^ argumentNodes first removeJumpBacksTo: destinationIp! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/4/2002 22:09'!
simplifyIfPossible

	self isInlineIf ifFalse: [^ super simplifyIfPossible].
	argumentNodes size = 2 ifFalse: [^ self].
	argumentNodes first statements isEmpty ifTrue: [
		argumentNodes _ argumentNodes allButFirst.
		selector = #ifTrue:ifFalse:
			ifTrue: [selector _ #ifFalse:]
			ifFalse: [selector _ #ifTrue:].  "was #ifFalse:ifTrue:"
	].
	argumentNodes last statements isEmpty ifTrue: [
		argumentNodes _ argumentNodes allButLast.
		selector = #ifTrue:ifFalse:
			ifTrue: [selector _ #ifTrue:]
			ifFalse: [selector _ #ifFalse:].  "was #ifFalse:ifTrue:"
	].
! !

!MessageNode2 methodsFor: 'inline if' stamp: 'ajh 10/1/2002 23:31'!
trueBlock

	^ selector caseOf: {
		[#ifTrue:] -> [argumentNodes first].
		[#ifFalse:] -> [nil].
		[#ifTrue:ifFalse:] -> [argumentNodes first].
		[#ifFalse:ifTrue:] -> [argumentNodes second]}! !

!MessageNode2 methodsFor: 'inline ifNil' stamp: 'ajh 10/1/2002 23:06'!
calcVarUsageIfNil: usage

	^ self calcVarUsageIf: usage! !

!MessageNode2 methodsFor: 'inline ifNil' stamp: 'ajh 10/1/2002 23:22'!
emitIfNil: method forValue: bool

	receiverNode emitValue: method.
	bool ifTrue: [method doDup].
	method mapTo: self.
	method pushLiteral: nil.
	method send: #==.
	method jumpAheadTo: #notNil if: false.
	method unMap.
	bool ifTrue: [method doPop].
	argumentNodes first emitStatements: method forValue: bool.
	method jumpAheadTarget: #notNil.
! !

!MessageNode2 methodsFor: 'inline ifNil' stamp: 'ajh 10/1/2002 23:22'!
emitIfNilIfNotNil: method forValue: bool

	| blockArgs |
	receiverNode emitValue: method.
	(blockArgs _ argumentNodes second argTempVars) isEmpty
		ifFalse: [blockArgs first emitStore: method].
	method mapTo: self.
	method pushLiteral: nil.
	method send: #==.
	method jumpAheadTo: #notNil if: false.
	method unMap.
	argumentNodes first emitStatements: method forValue: bool.
	method jumpAheadTo: #end.
	method jumpAheadTarget: #notNil.
	argumentNodes second emitStatements: method forValue: bool.
	method jumpAheadTarget: #end.
! !

!MessageNode2 methodsFor: 'inline ifNil' stamp: 'ajh 10/1/2002 23:22'!
emitIfNotNil: method forValue: bool

	| blockArgs |
	receiverNode emitValue: method.
	(blockArgs _ argumentNodes first argTempVars) isEmpty
		ifFalse: [blockArgs first emitStore: method].
	method mapTo: self.
	method pushLiteral: nil.
	method send: #==.
	method jumpAheadTo: #nil if: true.
	method unMap.
	argumentNodes first emitStatements: method forValue: bool.
	bool ifTrue: [method jumpAheadTo: #end].
	method jumpAheadTarget: #nil.
	bool ifTrue: [
		method pushLiteral: nil.
		method jumpAheadTarget: #end.
	].
! !

!MessageNode2 methodsFor: 'inline ifNil' stamp: 'ajh 10/1/2002 23:23'!
emitIfNotNilIfNil: method forValue: bool

	| blockArgs |
	receiverNode emitValue: method.
	(blockArgs _ argumentNodes first argTempVars) isEmpty
		ifFalse: [blockArgs first emitStore: method].
	method mapTo: self.
	method pushLiteral: nil.
	method send: #==.
	method jumpAheadTo: #nil if: true.
	method unMap.
	argumentNodes first emitStatements: method forValue: bool.
	method jumpAheadTo: #end.
	method jumpAheadTarget: #nil.
	argumentNodes second emitStatements: method forValue: bool.
	method jumpAheadTarget: #end.
! !

!MessageNode2 methodsFor: 'inline ifNil' stamp: 'ajh 10/1/2002 23:22'!
emitNil: method forValue: bool

	selector caseOf: {
		[#ifNil:] 		  -> [self emitIfNil: method forValue: bool].
		[#ifNil:ifNotNil:] -> [self emitIfNilIfNotNil: method forValue: bool].
		[#ifNotNil:] 		  -> [self emitIfNotNil: method forValue: bool].
		[#ifNotNil:ifNil:] -> [self emitIfNotNilIfNil: method forValue: bool]
	}
! !

!MessageNode2 methodsFor: 'inline ifNil' stamp: 'ajh 10/1/2002 23:27'!
isReturnIfNil
	"is returning if"

	^ (#(ifNil:ifNotNil: ifNotNil:ifNil:) includes: selector)
		and: [argumentNodes first returns and: [argumentNodes last returns]]! !

!MessageNode2 methodsFor: 'inline to-do' stamp: 'ajh 10/3/2002 13:48'!
calcVarUsageToDo: usage

	| doLoop |
	receiverNode calcVarUsage: usage.
	argumentNodes allButLastDo: [:node | node calcVarUsage: usage].
	doLoop _ argumentNodes last calcCapturedVarUsage withEmptyBranch.
	usage
		addUsage: doLoop;
		addUsage: doLoop.  "do twice for loop"
! !

!MessageNode2 methodsFor: 'inline to-do' stamp: 'ajh 1/20/2003 23:49'!
emitToDo: method forValue: bool
	"If step is not a constant we don't know which way to test #<= or #>=, so we generate code before the loop that evals the step expression then see if is positive or negative then store the selector #<= or #>= in an invisible variable which will be pushed before the sendTop test every time.  This adds a push and a primitive message send vs. just a quick bytecode send if it were constant.  I think this is not so bad. The alternative is to create a block, which wouldn't be so bad either unless it had inner blocks which would get created on each iteration.  By the way, the Slang-to-C translator (incorrectly) assumes variable steps are always positive."

	| limit step block iterator testSelector limitEmit stepEmit numDecls |
	limit _ argumentNodes first.
	step _ argumentNodes size = 3 
		ifTrue: [argumentNodes second]			"to:by:do:"
		ifFalse: [ConstantNode new object: 1].	"to:do:"
	block _ argumentNodes last.
	numDecls _ 0.

	"allocate and initialize iterator"
	receiverNode emitValue: method.  "eval receiver before limit and step"
	iterator _ block argTempVars first.
	method addTemp: iterator.
	method storeTemp: iterator; doPop.
	numDecls _ numDecls + 1.

	"evaluate limit only once if it's an expression or a changing variable"
	(limit isLeafNode and:
	 [(limit isMemberOf: VarReferenceNode) not
	  or: [(block containsAssignmentTo: limit var) not]]
	) ifTrue: [
		limitEmit _ [limit emitValue: method].
	] ifFalse: [
		limit emitValue: method.
		limit _ {iterator. #limit}.  "must be unique throughout method"
		method addTemp: limit.
		method storeTemp: limit; doPop.
		numDecls _ numDecls + 1.
		limitEmit _ [method pushTemp: limit].
	].

	"evaluate step expression if not literal and determine correct test selector"
	stepEmit _ [step emitValue: method].
	(step isMemberOf: ConstantNode) ifFalse: [		
		((step isMemberOf: VarReferenceNode) not
		 or: [block containsAssignmentTo: step var]) ifTrue: [
			step emitValue: method.
			step _ {iterator. #step}.  "must be unique throughout method"
			method addTemp: step.
			method storeTemp: step; doPop.
			numDecls _ numDecls + 1.
			stepEmit _ [method pushTemp: step]
		].
		stepEmit value.
		method doDup.
		method pushLiteral: 0.
		method send: #=.
		method jumpAheadTo: #ok if: false.
		method pushLiteral: 'to:by:do: step can''t be zero'.
		method send: #error:.
		method jumpAheadTarget: #ok.
		method pushLiteral: 0.
		method send: #>.
		method jumpAheadTo: #negative if: false.
		method pushLiteral: -1.
		method jumpAheadTo: #store.
		method jumpAheadTarget: #negative.
		method pushLiteral: 1.
		method jumpAheadTarget: #store.
		testSelector _ {iterator. #test}.
		method addTemp: testSelector.
		method storeTemp: testSelector; doPop.
		numDecls _ numDecls + 1.
	].

	iterator emitValue: method.  "initialized to receiver value"
	"loop"
	method jumpBackTarget: #start.
	limitEmit value.
	method mapTo: self.
	testSelector ifNil: [
		method send: (step object > 0 ifTrue: [#<=] ifFalse: [#>=]).
	] ifNotNil: [
		method pushTemp: testSelector.
		method pushLiteral: 0.
		method send: #>.
		method jumpAheadTo: #lessThan if: false.
		method send: #>=.
		method jumpAheadTo: #test.
		method jumpAheadTarget: #lessThan.
		method send: #<=.
		method jumpAheadTarget: #test.
	].
	method jumpAheadTo: #done if: false.
	method unMap.
	block emitStatementsEffect: method.
	iterator emitValue: method.
	stepEmit value.
	method send: #+.
	iterator emitStore: method.
	method jumpBackTo: #start.
	method jumpAheadTarget: #done.

	bool ifTrue: [method pushLiteral: nil].
! !

!MessageNode2 methodsFor: 'inline while' stamp: 'ajh 10/3/2002 13:48'!
calcVarUsageWhile: usage

	| testLoop doLoop |
	testLoop _ receiverNode calcCapturedVarUsage.
	usage addUsage: testLoop.
	argumentNodes isEmpty ifFalse: [
		doLoop _ argumentNodes first calcCapturedVarUsage withEmptyBranch.
		usage addUsage: doLoop.
	].
	"apply again for loop"
	usage addUsage: testLoop.
	doLoop ifNotNil: [usage addUsage: doLoop].
! !

!MessageNode2 methodsFor: 'inline while' stamp: 'ajh 10/1/2002 23:23'!
emitWhile: method forValue: bool

	method jumpBackTarget: #begin.
	receiverNode emitStatementsValue: method.
	method mapTo: self.
	selector caseOf: {
		[#whileTrue:]	-> [method jumpAheadTo: #end if: false].
		[#whileTrue]	-> [method jumpAheadTo: #end if: false].
		[#whileFalse:]	-> [method jumpAheadTo: #end if: true].
		[#whileFalse]	-> [method jumpAheadTo: #end if: true]
	}.
	method unMap.
	argumentNodes notEmpty ifTrue: [
		argumentNodes first emitStatementsEffect: method].
	method jumpBackTo: #begin.
	method jumpAheadTarget: #end.

	bool ifTrue: [method pushLiteral: nil].
! !

!MessageNode2 methodsFor: 'inline while' stamp: 'ajh 10/4/2002 00:01'!
toDoFromWhileWithInit: initStmt
	"Return nil, or a to:do: expression equivalent to this whileTrue:"

	| variable increment limit body test |
	(selector == #whileTrue:
		and: [(initStmt isKindOf: VarAssignmentNode) and:
				[initStmt variable isTemp]])
		ifFalse: [^ nil].
	body _ argumentNodes last.
	body statements isEmpty ifTrue: [^ nil].
	variable _ initStmt variable.
	increment _ body statements last toDoIncrement: variable.
	(increment == nil or: [receiverNode statements size ~= 1])
		ifTrue: [^ nil].
	test _ receiverNode statements first.
	"Note: test could really be checked that <= or >= comparison
	jibes with the sign of the (constant) increment"
	((test isMemberOf: MessageNode2)
		and: [(limit _ test toDoLimit: variable) notNil])
		ifFalse: [^ nil].
	body statements removeLast.
	body arguments: {variable}.
	^ (increment isConstantNode: [:o | o = 1])
		ifTrue: [MessageNode2 new
			receiver: initStmt valueNode
			selector: #to:do:
			arguments: {limit. body}]
		ifFalse: [MessageNode2 new
			receiver: initStmt valueNode
			selector: #to:by:do:
			arguments: {limit. increment. body}]
! !


!ParseStack methodsFor: 'accessing' stamp: 'ajh 6/12/2002 18:28'!
pop

	^ self pop: 1! !

!ParseStack methodsFor: 'accessing' stamp: 'ajh 7/23/2001 13:37'!
pop: n

	(position _ position - n) "< 0 
		ifTrue: [self error: 'Parse stack underflow']"! !

!ParseStack methodsFor: 'accessing' stamp: 'ajh 6/12/2002 18:27'!
push

	^ self push: 1! !

!ParseStack methodsFor: 'combining stacks' stamp: 'ajh 7/24/2001 22:47'!
= other

	^ self class == other class 
	  and: [position = other position
	  and: [length = other size]]! !

!ParseStack methodsFor: 'combining stacks' stamp: 'ajh 12/30/2001 02:04'!
addStack: aParseStack
	"self ends where aParseStack begins.  Update self to include aParseStack"

	length _ length max: position + aParseStack size.
	position _ position + aParseStack position.
	position < 0 ifTrue: [self errorParseStackUnderflow]! !

!ParseStack methodsFor: 'combining stacks' stamp: 'ajh 7/24/2001 22:48'!
hash

	^ position hash bitXor: length hash! !

!ParseStack methodsFor: 'combining stacks' stamp: 'ajh 8/8/2001 19:21'!
mergeStack: aParseStack
	"Both self and arg belong to basic blocks that call the same block so their stack positions must be the same when entering.  Return true if this merge causes self to change length, otherwise return false."

	position = aParseStack position ifFalse: [self error: 'Parse stack out of synch'].
	length < aParseStack size ifTrue: [
		length _ aParseStack size.
		^ true].
	^ false
! !


!ParseStack class methodsFor: 'as yet unclassified' stamp: 'ajh 7/24/2001 22:30'!
new

	^ super new init! !


!PopInstr methodsFor: 'as yet unclassified' stamp: 'ajh 8/7/2002 11:30'!
bytecodesOn: stream

	stream popTop! !

!PopInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/3/2002 20:59'!
printOn: stream

	stream nextPutAll: 'pop'.
! !

!PopInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/5/2002 11:54'!
stackEffect: parseStack

	parseStack pop! !


!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 7/14/2001 12:37'!
num

	^ primitiveNum! !

!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 7/14/2001 12:23'!
num: n

	primitiveNum _ n! !

!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 7/15/2002 14:11'!
printOn: aStream

	self printOn: aStream indent: 0! !

!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 7/24/2001 20:44'!
printOn: aStream indent: level

	aStream nextPutAll: 'primitive '; print: primitiveNum! !

!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 7/14/2001 12:37'!
spec

	^ spec! !

!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 7/14/2001 12:30'!
spec: literal

	spec _ literal! !


!PrimitiveNode class methodsFor: 'as yet unclassified' stamp: 'ajh 7/14/2001 12:47'!
null

	^ self new num: 0! !


!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 12:08'!
canAssign

	^ setterBlock notNil! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:01'!
getter: block

	getterBlock _ block! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:00'!
name: string

	name _ string! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:01'!
poolName

	^ name! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:01'!
poolValue

	^ getterBlock value! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:02'!
poolValue: obj

	setterBlock value: obj! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:01'!
setter: block

	setterBlock _ block! !


!PushConstantInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 17:41'!
addTo: sequence

	super addTo: sequence.
	isSpecial _ Compiler bytecodeWriteStream specialConstants identityIncludes: constant.
	isSpecial ifTrue: [isByte _ false.  ^ self].
	isByte _ constant isInteger and: [(constant between: 0 and: 255) and: [Compiler bytecodeWriteStream canUnderstand: #pushByte:]].
	isByte ifTrue: [^ self].
	sequence method addLiteral: constant.
! !

!PushConstantInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:50'!
bytecodesOn: stream

	isSpecial ifTrue: [^ stream pushConst: constant].
	isByte ifTrue: [^ stream pushByte: constant].
	stream pushLiteral: (owningSequence method literalSlotIndexOf: constant)! !

!PushConstantInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:51'!
constant

	^ constant! !

!PushConstantInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:50'!
constant: obj

	constant _ obj! !

!PushConstantInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:55'!
convertToLocalReturnConstant

	| obj |
	isSpecial ifFalse: [self error: 'only call this on special constants'].
	owningSequence localParseStack pop.
	self primitiveChangeClassTo: LocalReturnSpecialConstantInstr basicNew.
	"Associate self with mapToObject"
	obj _ owningSequence method mapToObject.
	obj ifNotNil: [obj mapInstr: self].
! !

!PushConstantInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:43'!
isPushSpecialConstant

	^ isSpecial! !

!PushConstantInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:37'!
printOn: stream

	stream nextPutAll: 'pushLit '.
	self constant printOn: stream.
! !

!PushConstantInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:37'!
stackEffect: parseStack

	parseStack push! !


!PushInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 21:03'!
bytecodesOn: stream

	stream pushInstVar: instVarIndex! !

!PushInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 21:00'!
printOn: stream

	stream nextPutAll: 'push instVar '.
	instVarIndex printOn: stream.
! !

!PushInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/27/2002 12:57'!
slot

	^ instVarIndex! !

!PushInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/24/2002 14:28'!
slot: index

	instVarIndex _ index! !

!PushInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 20:56'!
stackEffect: parseStack

	parseStack push! !


!PushThisContextInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/15/2002 02:47'!
bytecodesOn: stream

	stream pushThisContext: self spBefore! !

!PushThisContextInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/7/2002 15:26'!
printOn: stream

	stream nextPutAll: 'pushThisContext'! !

!PushThisContextInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/7/2002 15:26'!
stackEffect: parseStack

	parseStack push! !


!ReadOnlyVariableBinding methodsFor: 'accessing' stamp: 'ajh 9/12/2002 12:06'!
canAssign

	^ false! !

!ReadOnlyVariableBinding methodsFor: 'accessing' stamp: 'ajh 9/19/2002 23:58'!
poolValue

	^ value! !


!RemoteReturnTopInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/17/2003 23:13'!
addTo: sequence

	super addTo: sequence.
	isSpecial _ Compiler bytecodeWriteStream specialSelectors includesKey: #remoteReturnTo:.
	isSpecial ifFalse: [sequence method addLiteral: #remoteReturnTo:].
! !

!RemoteReturnTopInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/17/2003 23:15'!
bytecodesOn: stream

	isSpecial
		ifTrue: [stream sendSpecial: #remoteReturnTo:]
		ifFalse: [stream
			sendLiteral: (owningSequence method literalSlotIndexOf: #remoteReturnTo:)
			numArgs: 1]
! !

!RemoteReturnTopInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/22/2002 16:09'!
printOn: stream

	stream nextPutAll: 'remoteReturn'.
! !

!RemoteReturnTopInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/17/2003 23:16'!
stackEffect: parseStack
	"home contextTag is on top with return value underneath"

	parseStack pop: 1! !

!RemoteReturnTopInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/13/2002 00:13'!
successorSequences

	^ #()! !

!RemoteReturnTopInstr methodsFor: 'testing' stamp: 'ajh 6/13/2002 00:03'!
isJumpOrReturn

	^ true! !

!RemoteReturnTopInstr methodsFor: 'testing' stamp: 'ajh 6/13/2002 10:28'!
isReturnInstr

	^ true! !


!ReturnNode2 methodsFor: 'initializing' stamp: 'ajh 10/3/2002 18:33'!
valueNode: aParseNode fromContext: contextBlockVar

	valueNode _ aParseNode.
	contextVar _ contextBlockVar.
! !

!ReturnNode2 methodsFor: 'accessing' stamp: 'ajh 6/12/2002 17:09'!
childrenParseNodesDo: block

	block value: valueNode! !

!ReturnNode2 methodsFor: 'accessing' stamp: 'ajh 9/25/2002 12:48'!
contextVar

	^ contextVar! !

!ReturnNode2 methodsFor: 'accessing' stamp: 'ajh 6/29/2002 16:14'!
valueNode

	^ valueNode! !

!ReturnNode2 methodsFor: 'testing' stamp: 'ajh 6/10/2002 13:10'!
isReturn

	^ true! !

!ReturnNode2 methodsFor: 'var usage' stamp: 'ajh 2/1/2003 00:48'!
calcVarUsage: usage

	valueNode calcVarUsage: usage.
	(BlockNodeSignal new blockNode: contextVar owningBlockNode; signal)
		ifFalse: [contextVar usage: usage nowHasRef: self].
! !

!ReturnNode2 methodsFor: 'code generation' stamp: 'ajh 10/25/2002 09:49'!
emit: method forValue: bool
	"If local return push value and return.
	 If remote return push value and captured home context tag then remote return"

	valueNode emitValue: method.
	method mapTo: self.
	(method isLocalTemp: contextVar) ifTrue: [
		method localReturnTop.
	] ifFalse: [
		contextVar emitValue: method.
		method remoteReturn.
	].
	method unMap.
! !

!ReturnNode2 methodsFor: 'evaluating' stamp: 'ajh 6/4/2002 16:34'!
evaluate

	(thisContext env at: #returnBlock) value: valueNode evaluate! !

!ReturnNode2 methodsFor: 'printing' stamp: 'ajh 1/31/2003 19:37'!
printOn: aStream indent: level

	self startSourceRange: aStream position.
	aStream dialect = #SQ00
		ifTrue: ["Add prefix keyword"
				aStream withStyleFor: #setOrReturn do: [aStream nextPutAll: 'Answer '].
				valueNode printOn: aStream indent: level]
		ifFalse: [aStream nextPutAll: '^ '.
				valueNode printOn: aStream indent: level].
	valueNode printCommentOn: aStream indent: level.
	self endSourceRange: aStream position.
! !

!ReturnNode2 methodsFor: 'C translation' stamp: 'ajh 2/9/2003 13:41'!
asTranslatorNode

	^ TReturnNode new 
		setExpression: valueNode asTranslatorNode;
		comment: comment! !


!Parser2 methodsFor: 'public access' stamp: 'ajh 9/14/2002 09:22'!
parse: sourceStreamOrString class: behavior

	^ self parse: sourceStreamOrString readStream class: behavior
		noPattern: false notifying: nil ifFail: [self parseError]! !

!Parser2 methodsFor: 'public access' stamp: 'ajh 9/14/2002 09:28'!
parse: sourceStream class: behavior noPattern: noPattern context: context notifying: req ifFail: aBlock
	"Backwards compatibilty"

	context ifNotNil: [self error: 'Use Compiler-evaluate... instead'].
	^ self parse: sourceStream class: behavior noPattern: noPattern notifying: req ifFail: aBlock! !

!Parser2 methodsFor: 'public access' stamp: 'ajh 1/22/2003 17:56'!
parse: sourceStream class: parseScope noPattern: noPattern notifying: req ifFail: aBlock
	"Parse sourceStream into a embedded BlockNode if noPattern is true (no method header) or a MethodNode if noPattern is false.  Parsing is done with respect to parseScope to find non-local variables.  Errors in parsing are reported to req if not nil followed by executing the fail block."

	| repeatNeeded myStream parser s |
	(req notNil and: [RequestAlternateSyntaxSetting signal and: [(sourceStream isKindOf: FileStream) not]])
		ifTrue: [parser _ self as: DialectParser2]
		ifFalse: [parser _ self].
	myStream _ sourceStream.
	[	repeatNeeded _ false.
		s _ myStream rest.
		parser init: myStream notifying: req failBlock: [^ aBlock value].
		doitFlag _ noPattern.
		[[parser method: noPattern class: parseScope] 
			on: ParserRemovedUnusedTemps do: [:ex | 
				repeatNeeded _ (requestor isKindOf: TextMorphEditor) not.
				myStream _ ReadStream on: requestor text string.
				ex resume]]
			on: ParseError do: [:ex | self notify: ex description].
		repeatNeeded
	] whileTrue.
	^ self blockNode sourceText: s! !

!Parser2 methodsFor: 'public access' stamp: 'ajh 2/7/2003 22:39'!
parseArgsAndTemps: aString notifying: req 
	"No initialization required.  Parse the argument, aString, notifying req if an error occurs. Otherwise, answer a two-element Array containing Arrays of strings (the argument names and temporary variable names)."

	(req notNil and: [RequestAlternateSyntaxSetting signal]) ifTrue:
		[^ (self as: DialectParser2) parseArgsAndTemps: aString notifying: req].
	aString == nil ifTrue: [^ #()].
	doitFlag _ false.               "Don't really know if a doit or not!!"
	^ self initPattern: aString
		notifying: req
		return: [:blockNode |
			self temporaries.
			blockNode tempVars collect: [:v | v name]
		]! !

!Parser2 methodsFor: 'public access' stamp: 'ajh 6/10/2002 10:58'!
parseMethodComment: aString setPattern: aBlock
	"No initialization required.  Answer the method comment for the argument, aString. Evaluate aBlock with the methodNode containing the selector and arguments"

	self
		initPattern: aString
		notifying: nil
		return: aBlock.
	currentComment==nil
		ifTrue:	[^OrderedCollection new]
		ifFalse:	[^currentComment]! !

!Parser2 methodsFor: 'public access' stamp: 'ajh 6/10/2002 10:59'!
parseSelector: aString 
	"No initialization required.  Answer the message selector for the argument, aString, which should parse successfully up to the temporary declaration or the end of the method header."

	^self
		initPattern: aString
		notifying: nil
		return: [:methodNode | methodNode selector]! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 6/9/2002 19:47'!
argumentName

	hereType == #word
		ifFalse: [^self expected: 'Argument name'].
	^self advance! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 8/17/2002 10:30'!
assignment
	" var '_' expression => AssignmentNode."

	| varNode start |
	start _ self startOfNextToken + requestorOffset.
	varNode _ self variable.
	varNode canAssign ifFalse: [^ self notify: 'Cannot store into' at: prevMark + requestorOffset].
	self advance.
	self expression ifFalse: [^ self expected: 'Expression'].
	parseNode _ VarAssignmentNode new variable: varNode value: parseNode.
	self noteSourceRange: (start to: self endOfLastToken) forNode: parseNode.
	^ true! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 1/31/2003 22:50'!
blockExpression
	"[ ({:var} |) (| {temps} |) (statements) ] => BlockNode."

	| start argsSize |
	self pushNewBlockNode.
	start _ prevMark.

	"Gather parameters."
	argsSize _ 0.
	[self match: #colon] whileTrue: [
		self declareArg: self argumentName.
		argsSize _ argsSize + 1].
	(argsSize > 0 & (hereType ~~ #rightBracket) and: [(self match: #verticalBar) not]) ifTrue: [^ self expected: 'Vertical bar'].

	self temporaryBlockVariables.
	self statementsInnerBlock: true.

	(self match: #rightBracket) ifFalse: [^ self expected: 'Period or right bracket'].

	parseNode _ self blockNode.
	self noteSourceRange: (start to: self endOfLastToken) forNode: parseNode.
	self popBlockNode.
! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 1/31/2003 22:40'!
braceExpression
	" { elements } => BraceNode."

	| elements locations loc more start |
	elements _ OrderedCollection new.
	locations _ OrderedCollection new.
	self advance.
	start _ prevMark.
	more _ hereType ~~ #rightBrace.
	[more]
		whileTrue: 
			[loc _ hereMark + requestorOffset.
			self expression
				ifTrue: 
					[elements addLast: parseNode.
					locations addLast: loc]
				ifFalse:
					[^self expected: 'Variable or expression'].
			(self match: #period)
				ifTrue: [more _ hereType ~~ #rightBrace]
				ifFalse: [more _ false]].
	parseNode _ ArrayNode new elements: elements "sourceLocations: locations".
	(self match: #rightBrace)
		ifFalse: [^self expected: 'Period or right brace'].
	self noteSourceRange: (start to: self endOfLastToken) forNode: parseNode.
	^true! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 6/10/2002 07:34'!
cascade
	" {; message} => CascadeNode."

	| rcvr msgs |
	parseNode canCascade
		ifFalse: [^self expected: 'Cascading not'].
	rcvr _ parseNode cascadeReceiver.
	msgs _ OrderedCollection with: parseNode.
	[self match: #semicolon]
		whileTrue: 
			[parseNode _ rcvr.
			(self messagePart: 3 repeat: false)
				ifFalse: [^self expected: 'Cascade'].
			parseNode canCascade
				ifFalse: [^self expected: '<- No special messages'].
			parseNode cascadeReceiver.
			msgs addLast: parseNode].
	parseNode _ CascadeMessageNode new receiver: rcvr messages: msgs! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 6/9/2002 19:47'!
expression

	(hereType == #word and: [tokenType == #leftArrow])
		ifTrue: [^ self assignment].
	hereType == #leftBrace
		ifTrue: [self braceExpression]
		ifFalse: [self primaryExpression ifFalse: [^ false]].
	(self messagePart: 3 repeat: true)
		ifTrue: [hereType == #semicolon ifTrue: [self cascade]].
	^ true! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 8/23/2002 01:16'!
messagePart: level repeat: repeat

  | start receiver selector args words keywordStart |
  [	receiver _ parseNode.
	(hereType == #keyword and: [level >= 3]) ifTrue: [
		start _ self startOfNextToken.
		selector _ WriteStream on: (String new: 32).
		args _ OrderedCollection new.
		words _ OrderedCollection new.
		[hereType == #keyword] whileTrue: [
			keywordStart _ self startOfNextToken + requestorOffset.
			selector nextPutAll: self advance.
			words addLast: (keywordStart to: self endOfLastToken + requestorOffset).
			self primaryExpression ifFalse: [^ self expected: 'Argument'].
			self messagePart: 2 repeat: true.
			args addLast: parseNode
		].
		selector _ self lookupSelector: selector contents wordIntervals: words exprInterval: (start to: self endOfLastToken).
	] ifFalse: [((hereType == #binary or: [hereType == #verticalBar]) and: [level >= 2]) ifTrue: [
		start _ self startOfNextToken.
		selector _ self advance.
		words _ OrderedCollection with: (start  + requestorOffset to: self endOfLastToken + requestorOffset).
		self primaryExpression ifFalse: [^ self expected: 'Argument'].
		self messagePart: 1 repeat: true.
		args _ Array with: parseNode.
		selector _ self lookupSelector: selector wordIntervals: words exprInterval: (start to: self endOfLastToken).
	] ifFalse: [hereType == #word ifTrue: [
		start _ self startOfNextToken.
		selector _ self advance.
		args _ #().
		words _ OrderedCollection with: (start  + requestorOffset to: self endOfLastToken + requestorOffset).
		selector _ self lookupSelector: selector wordIntervals: words exprInterval: (start to: self endOfLastToken).
	] ifFalse: [
		^ args notNil
	]]].
	parseNode _ MessageNode2 new receiver: receiver selector: selector arguments: args.
	self noteSourceRange: (start to: self endOfLastToken) forNode: parseNode.
	repeat
  ] whileTrue: [].
  ^ true! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 9/13/2002 13:13'!
messagePattern
	" unarySelector | binarySelector arg | keyword arg {keyword arg} =>  
	{selector, arguments, precedence}."

	| selector start words keywordStart |
	hereType == #word ifTrue: [
		start _ self startOfNextToken.
		selector _ self advance.
		words _ OrderedCollection with: (start  + requestorOffset to: self endOfLastToken + requestorOffset).
		self blockNode selector: selector asSymbol.
		^ self
	].
	(hereType == #binary or: [hereType == #verticalBar]) ifTrue: [
		start _ self startOfNextToken.
		selector _ self advance.
		words _ OrderedCollection with: (start  + requestorOffset to: self endOfLastToken + requestorOffset).
		self declareArg: self argumentName.
		self blockNode selector: selector asSymbol.
		^ self
	].
	hereType == #keyword ifTrue: [
		start _ self startOfNextToken.
		selector _ WriteStream on: (String new: 32).
		words _ OrderedCollection new.
		[hereType == #keyword] whileTrue: [
			keywordStart _ self startOfNextToken + requestorOffset.
			selector nextPutAll: self advance.
			words addLast: (keywordStart to: self endOfLastToken + requestorOffset).
			self declareArg: self argumentName.
		].
		self blockNode selector: selector contents asSymbol.
		^ self
	].
	^ self expected: 'Message pattern'! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 2/8/2003 18:09'!
method: doit class:  aParseEnvOrClass
	" pattern [ | temporaries ] block => MethodNode."

	nestedBlockNodes push: self newMethodNode.
	parseEnv _ aParseEnvOrClass parseEnv.
	parseEnv receiverVar: self blockNode receiverTempVar.
	(doitFlag _ doit) ifFalse: [self messagePattern].
	homeBlockNode _ self blockNode.
	self temporaries.
	self blockNode comment: currentComment.
	doit ifFalse: [self blockNode primitive: self primitive].
	self statementsInnerBlock: doit.
	doit ifFalse: [self blockNode returnSelfIfNoOther].
	hereType == #doIt ifFalse: [^ self expected: 'Nothing more'].
	self interactive ifTrue: [self removeUnusedTemps].
! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 8/17/2002 10:29'!
primaryExpression

	hereType == #word ifTrue: [
		self variableRef.
		^ true].
	hereType == #leftBracket ifTrue: [
		self advance.
		self blockExpression.
		^ true].
	hereType == #leftBrace ifTrue: [
		self braceExpression.
		^ true].
	hereType == #leftParenthesis ifTrue: [
		self advance.
		self expression ifFalse: [^ self expected: 'expression'].
		(self match: #rightParenthesis) ifFalse: [^ self expected: 'right parenthesis'].
		^ true].
	(hereType == #string or: [hereType == #number or: [hereType == #literal]]) ifTrue: [
		parseNode _ self literalNode: self advance.
		^ true].
	(here == #- and: [tokenType == #number]) ifTrue: [
		self advance.
		parseNode _ self literalNode: self advance negated.
		^ true].
	^ false! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 10/4/2002 13:00'!
statementsInnerBlock: inner

	| stmts returns start more blockComment |
	stmts _ OrderedCollection new.
	"give initial comment to block, since others trail statements"
	blockComment _ currentComment.
	currentComment _ nil.
	returns _ false.
	more _ hereType ~~ #rightBracket.
	[more] whileTrue: [
		start _ self startOfNextToken.
		(returns _ self matchReturn) ifTrue: [
			self expression ifFalse: [^ self expected: 'Expression to return'].
			self addComment.
			stmts addLast: (parseNode isReturn
				ifTrue: [parseNode]
				ifFalse: [self noteSourceRange: (start to: self endOfLastToken)
					forNode: (homeBlockNode returnNode: parseNode)]).
			self match: #period.
			(hereType == #rightBracket or: [hereType == #doIt])
				ifFalse: [^ self expected: 'End of block'].
		] ifFalse: [
			self expression ifTrue: [
				self addComment.
				stmts addLast: parseNode
			] ifFalse: [
				self addComment.
				stmts size = 0 ifTrue: [
					stmts addLast: (inner ifTrue: [self nilNode] ifFalse: [self selfNode])]
			]
		].
		more _ returns not and: [self match: #period]
	].

	self blockNode statements: stmts; comment: blockComment.
	^ true! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 6/10/2002 11:15'!
temporaries
	" [ '|' (variable)* '|' ]"

	| theActualText |
	(self match: #verticalBar) ifFalse: ["no temps"
		doitFlag ifTrue: [
			tempsMark _ requestor 
				ifNil: [1]
				ifNotNil: [requestor selectionInterval first].
			^ self
		].
		tempsMark _ hereMark	"formerly --> prevMark + prevToken".
		tempsMark > 0 ifTrue:
			[theActualText _ source contents.
			[tempsMark < theActualText size and: [(theActualText at: tempsMark) isSeparator]]
				whileTrue: [tempsMark _ tempsMark + 1]].
			^ self
	].
	[hereType == #word]
		whileTrue: [self declareTemp: self advance].
	(self match: #verticalBar) ifTrue: [tempsMark _ prevMark.  ^ self].
	^ self expected: 'Vertical bar'! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 6/10/2002 11:15'!
temporaryBlockVariables
	"Scan and answer temporary block variables."

	(self match: #verticalBar) ifFalse: [
		"There are't any temporary variables."
		^ self].

	[hereType == #word] whileTrue: [self declareTemp: self advance].
	(self match: #verticalBar) ifTrue: [^ self].
	^self expected: 'Vertical bar'! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 8/17/2002 10:32'!
variable

	| varName varStart varEnd varNode |
	varStart _ self startOfNextToken + requestorOffset.
	varName _ self advance.
	varEnd _ self endOfLastToken + requestorOffset.
	varNode _ self lookupVar: varName ifAbsent: [
		self correctVariable: varName interval: (varStart to: varEnd)].
	^ varNode! !

!Parser2 methodsFor: 'expression types' stamp: 'ajh 10/2/2002 00:58'!
variableRef

	| start end |
	start _ self startOfNextToken + requestorOffset.
	parseNode _ self variable referenceNode.
	end _ self endOfLastToken + requestorOffset.
	self noteSourceRange: (start to: end) forNode: parseNode.
! !

!Parser2 methodsFor: 'scanning' stamp: 'ajh 6/9/2002 19:47'!
advance
	| this |
	prevMark _ hereMark.
	prevEnd _ hereEnd.
	this _ here.
	here _ token.
	hereType _ tokenType.
	hereMark _ mark.
	hereEnd _ source position - (source atEnd ifTrue: [hereChar == 30 asCharacter ifTrue: [0] ifFalse: [1]] ifFalse: [2]).
	self scanToken.
	"Transcript show: 'here: ', here printString, ' mark: ', hereMark printString, ' end: ', hereEnd printString; cr."
	^this! !

!Parser2 methodsFor: 'scanning' stamp: 'ajh 6/9/2002 19:47'!
endOfLastToken

	^ prevEnd ifNil: [mark]! !

!Parser2 methodsFor: 'scanning' stamp: 'ajh 6/9/2002 19:47'!
match: type 
	"Answer with true if next tokens type matches."

	hereType == type
		ifTrue: 
			[self advance.
			^true].
	^false! !

!Parser2 methodsFor: 'scanning' stamp: 'ajh 6/9/2002 19:47'!
matchReturn

	^ self match: #upArrow! !

!Parser2 methodsFor: 'scanning' stamp: 'ajh 6/9/2002 19:47'!
matchToken: thing 
	"Matches the token, not its type."

	here = thing ifTrue: [self advance. ^true].
	^false! !

!Parser2 methodsFor: 'scanning' stamp: 'ajh 6/9/2002 19:47'!
startOfNextToken
	"Return starting position in source of next token."

	hereType == #doIt ifTrue: [^source position + 1].
	^hereMark! !

!Parser2 methodsFor: 'vars' stamp: 'ajh 6/10/2002 11:36'!
declareArg: name

	^ self declareTemp: name asArg: true! !

!Parser2 methodsFor: 'vars' stamp: 'ajh 6/10/2002 11:36'!
declareTemp: name

	^ self declareTemp: name asArg: false! !

!Parser2 methodsFor: 'vars' stamp: 'ajh 8/16/2002 19:47'!
declareTemp: name asArg: bool
	"Add a new temporary variable (could be an argument) to the current scope"

	self interactive ifTrue: [
		self lookupVar: name ifPresent: [:var |
			(var isMemberOf: InstVarNode) ifTrue: [
				^ self notify: 'Can''t shadow instance variable'].
			(var isMemberOf: PoolVarNode) ifTrue: [
				^ self notify: 'Can''t shadow pool variable'].
			(var isMemberOf: TempVarNode) ifFalse: [
				^ self notify: 'Can''t shadow reserved variable'].
			(var owningBlockNode == self blockNode) ifTrue: [
				^ self notify: 'Duplicate declaration'].
			"var must be local to an outer block, it is ok to override it"
	]].
	^ self declareTempUnchecked: name asArg: bool! !

!Parser2 methodsFor: 'vars' stamp: 'ajh 7/29/2002 15:05'!
declareTempUnchecked: name asArg: bool

	^ self blockNode addTempVarNamed: name asArg: bool! !

!Parser2 methodsFor: 'vars' stamp: 'ajh 2/8/2003 18:55'!
lookupVar: name
	"'self' will be found in a nestedBlockNode chain (in MethodNode2)"

	| var |
	nestedBlockNodes topToBottomDo: [:blockNode |
		(var _ blockNode lookupVar: name) ifNotNil: [^ var]].
	(var _ parseEnv lookupVar: name) ifNotNil: [^ var].
	^ name caseOf: {
		['nil'] 	-> [self nilNode].
		['false'] -> [self falseNode].
		['true'] 	-> [self trueNode].
		['thisContext'] -> [self blockNode thisContextTempVar].
		['super'] -> [SuperNode new var: (self lookupVar: 'self'); varClass: self classEncoding]}
		otherwise: [nil]
! !

!Parser2 methodsFor: 'vars' stamp: 'ajh 9/12/2002 00:22'!
lookupVar: string ifAbsent: block

	^ (self lookupVar: string) ifNil: block! !

!Parser2 methodsFor: 'vars' stamp: 'ajh 6/10/2002 12:56'!
lookupVar: string ifPresent: block

	^ block value: (self lookupVar: string ifAbsent: [^ nil])! !

!Parser2 methodsFor: 'vars' stamp: 'ajh 8/17/2002 11:17'!
poolVar: assoc

	^ PoolVarNode new poolVariable: assoc! !

!Parser2 methodsFor: 'vars' stamp: 'ajh 9/12/2002 00:16'!
possibleVariablesFor: string

	| list |
	list _ self possibleVarsFor: string continued: nil.
	^ string correctAgainst: nil continuedFrom: list
! !

!Parser2 methodsFor: 'vars' stamp: 'ajh 2/8/2003 18:20'!
possibleVarsFor: name continued: listOrNil
	"'self' found in nestedBlockNode chain above (in MethodNode2)"

	| list |
	list _ listOrNil.
	nestedBlockNodes topToBottomDo: [:blockNode |
		list _ blockNode possibleVarsFor: name continued: listOrNil].
	list _ parseEnv possibleVarsFor: name continued: list.
	^ name correctAgainst: #('nil' 'false' 'true' 'thisContext' 'super') continuedFrom: list
! !

!Parser2 methodsFor: 'error handling' stamp: 'ajh 6/9/2002 19:47'!
expected: aString 
	"Notify a problem at token 'here'."

	tokenType == #doIt ifTrue: [hereMark _ hereMark + 1].
	hereType == #doIt ifTrue: [hereMark _ hereMark + 1].
	^ self notify: aString , ' expected' at: hereMark + requestorOffset! !

!Parser2 methodsFor: 'error handling' stamp: 'ajh 6/9/2002 19:47'!
fail

	| exitBlock |
	exitBlock _ failBlock.
	failBlock _ nil.
	^exitBlock value! !

!Parser2 methodsFor: 'error handling' stamp: 'ajh 6/9/2002 19:47'!
interactive

	^ (requestor == nil or: [requestor isKindOf: SyntaxError]) not! !

!Parser2 methodsFor: 'error handling' stamp: 'ajh 6/9/2002 19:47'!
notify: aString 
	"Notify problem at token before 'here'."

	^self notify: aString at: prevMark + requestorOffset! !

!Parser2 methodsFor: 'error handling' stamp: 'ajh 9/12/2002 01:10'!
notify: string at: location
	requestor isNil
		ifTrue: [SyntaxError 
					errorInClass: (self classEncoding ifNil: [FakeClassPool])
					withCode: 
						(source contents
							copyReplaceFrom: location
							to: location - 1
							with: string , ' ->')
					doitFlag: doitFlag]
		ifFalse: [requestor
					notify: string , ' ->'
					at: location
					in: source].
	^self fail! !

!Parser2 methodsFor: 'error handling' stamp: 'ajh 6/9/2002 19:47'!
offEnd: aString 
	"Notify a problem beyond 'here' (in lookAhead token). Don't be offEnded!!"

	requestorOffset == nil
		ifTrue: [^ self notify: aString at: mark]
		ifFalse: [^ self notify: aString at: mark + requestorOffset]
! !

!Parser2 methodsFor: 'var correction' stamp: 'ajh 6/9/2002 19:57'!
choose: alternatives variable: proposedVariable interval: spot
	"Correct the proposedVariable to one in alternatives.
	spot is the interval of the variable within the test stream."

	| aStream choice userSelection |
	"If we can't ask the user for correction, make it ambiguous"
	self interactive ifFalse: [^ self ambiguousVar: proposedVariable].

	userSelection _ requestor selectionInterval.
	requestor selectFrom: spot first to: spot last.
	requestor select.

	aStream _ WriteStream on: (String new: 200).
	alternatives do: [:varNode | aStream nextPutAll: varNode fullName; cr].
	aStream nextPutAll: 'cancel'.

	choice _ (PopUpMenu labels: aStream contents
				lines: {alternatives size})
		startUpWithCaption:
(('Ambiguous variable: ', proposedVariable, '
please choose, or cancel:') asText makeBoldFrom: 21 to: 21 + proposedVariable size).

	(choice = 0 or: [choice > alternatives size]) ifTrue: [^ self fail].

	requestor deselect.
	requestor selectInvisiblyFrom: userSelection first to: userSelection last.

	^ alternatives at: choice
! !

!Parser2 methodsFor: 'var correction' stamp: 'ajh 9/12/2002 01:12'!
correctVariable: proposedVariable interval: spot
	"Correct the proposedVariable to a known variable, or declare it as a new
	variable if such action is requested.  We support declaring lowercase
	variables as temps, and uppercase variables as Globals or ClassVars,
	depending on whether the context is nil (class=UndefinedObject).
	Spot is the interval within the test stream of the variable."

	| alternatives aStream choice userSelection temp binding declareSize variableNode |
	"If we can't ask the user for correction, make it undeclared"
	self interactive ifFalse: [^ self undeclaredVariable: proposedVariable].

	temp _ proposedVariable first isLowercase.
	"First check to see if the requestor knows anything about the variable"
	(temp and: [(binding _ requestor bindingOf: proposedVariable) notNil])
		ifTrue: [^ self poolVar: binding].
	userSelection _ requestor selectionInterval.
	requestor selectFrom: spot first to: spot last.
	requestor select.

	alternatives _ self possibleVariablesFor: proposedVariable.

	aStream _ WriteStream on: (String new: 200).
	temp ifTrue: [
		declareSize _ 1.
		aStream nextPutAll: 'declare temp'; cr.
	] ifFalse: [
		declareSize _ 2.
		aStream nextPutAll: 'declare global'; cr.
		aStream nextPutAll: 'declare class'; cr.
		self classEncoding ifNotNil: [
			declareSize _ declareSize + 1.  
			aStream nextPutAll: 'declare class variable'; cr].
	].
	alternatives do: [:name | aStream nextPutAll: name; cr].
	aStream nextPutAll: 'cancel'.

	choice _ (PopUpMenu labels: aStream contents
				lines: {declareSize. declareSize + alternatives size})
		startUpWithCaption:
(('Unknown variable: ', proposedVariable, '
please correct, or cancel:') asText makeBoldFrom: 19 to: 19 + proposedVariable size).

	(choice = 0 or: [choice > (declareSize + alternatives size)]) ifTrue: [^ self fail].

	requestor deselect.
	requestor selectInvisiblyFrom: userSelection first to: userSelection last.
	temp ifTrue: [
		choice = 1 ifTrue: [^ self declareTempAndPaste: proposedVariable]
	] ifFalse: [
		choice = 1 ifTrue: [^ self declareGlobal: proposedVariable].
		choice = 2 ifTrue: [^ self declareClass: proposedVariable].
		(declareSize = 3 and: [choice = 3]) ifTrue: [^ self declareClassVar: proposedVariable]
	].

	"Spelling correction"
	variableNode _ self lookupVar: (alternatives at: choice - declareSize)
			ifAbsent: [self halt: 'var should have been found'].
	self substituteWord: variableNode name wordInterval: spot offset: 0.
	^ variableNode! !

!Parser2 methodsFor: 'var correction' stamp: 'ajh 8/17/2002 11:26'!
declareClass: name

	| s |
	s _ (Object newSubclass: name) name.
	^ self poolVar: (Smalltalk associationAt: s)! !

!Parser2 methodsFor: 'var correction' stamp: 'ajh 9/12/2002 00:00'!
declareClassVar: name

	| s |
	s _ name asSymbol.
	self classEncoding theNonMetaClass addClassVarName: s.
	^ self lookupVar: s! !

!Parser2 methodsFor: 'var correction' stamp: 'ajh 8/17/2002 11:25'!
declareGlobal: name

	| s |
	s _ name asSymbol.
	Smalltalk at: s put: nil.
	^ self poolVar: (Smalltalk associationAt: s)! !

!Parser2 methodsFor: 'var correction' stamp: 'ajh 9/14/2002 09:11'!
declareTempAndPaste: name
	| insertion delta theTextString characterBeforeMark |

	theTextString _ requestor text string.
	characterBeforeMark _ theTextString at: tempsMark-1 ifAbsent: [$ ].
	(theTextString at: tempsMark) = $| ifTrue: [
  		"Paste it before the second vertical bar"
		insertion _ name, ' '.
		characterBeforeMark isSeparator ifFalse: [insertion _ ' ', insertion].
		delta _ 0.
	] ifFalse: [
		"No bars - insert some with CR, tab"
		insertion _ '| ' , name , ' |',String cr.
		delta _ 2.	"the bar and CR"
		characterBeforeMark = Character tab ifTrue: [
			insertion _ insertion , String tab.
			delta _ delta + 1.	"the tab"
		].
	].
	tempsMark _ tempsMark +
		(self substituteWord: insertion
			wordInterval: (tempsMark to: tempsMark-1)
			offset: 0) - delta.
	^ homeBlockNode addTempVarNamed: name asArg: false! !

!Parser2 methodsFor: 'var correction' stamp: 'ajh 2/1/2003 01:38'!
removeUnusedTemps

	| str end start madeChanges tempName vars | 
	madeChanges _ false.
	str _ requestor text string.
	((tempsMark between: 1 and: str size) and: [(str at: tempsMark) = $|]) ifFalse: [^ self].
	[vars _ self blockNode unusedTemps] on: UndefinedVarSignal do: [:ex |
		tempName _ ex node var name.
		requestor selectFrom: ex node sourceRange first to: ex node sourceRange last.
		((PopUpMenu labels:
'yes
no') startUpWithCaption:
		((tempName , ' appears to be
undefined at this point.
Proceed anyway?') asText makeBoldFrom: 1 to: tempName size))
		= 1 ifTrue: [ex resume] ifFalse: [^ self fail]
	].
	vars do: [:temp |
		tempName _ temp name.
		"temp isArg ifTrue: [
		 PopUpMenu notify: ((tempName, ' appears to be
unused in this method.') asText makeBoldFrom: 1 to: tempName size)
		] ifFalse: ["
		 ((PopUpMenu labels: 'yes\no' withCRs) startUpWithCaption:
			((tempName , ' appears to be
unused in this method.
OK to remove it?') asText makeBoldFrom: 1 to: tempName size))
			= 1 ifTrue: [
			temp varUse isUnassigned ifTrue: [
				end _ tempsMark.
				["Beginning at right temp marker..."
				  start _ end - tempName size + 1.
				  end < tempName size or: [tempName = (str copyFrom: start to: end)
					and: [(str at: start-1) isSeparator & (str at: end+1) isSeparator]]
				] whileFalse: [
				  "Search left for the unused temp"
				  end _ requestor nextTokenFrom: end direction: -1
				].
				end < tempName size ifFalse: [
					(str at: start-1) = $  ifTrue: [start _ start-1].
					requestor correctFrom: start to: end with: ''.
					str _ str copyReplaceFrom: start to: end with: ''. 
					madeChanges _ true.
					tempsMark _ tempsMark - (end-start+1)]
			] ifFalse: [
				PopUpMenu notify: 'You''ll first have to remove the
statement where it''s stored into'
			]
		  ]
		"]"
	].
	madeChanges ifTrue: [ParserRemovedUnusedTemps signal].
! !

!Parser2 methodsFor: 'var correction' stamp: 'ajh 6/9/2002 19:47'!
substituteWord: correctWord wordInterval: spot offset: o
	"Substitute the correctSelector into the (presuamed interactive) receiver."

	requestor correctFrom: (spot first + o)
					to: (spot last + o)
					with: correctWord.

	requestorOffset _ requestorOffset + correctWord size - spot size.
	^ o + correctWord size - spot size! !

!Parser2 methodsFor: 'var correction' stamp: 'ajh 8/17/2002 11:32'!
undeclaredVariable: string

	| symbol |
	Transcript show: ' (' , string , ' is Undeclared) '.
	symbol _ string asSymbol.
	Undeclared at: symbol put: nil.
	^ self poolVar: (Undeclared associationAt: symbol)! !

!Parser2 methodsFor: 'private' stamp: 'ajh 6/9/2002 19:47'!
addComment

	parseNode ~~ nil
		ifTrue: 
			[parseNode comment: currentComment.
			currentComment _ nil]! !

!Parser2 methodsFor: 'private' stamp: 'ajh 10/3/2002 17:31'!
blockNode

	^ nestedBlockNodes top! !

!Parser2 methodsFor: 'private' stamp: 'ajh 10/3/2002 17:27'!
init: sourceStream notifying: req failBlock: aBlock

	requestor _ req.
	failBlock _ aBlock.
	super scan: sourceStream.
	prevMark _ hereMark _ mark.
	requestorOffset _ 0.
	nestedBlockNodes _ OrderedCollection new.  "stack"
	self advance.! !

!Parser2 methodsFor: 'private' stamp: 'ajh 2/9/2003 17:09'!
initPattern: aString notifying: req return: aBlock

	| result |
	self
		init: (ReadStream on: aString asString)
		notifying: req
		failBlock: [^nil].
	nestedBlockNodes push: BlockNode2 new.
	parseEnv _ nil parseEnv.
	doitFlag _ false.
	self messagePattern.
	result _ aBlock value: self blockNode.
	self popBlockNode.
	^ result! !

!Parser2 methodsFor: 'private' stamp: 'ajh 2/8/2003 18:22'!
newMethodNode

	^ BlockNode2 new! !

!Parser2 methodsFor: 'private' stamp: 'ajh 10/3/2002 17:28'!
popBlockNode

	^ nestedBlockNodes pop! !

!Parser2 methodsFor: 'private' stamp: 'ajh 10/3/2002 17:33'!
pushNewBlockNode

	^ nestedBlockNodes push: BlockNode2 new! !

!Parser2 methodsFor: 'primitives' stamp: 'ajh 6/9/2002 19:47'!
externalFunctionDeclaration
	"Parse the function declaration for a call to an external library."

	| descriptorClass callType retType externalName args argType module primNode |
	descriptorClass _ Smalltalk at: #ExternalFunction ifAbsent:[nil].
	descriptorClass == nil ifTrue:[^0].
	callType _ descriptorClass callingConventionFor: here.
	callType == nil ifTrue:[^0].
	"Parse return type"
	self advance.
	retType _ self externalType: descriptorClass.
	retType == nil ifTrue:[^self expected:'return type'].
	"Parse function name or index"
	externalName _ here.
	(self match: #string) 
		ifTrue:[externalName _ externalName asSymbol]
		ifFalse:[(self match:#number) ifFalse:[^self expected:'function name or index']].
	(self matchToken:'(' asSymbol) ifFalse:[^self expected:'argument list'].
	args _ WriteStream on: Array new.
	[here == #)] whileFalse:[
		argType _ self externalType: descriptorClass.
		argType == nil ifTrue:[^self expected:'argument'].
		argType isVoid & argType isPointerType not ifFalse:[args nextPut: argType].
	].
	(self matchToken:')' asSymbol) ifFalse:[^self expected:')'].
	(self matchToken: 'module:') ifTrue:[
		module _ here.
		(self match: #string) ifFalse:[^self expected: 'String'].
		module _ module asSymbol].
	primNode _ PrimitiveNode new num: 120.
	Smalltalk at: #ExternalLibraryFunction ifPresent:[:xfn |
		primNode spec: (xfn name: externalName 
						module: module 
						callType: callType
						returnType: retType
						argumentTypes: args contents)
	].
	^ primNode! !

!Parser2 methodsFor: 'primitives' stamp: 'ajh 6/9/2002 19:47'!
externalType: descriptorClass
	"Parse an return an external type"
	| xType |
	xType _ descriptorClass atomicTypeNamed: here.
	xType == nil ifTrue:["Look up from class scope"
		Symbol hasInterned: here ifTrue:[:sym|
			xType _ descriptorClass structTypeNamed: sym]].
	xType == nil ifTrue:[
		"Raise an error if user is there"
		self interactive ifTrue:[^nil].
		"otherwise go over it silently"
		xType _ descriptorClass forceTypeNamed: here].
	self advance.
	(self matchToken:#*)
		ifTrue:[^xType asPointerType]
		ifFalse:[^xType]! !

!Parser2 methodsFor: 'primitives' stamp: 'ajh 6/9/2002 19:47'!
primitive

	| primNode |
	(self matchToken: #<) ifFalse: [^ PrimitiveNode null].
	primNode _ self primitiveDeclarations.
	(self matchToken: #>) ifFalse: [^ self expected: '>'].
	^ primNode! !

!Parser2 methodsFor: 'primitives' stamp: 'ajh 6/9/2002 19:47'!
primitiveDeclarations

	| prim module |
	(self matchToken: 'primitive:') ifFalse: [
		^ self externalFunctionDeclaration].
	prim _ here.
	(self match: #number) ifTrue: [
		"Indexed primitives"
		^ PrimitiveNode new num: prim].
	(self match: #string) ifFalse: [^ self expected: 'Integer or String'].
	(self matchToken: 'module:') ifTrue:[
		module _ here.
		(self match: #string) ifFalse: [^ self expected: 'String'].
		module _ module asSymbol].
	^ PrimitiveNode new
		num: 117;
		spec: {module. prim asSymbol. 0. 0}! !

!Parser2 methodsFor: 'accessing' stamp: 'ajh 2/8/2003 18:18'!
classEncoding
	"what behavior am I parsing for, if none return nil"

	^ parseEnv classEncoding! !

!Parser2 methodsFor: 'accessing' stamp: 'ajh 6/9/2002 19:47'!
encoder

	^ self! !

!Parser2 methodsFor: 'accessing' stamp: 'ajh 6/9/2002 19:47'!
requestor
	"Return the source code editor"

	^ requestor! !

!Parser2 methodsFor: 'accessing' stamp: 'ajh 6/9/2002 19:47'!
requestor: editor
	"set the source code editor"

	requestor _ editor! !

!Parser2 methodsFor: 'source mapping' stamp: 'ajh 8/17/2002 10:17'!
noteSourceRange: range forNode: node

	node sourceRange: range.
	^ node! !

!Parser2 methodsFor: 'selectors' stamp: 'ajh 6/10/2002 06:23'!
choose: alternatives selector: proposedName interval: spot
	"Correct the proposedVariable to one in alternatives.
	spot is the interval of the variable within the test stream."

	| aStream choice userSelection |
	"If we can't ask the user for correction, make it ambiguous"
	self interactive ifFalse: [^ self ambiguousSelector: proposedName].

	userSelection _ requestor selectionInterval.
	requestor selectFrom: spot first to: spot last.
	requestor select.

	aStream _ WriteStream on: (String new: 200).
	alternatives do: [:selNode | aStream nextPutAll: selNode fullName; cr].
	aStream nextPutAll: 'cancel'.

	choice _ (PopUpMenu labels: aStream contents
				lines: {alternatives size})
		startUpWithCaption:
(('Ambiguous selector: ', proposedName, '
please choose, or cancel:') asText makeBoldFrom: 21 to: 21 + proposedName size).

	(choice = 0 or: [choice > alternatives size]) ifTrue: [^ self fail].

	requestor deselect.
	requestor selectInvisiblyFrom: userSelection first to: userSelection last.

	^ alternatives at: choice
! !

!Parser2 methodsFor: 'selectors' stamp: 'ajh 8/22/2002 21:53'!
correctSelector: proposedKeyword wordIntervals: spots exprInterval: expInt ifAbort: abortAction
	"Correct the proposedKeyword to some selector symbol, correcting the original text if such action is indicated.  abortAction is invoked if the proposedKeyword couldn't be converted into a valid selector.  Spots is an ordered collection of intervals within the test stream of the for each of the keyword parts."

	| alternatives aStream choice correctSelector userSelection lines firstLine |
	"If we can't ask the user, assume that the keyword will be defined later"
	self interactive ifFalse: [^ self unknownSelector: proposedKeyword].

	userSelection _ requestor selectionInterval.
	requestor selectFrom: spots first first to: spots last last.
	requestor select.
	alternatives _ self possibleSelectorsFor: proposedKeyword.

	aStream _ WriteStream on: (String new: 200).
	aStream nextPutAll: (proposedKeyword contractTo: 35); cr.
	firstLine _ 1.
 	alternatives do: [:sel | aStream nextPutAll: sel; cr].
	aStream nextPutAll: 'cancel'.
	lines _ Array with: firstLine with: (alternatives size + firstLine).
	
	choice _ (PopUpMenu labels: aStream contents lines: lines)
		startUpWithCaption: 
'Unknown selector, please 
confirm, correct, or cancel'.

	(choice = 0) | (choice > (lines at: 2)) ifTrue: [^ abortAction value].

	requestor deselect.
	requestor selectInvisiblyFrom: userSelection first to: userSelection last.

	choice = 1 ifTrue: [^ self createNewSelector: proposedKeyword].
	correctSelector _ alternatives at: choice - 1.
	self substituteSelector: correctSelector keywords wordIntervals: spots.
	((proposedKeyword last ~~ $:) and: [correctSelector last == $:]) ifTrue: [
		^ abortAction value].
	^ correctSelector.
! !

!Parser2 methodsFor: 'selectors' stamp: 'ajh 8/17/2002 11:13'!
createNewSelector: string

	^ string asSymbol! !

!Parser2 methodsFor: 'selectors' stamp: 'ajh 8/17/2002 11:09'!
lookupSelector: name ifNone: block ifMany: manyBlock

	Symbol hasInterned: name ifTrue: [ :sym | ^ sym].
	^ block value! !

!Parser2 methodsFor: 'selectors' stamp: 'ajh 6/10/2002 11:25'!
lookupSelector: string wordIntervals: wordInts exprInterval: exprInt

	^ self lookupSelector: string
		ifNone: [self correctSelector: string
					wordIntervals: wordInts
					exprInterval: exprInt
					ifAbort: [^ self fail]]
		ifMany: [:set | self choose: set selector: string
					wordIntervals: wordInts
					exprInterval: exprInt
					ifAbort: [^ self fail]]! !

!Parser2 methodsFor: 'selectors' stamp: 'ajh 8/17/2002 11:12'!
possibleSelectorsFor: string

	^ Symbol possibleSelectorsFor: string! !

!Parser2 methodsFor: 'selectors' stamp: 'ajh 6/9/2002 19:47'!
substituteSelector: selectorParts wordIntervals: spots
	"Substitute the correctSelector into the (presuamed interactive) receiver."
	| offset |
	offset _ 0.
	selectorParts with: spots do:
		[ :word :interval |
		offset _ self substituteWord: word wordInterval: interval offset: offset ]
! !

!Parser2 methodsFor: 'selectors' stamp: 'ajh 8/17/2002 11:11'!
unknownSelector: string

	^ string asSymbol! !

!Parser2 methodsFor: 'basic nodes' stamp: 'ajh 8/2/2002 18:53'!
falseNode

	^ ConstantNode new object: false! !

!Parser2 methodsFor: 'basic nodes' stamp: 'ajh 6/10/2002 12:35'!
literalNode: object

	^ ConstantNode new object: object! !

!Parser2 methodsFor: 'basic nodes' stamp: 'ajh 8/2/2002 18:53'!
nilNode

	^ ConstantNode new object: nil! !

!Parser2 methodsFor: 'basic nodes' stamp: 'ajh 9/12/2002 00:41'!
selfNode

	^ (self lookupVar: 'self') referenceNode! !

!Parser2 methodsFor: 'basic nodes' stamp: 'ajh 8/2/2002 18:53'!
trueNode

	^ ConstantNode new object: true! !


!DialectParser2 methodsFor: 'as yet unclassified' stamp: 'ajh 9/22/2002 23:20'!
blockExpression
	"[ ({:var} |) (| {temps} |) (statements) ] => BlockNode."

	| start |
	self pushNewBlockNode.
	start _ prevMark.

	"Gather parameters."
	(self matchToken: 'With') ifTrue:
		[[self match: #period]
			whileFalse: [self declareArg: self argumentName]].

	self temporaryBlockVariables.
	self statementsInnerBlock: true.

	(self match: #rightBracket) ifFalse: [^ self expected: 'Period or right bracket'].

	self noteSourceRange: (start to: self endOfLastToken) forNode: self blockNode.
	parseNode _ self blockNode.
	self popBlockNode.
! !

!DialectParser2 methodsFor: 'as yet unclassified' stamp: 'ajh 9/11/2002 12:26'!
expression

	^ self expressionWithInitialKeyword: ''
! !

!DialectParser2 methodsFor: 'as yet unclassified' stamp: 'ajh 9/11/2002 12:38'!
expressionWithInitialKeyword: kwdIfAny

	| checkpoint |
	(hereType == #word and: [here = 'Set' and: [tokenType == #word]]) ifTrue:
			["Parse assignment statement 'Set' var 'to' expression"
			checkpoint _ self checkpoint.
			self advance.
			token = 'to'
				ifTrue: [^ self assignment]
				ifFalse: [self revertToCheckpoint: checkpoint]].
	self matchKeyword
		ifTrue: ["It's an initial keyword."
				kwdIfAny isEmpty ifFalse: [self error: 'compiler logic error'].
				^ self expressionWithInitialKeyword: ':' , self advance , ':'].
	hereType == #leftBrace
		ifTrue: [self braceExpression]
		ifFalse: [self primaryExpression ifFalse: [^ false]].
	(self messagePart: 3 repeat: true initialKeyword: kwdIfAny)
		ifTrue: [hereType == #semicolon ifTrue: [self cascade]].
	^ true! !

!DialectParser2 methodsFor: 'as yet unclassified' stamp: 'ajh 9/11/2002 12:26'!
matchKeyword
	"Return true if we are looking at a keyword (and its argument)."

	hereType == #word ifFalse: [^ false].
	tokenType == #leftParenthesis ifTrue: [^ true].
	tokenType == #leftBracket ifTrue: [^ true].
	tokenType == #leftBrace ifTrue: [^ true].
	^ false! !

!DialectParser2 methodsFor: 'as yet unclassified' stamp: 'ajh 9/11/2002 12:26'!
matchReturn

	^ self matchToken: 'Answer'! !

!DialectParser2 methodsFor: 'as yet unclassified' stamp: 'ajh 9/11/2002 12:26'!
messagePart: level repeat: repeat

	^ self messagePart: level repeat: repeat initialKeyword: ''! !

!DialectParser2 methodsFor: 'as yet unclassified' stamp: 'ajh 9/11/2002 12:53'!
messagePart: level repeat: repeat initialKeyword: kwdIfAny

	| start receiver selector args words keywordStart |
	[receiver _ parseNode.
	(self matchKeyword and: [level >= 3])
		ifTrue: 
			[start _ self startOfNextToken.
			selector _ WriteStream on: (String new: 32).
			selector nextPutAll: kwdIfAny.
			args _ OrderedCollection new.
			words _ OrderedCollection new.
			[self matchKeyword]
				whileTrue: 
					[keywordStart _ self startOfNextToken + requestorOffset.
					selector nextPutAll: self advance , ':'.
					words addLast: (keywordStart to: hereEnd + requestorOffset).
					self primaryExpression ifFalse: [^ self expected: 'Argument'].
					args addLast: parseNode].
			selector _ self lookupSelector: selector contents wordIntervals: words exprInterval: (start to: self endOfLastToken)]
		ifFalse: [((hereType == #binary or: [hereType == #verticalBar])
				and: [level >= 2])
				ifTrue: 
					[start _ self startOfNextToken.
					selector _ self advance.
					self primaryExpression ifFalse: [^self expected: 'Argument'].
					self messagePart: 1 repeat: true.
					args _ Array with: parseNode.
					selector _ self lookupSelector: selector wordIntervals: words exprInterval: (start to: self endOfLastToken)]
				ifFalse: [(hereType == #word
							and: [(#(leftParenthesis leftBracket leftBrace) includes: tokenType) not])
						ifTrue: 
							[start _ self startOfNextToken.
							selector _ self advance.
							args _ #().
							words _ OrderedCollection with: (start  + requestorOffset to: self endOfLastToken + requestorOffset).
							selector _ self lookupSelector: selector wordIntervals: words exprInterval: (start to: self endOfLastToken)]
						ifFalse: [^args notNil]]].
	parseNode _ MessageNode2 new receiver: receiver selector: selector arguments: args.
	self noteSourceRange: (start to: self endOfLastToken) forNode: parseNode.
	repeat]
		whileTrue: [].
	^true! !

!DialectParser2 methodsFor: 'as yet unclassified' stamp: 'ajh 9/11/2002 12:56'!
newMethodNode

	^ DialectMethodNode2 new setDialect: #SQ00! !

!DialectParser2 methodsFor: 'as yet unclassified' stamp: 'ajh 2/7/2003 22:40'!
parseArgsAndTemps: aString notifying: req 
	"Parse the argument, aString, notifying req if an error occurs. Otherwise, 
	answer a two-element Array containing Arrays of strings (the argument 
	names and temporary variable names)."

	aString == nil ifTrue: [^ #()].
	doitFlag _ false.		"Don't really know if a doit or not!!"
	^ self initPattern: aString
		notifying: req
		return: [:blockNode |
			self temporaries.
			blockNode tempVars collect: [:v | v name]
		]! !

!DialectParser2 methodsFor: 'as yet unclassified' stamp: 'ajh 9/11/2002 13:13'!
pattern: fromDoit inContext: ctxt 
	" unarySelector | binarySelector arg | keyword arg {keyword arg} =>  
	{selector, arguments, precedence}."

	| args selector checkpoint |
	doitFlag _ fromDoit.
	fromDoit ifTrue: [
		ctxt == nil ifTrue: [
			self blockNode selector: (self lookupSelector: #DOIT ifNone: [self halt: 'DOIT should be a selector seen in ProtoObject'] ifMany: [:set | self halt: 'Multiple DOIT selectors']).
		] ifFalse: [
			self blockNode selector: (self lookupSelector: #DOITIN: ifNone: [self halt: 'DOIT should be a selector seen in ProtoObject'] ifMany: [:set | self halt: 'Multiple DOIT selectors']).
			self declareArg: 'homeContext'.
		].
		^ self
].

	"NOTE: there is now an ambiguity between
	keywordSelector (argName) -and- unarySelector (first expression).
	Also, there is an amibuity (if there are no temp declarations) between
	keywordSelector (argName) -and- PrefixKeyword (some expression).
	We use duct tape for now."
	(hereType == #word and: [tokenType == #leftParenthesis]) ifTrue:
		[checkpoint _ self checkpoint.  "in case we have to back out"
		selector _ WriteStream on: (String new: 32).
			args _ OrderedCollection new.
			[hereType == #word
				and: [tokenType == #leftParenthesis
				and: [here first isLowercase
						or: [(#('Test' 'Repeat' 'Answer') includes: here) not]]]]
				whileTrue: 
					[selector nextPutAll: self advance , ':'.  "selector part"
					self advance.  "open paren"
					(args size = 0 and: [tokenType ~~ #rightParenthesis]) ifTrue:
						["This is really a unary selector on a method that
						begins with a parenthesized expression.  Back out now"
						self revertToCheckpoint: checkpoint.
						self blockNode selector: self advance asSymbol.
						^ self].
					self declareArg: self argumentName.
					(self match: #rightParenthesis)
						ifFalse: [^ self expected: 'right parenthesis']].
			self blockNode selector: selector contents asSymbol.
			^ self].

	hereType == #word ifTrue: [
		self blockNode selector: self advance asSymbol.
		^ self].

	(hereType == #binary or: [hereType == #verticalBar])
		ifTrue: 
			[self blockNode selector: self advance asSymbol.
			self declareArg: self argumentName.
			^ self].

	^ self expected: 'Message pattern'! !

!DialectParser2 methodsFor: 'as yet unclassified' stamp: 'ajh 2/7/2003 22:34'!
temporaries
	" [ 'Use' (variable)* '.' ]"

	| theActualText |
	(self matchToken: #'Use') ifFalse: 
		["no temps"
		doitFlag ifTrue: [requestor
				ifNil: [tempsMark _ 1]
				ifNotNil: [tempsMark _ requestor selectionInterval first].
			^ self].
		tempsMark _ prevEnd+1.
		tempsMark > 0 ifTrue:
			[theActualText _ source contents.
			[tempsMark < theActualText size and: [(theActualText at: tempsMark) isSeparator]]
				whileTrue: [tempsMark _ tempsMark + 1]].
			^ self].
	[hereType == #word]
		whileTrue: [self declareTemp: self advance].
	(self match: #period) ifTrue: 
		[tempsMark _ prevMark.
		^ self].
	^ self expected: 'Period'! !


!Parser2 class methodsFor: 'as yet unclassified' stamp: 'ajh 8/2/2002 18:52'!
falseNode

	^ ConstantNode new object: false! !

!Parser2 class methodsFor: 'as yet unclassified' stamp: 'ajh 8/2/2002 18:52'!
nilNode

	^ ConstantNode new object: nil! !

!Parser2 class methodsFor: 'as yet unclassified' stamp: 'ajh 8/2/2002 18:52'!
trueNode

	^ ConstantNode new object: true! !


!DialectParser2 class methodsFor: 'as yet unclassified' stamp: 'ajh 9/11/2002 12:26'!
test    "DialectParser test"

"PrettyPrints the source for every method in the system in the alternative syntax, and then compiles that source and verifies that it generates identical code.  No changes are actually made to the system.  At the time of this writing, only two methods caused complaints (reported in Transcript and displayed in browse window after running):

	BalloonEngineSimulation circleCosTable and
	BalloonEngineSimulation circleSinTable.

These are not errors, but merely a case of Floats embedded in literal arrays, and thus not specially checked for roundoff errors.

Note that if an error or interruption occurs during execution of this method, the alternativeSyntax preference will be left on.

NOTE:  Some methods may not compare properly until the system has been recompiled once.  Do this by executing...
		Smalltalk recompileAllFrom: 'AARDVAARK'.
"

	 | newCodeString methodNode oldMethod newMethod badOnes n heading |
	Preferences enable: #printAlternateSyntax.
	badOnes _ OrderedCollection new.
	Transcript clear.
	Smalltalk forgetDoIts.
'Formatting and recompiling all classes...'
displayProgressAt: Sensor cursorPoint
from: 0 to: Smalltalk allMethodsCount
during: [:bar | n _ 0.
	Smalltalk allClassesDo:  "{MethodNode} do:"  "<- to check one class"
		[:nonMeta |  "Transcript cr; show: nonMeta name."
		{nonMeta. nonMeta class} do:
		[:cls |
		cls selectors do:
			[:selector | (n _ n+1) \\ 100 = 0 ifTrue: [bar value: n].
			newCodeString _ (cls compilerClass new)
				format: (cls sourceCodeAt: selector)
				in: cls notifying: nil decorated: Preferences colorWhenPrettyPrinting.
			heading _ cls organization categoryOfElement: selector.
			methodNode _ cls compilerClass new
						compile: newCodeString
						in: cls notifying: (SyntaxError new category: heading)
						ifFail: [].
			newMethod _ methodNode generate: #(0 0 0 0).
			oldMethod _ cls compiledMethodAt: selector.
			"Transcript cr; show: cls name , ' ' , selector."
			oldMethod = newMethod ifFalse:
				[Transcript cr; show: '***' , cls name , ' ' , selector.
				oldMethod size = newMethod size ifFalse:
					[Transcript show: ' difft size'].
				oldMethod header = newMethod header ifFalse:
					[Transcript show: ' difft header'].
				oldMethod literals = newMethod literals ifFalse:
					[Transcript show: ' difft literals'].
				Transcript endEntry.
				badOnes add: cls name , ' ' , selector]]]].
].
	Smalltalk browseMessageList: badOnes asSortedCollection name: 'Formatter Discrepancies'.
	Preferences disable: #printAlternateSyntax.
! !


!SendSelectorInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 17:40'!
addTo: sequence

	super addTo: sequence.
	isSpecial _ Compiler bytecodeWriteStream specialSelectors includesKey: selector.
	isSpecial ifFalse: [sequence method addLiteral: selector].
! !

!SendSelectorInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:04'!
bytecodesOn: stream

	isSpecial
		ifTrue: [stream sendSpecial: selector]
		ifFalse: [stream
			sendLiteral: (owningSequence method literalSlotIndexOf: selector)
			numArgs: selector numArgs]
! !

!SendSelectorInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:02'!
printOn: stream

	stream nextPutAll: 'send '.
	selector printOn: stream.
! !

!SendSelectorInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:02'!
selector: symbol

	selector _ symbol! !

!SendSelectorInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 15:02'!
stackEffect: parseStack
	"selector is on top of stack, with args underneath, pop them"

	parseStack pop: selector numArgs! !


!SendSelectorSuperInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/21/2003 12:23'!
addTo: sequence

	super addTo: sequence.
	isSpecial ifTrue: [
		"super calls are never special"
		sequence method addLiteral: selector.
		isSpecial _ false.  
	].
	sequence method addLastLiteral: behavior holder.
! !

!SendSelectorSuperInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/12/2002 17:29'!
behavior: aBehavior

	behavior _ aBehavior! !

!SendSelectorSuperInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/21/2003 12:36'!
bytecodesOn: stream

	stream
		sendLiteral: (owningSequence method literalSlotIndexOf: selector)
		toSuperOfLastLiteralNumArgs: selector numArgs
! !

!SendSelectorSuperInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/12/2002 17:29'!
printOn: stream

	stream nextPutAll: 'sendSuper '.
	selector printOn: stream.
! !


!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 9/17/2001 21:54'!
printOn: stream

	super printOn: stream.
	stream nextPut: $(.
	value printOn: stream.
	stream nextPut: $).
! !

!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 20:25'!
tempHolderValue

	^ value! !

!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 20:07'!
tempValue

	^ value! !

!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 20:07'!
tempValue: obj

	value _ obj! !

!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 9/15/2001 19:50'!
value

	^ value! !

!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 9/15/2001 19:50'!
value: obj

	^ value _ obj! !


!SharedTempHolder class methodsFor: 'as yet unclassified' stamp: 'ajh 6/5/2002 22:55'!
valueSlotIndex
	"Return the slot index of 'value' in my instances"

	^ 1! !

!SharedTempHolder class methodsFor: 'as yet unclassified' stamp: 'ajh 9/26/2001 09:37'!
with: value

	^ self new
		value: value;
		yourself! !


!StoreInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 21:01'!
bytecodesOn: stream

	stream storeInstVar: instVarIndex pop: self pops! !

!StoreInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/24/2002 14:05'!
doPop

	pop _ true! !

!StoreInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/24/2002 14:05'!
pops

	^ pop == true! !

!StoreInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 20:58'!
pops: bool

	pop _ bool! !

!StoreInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 21:00'!
printOn: stream

	stream nextPutAll: 'store'.
	stream nextPutAll: (self pops ifTrue: ['Pop '] ifFalse: [' ']).
	stream nextPutAll: 'instVar '.
	instVarIndex printOn: stream.
! !

!StoreInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/26/2002 19:35'!
slot

	^ instVarIndex! !

!StoreInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/26/2002 19:36'!
slot: index

	instVarIndex _ index! !

!StoreInstVarInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 21:00'!
stackEffect: parseStack

	self pops ifTrue: [parseStack pop]! !


!String methodsFor: 'compiling' stamp: 'ajh 6/14/2002 17:55'!
isAndOrSelector

	^ #(and: or:) includes: self! !

!String methodsFor: 'compiling' stamp: 'ajh 6/14/2002 17:55'!
isCaseSelector

	^ #(caseOf: caseOf:otherwise:) includes: self! !

!String methodsFor: 'compiling' stamp: 'ajh 6/14/2002 17:56'!
isIfNilSelector

	^ #(ifNil: ifNotNil: ifNil:ifNotNil: ifNotNil:ifNil:) includes: self! !

!String methodsFor: 'compiling' stamp: 'ajh 6/14/2002 17:56'!
isIfSelector

	^ #(ifTrue: ifFalse: ifTrue:ifFalse: ifFalse:ifTrue:) includes: self! !

!String methodsFor: 'compiling' stamp: 'ajh 6/14/2002 17:59'!
isToDoSelector

	^ #(to:do: to:by:do:) includes: self! !

!String methodsFor: 'compiling' stamp: 'ajh 6/14/2002 18:02'!
isWhileSelector

	^ #(whileFalse: whileTrue: whileFalse whileTrue) includes: self! !


!TMethod methodsFor: 'initialization' stamp: 'ajh 1/21/2003 13:10'!
setSelector2: sel args: argList locals: localList block: tBlockNode primitive: aNumber
	"Initialize this method using the given information."

	selector _ sel.
	returnType _ 'int'. 	 "assume return type is int for now"
	args _ argList asOrderedCollection.
	locals _ localList asOrderedCollection.
	declarations _ Dictionary new.
	primitive _ aNumber.
	parseTree _ tBlockNode.
	labels _ OrderedCollection new.
	complete _ false.  "set to true when all possible inlining has been done"
	export _ self extractExportDirective.
	static _ self extractStaticDirective.
	self removeFinalSelfReturn.
	self recordDeclarations.
! !


!TempVarInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 9/24/2002 15:52'!
tempVar

	^ tempVar! !

!TempVarInstruction methodsFor: 'as yet unclassified' stamp: 'ajh 9/24/2002 15:49'!
tempVar: aTempVar

	tempVar _ aTempVar! !


!DeclareTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/27/2002 13:50'!
assignTempStackPosition

	tempVar fpOffset: self tempAllocPosition! !

!DeclareTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/27/2002 13:58'!
bytecodesOn: stream
	"no effect"
! !

!DeclareTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/4/2002 18:18'!
deallocInstr

	^ deallocInstr! !

!DeclareTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/4/2002 18:18'!
deallocInstr: popInstr

	deallocInstr _ popInstr! !

!DeclareTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/4/2002 17:04'!
printOn: stream

	stream nextPutAll: 'name '; print: tempVar.
	tempVar fpOffset ifNotNil: [
		stream nextPutAll: ' (fp+'.
		tempVar fpOffset printOn: stream.
		stream nextPut: $).
	].
! !

!DeclareTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/26/2002 19:48'!
stackEffect: parseStack

	"no effect"! !

!DeclareTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 9/27/2002 12:51'!
tempAllocPosition

	^ self spBefore! !


!PushTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 21:02'!
bytecodesOn: stream

	stream pushTemp: tempVar fpOffset! !

!PushTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 10/12/2002 13:47'!
convertToLocalReturnSelf

	| obj |
	owningSequence localParseStack pop.
	self primitiveChangeClassTo: LocalReturnSelfInstr basicNew.
	"Associate self with mapToObject"
	obj _ owningSequence method mapToObject.
	obj ifNotNil: [obj mapInstr: self].
! !

!PushTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/21/2003 00:03'!
convertToPushInstVar: instVarIndex

	| obj |
	owningSequence removeLast.
	owningSequence localParseStack pop.
	obj _ owningSequence method mapToObject.
	obj ifNotNil: [obj initInstrMap].
	PushInstVarInstr new slot: instVarIndex; addTo: owningSequence.
! !

!PushTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/21/2003 00:03'!
convertToStoreInstVar: instVarIndex

	| obj |
	owningSequence removeLast.
	owningSequence localParseStack pop.
	obj _ owningSequence method mapToObject.
	obj ifNotNil: [obj initInstrMap].
	StoreInstVarInstr new slot: instVarIndex; addTo: owningSequence.
! !

!PushTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/19/2002 23:17'!
printOn: stream

	stream nextPutAll: 'push '.
	tempVar printOn: stream.
! !

!PushTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/12/2002 16:46'!
stackEffect: parseStack

	parseStack push! !


!StoreTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 21:02'!
bytecodesOn: stream

	stream storeTemp: tempVar fpOffset pop: self pops! !

!StoreTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/5/2002 22:02'!
doPop

	pop _ true! !

!StoreTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/5/2002 22:02'!
pops

	^ pop == true! !

!StoreTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/17/2002 17:30'!
pops: bool

	pop _ bool! !

!StoreTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/19/2002 23:20'!
printOn: stream

	stream nextPutAll: 'store'.
	stream nextPutAll: (self pops ifTrue: ['Pop '] ifFalse: [' ']).
	tempVar printOn: stream.
! !

!StoreTempInstr methodsFor: 'as yet unclassified' stamp: 'ajh 6/12/2002 16:48'!
stackEffect: parseStack

	self pops ifTrue: [parseStack pop]! !


!TestCodeGenerator methodsFor: 'coercing' stamp: 'ajh 2/7/2003 21:27'!
ccgTVarBlock: anInteger

	| ctxtString |
	ctxtString _ Preferences compileBlocksAsClosures
		ifTrue: ['thisContext']  "thisContext refers to captured context not eval context"
		ifFalse: ['(thisContext tempAt: 1)'].  "thisContext refers to eval context"
	^ [:expr | ctxtString, ' tempAt: ', anInteger asString, ' put: (', expr, ')']! !


!TestTMethod methodsFor: 'initializing' stamp: 'ajh 1/21/2003 13:11'!
setSelector2: sel args: argList locals: localList block: tBlockNode primitive: aNumber
	"Initialize this method using the given information."

	selector _ sel.
	returnType _ 'int'. 	 "assume return type is int for now"
	args _ argList asOrderedCollection.
	locals _ localList asOrderedCollection.
	declarations _ Dictionary new.
	primitive _ aNumber.
	parseTree _ tBlockNode.
	labels _ OrderedCollection new.
	complete _ false.  "set to true when all possible inlining has been done"
	export _ self extractExportDirective.
	static _ self extractStaticDirective.
	isPrimitive _ false.  "set to true only if you find a primtive direction."
	suppressingFailureGuards _ self extractSuppressFailureGuardDirective.
	self recordDeclarations.
	self extractPrimitiveDirectives.
! !


!UndefinedObject methodsFor: 'class hierarchy' stamp: 'ajh 10/15/2002 04:10'!
parseEnv

	^ ParseEnv new! !

!UndefinedObject methodsFor: 'class hierarchy' stamp: 'ajh 1/21/2003 13:03'!
parserClass

	^ Compiler parserClass! !


!UndefinedVarSignal methodsFor: 'as yet unclassified' stamp: 'ajh 2/1/2003 00:45'!
node

	^ refNode! !

!UndefinedVarSignal methodsFor: 'as yet unclassified' stamp: 'ajh 2/1/2003 00:45'!
node: refOrAssignmentOrReturnNode

	refNode _ refOrAssignmentOrReturnNode! !

!UndefinedVarSignal methodsFor: 'as yet unclassified' stamp: 'ajh 2/1/2003 00:46'!
var

	^ refNode var! !

!UndefinedVarSignal methodsFor: 'as yet unclassified' stamp: 'ajh 2/1/2003 00:46'!
varName

	^ refNode var name! !


!VarAssignmentNode methodsFor: 'initializing' stamp: 'ajh 2/8/2003 18:52'!
variable: aVarNode value: aParseNode

	var _ aVarNode.
	valueNode _ aParseNode.
! !

!VarAssignmentNode methodsFor: 'accessing' stamp: 'ajh 6/12/2002 17:10'!
childrenParseNodesDo: block

	block value: valueNode! !

!VarAssignmentNode methodsFor: 'accessing' stamp: 'ajh 6/29/2002 16:13'!
valueNode

	^ valueNode! !

!VarAssignmentNode methodsFor: 'accessing' stamp: 'ajh 2/8/2003 18:52'!
var

	^ var! !

!VarAssignmentNode methodsFor: 'accessing' stamp: 'ajh 2/8/2003 18:52'!
variable

	^ var! !

!VarAssignmentNode methodsFor: 'var usage' stamp: 'ajh 2/8/2003 18:52'!
calcVarUsage: usage

	valueNode calcVarUsage: usage.
	var usage: usage nowHasDef: self.
! !

!VarAssignmentNode methodsFor: 'code generation' stamp: 'ajh 2/8/2003 18:52'!
emit: method forValue: bool

	var emit: method assignment: self forValue: bool.
! !

!VarAssignmentNode methodsFor: 'evaluating' stamp: 'ajh 2/8/2003 18:52'!
evaluate

	^ var evaluateAssignTo: valueNode evaluate! !

!VarAssignmentNode methodsFor: 'printing' stamp: 'ajh 2/8/2003 18:51'!
printOn: aStream indent: level

	aStream dialect = #SQ00
		ifTrue: [aStream withStyleFor: #setOrReturn do: [aStream nextPutAll: 'Set '].
				var printOn: aStream indent: level.
				self startSourceRange: aStream position.
				aStream withStyleFor: #setOrReturn do: [aStream nextPutAll: ' to '].
				valueNode printOn: aStream indent: level + 2]
		ifFalse: [var printOn: aStream indent: level.
				self startSourceRange: aStream position.
				aStream nextPutAll: ' _ '.
				valueNode printOn: aStream indent: level + 2].
	self endSourceRange: aStream position.
! !

!VarAssignmentNode methodsFor: 'printing' stamp: 'ajh 6/10/2002 15:58'!
printOn: aStream indent: level precedence: p

	(aStream dialect = #SQ00
			ifTrue: [p < 3]
			ifFalse: [p < 4])
		ifTrue: [aStream nextPutAll: '('.
				self printOn: aStream indent: level.
				aStream nextPutAll: ')']
		ifFalse: [self printOn: aStream indent: level]! !

!VarAssignmentNode methodsFor: 'decompiling' stamp: 'ajh 2/8/2003 18:51'!
toDoIncrement: aVarNode

	var = aVarNode ifFalse: [^ nil].
	^ (valueNode isKindOf: MessageNode2) 
		ifTrue: [valueNode toDoIncrement: var]
		ifFalse: [nil]! !

!VarAssignmentNode methodsFor: 'decompiling' stamp: 'ajh 8/2/2002 22:27'!
value

	^ valueNode! !

!VarAssignmentNode methodsFor: 'C translation' stamp: 'ajh 2/9/2003 13:41'!
asTranslatorNode

	^ TAssignmentNode new
		setVariable: var asTranslatorNode
		expression: valueNode asTranslatorNode;
		comment: comment! !


!VarNode methodsFor: 'accessing' stamp: 'ajh 9/14/2002 15:45'!
instVarNode: varName at: index

	^ MessageVarNode new
		name: varName;
		getMessageNode: (MessageNode2 new
			receiver: self referenceNode
			selector: #instVarAt:
			arguments: {ConstantNode new object: index});
		setMessageNode: (MessageNode2 new
			receiver: self referenceNode
			selector: #instVarAt:put:
			arguments: {ConstantNode new object: index. TempVarNode dummy})! !

!VarNode methodsFor: 'accessing' stamp: 'ajh 6/14/2002 17:48'!
name

	^ self subclassResponsibility! !

!VarNode methodsFor: 'testing' stamp: 'ajh 9/12/2002 00:37'!
canAssign

	^ true! !

!VarNode methodsFor: 'testing' stamp: 'ajh 6/10/2002 15:08'!
isSelfPseudoVariable

	^ false! !

!VarNode methodsFor: 'testing' stamp: 'ajh 7/29/2002 10:42'!
isTemp

	^ false! !

!VarNode methodsFor: 'var usage' stamp: 'ajh 10/2/2002 00:19'!
usage: usage nowHasDef: assignmentNode
	"Do nothing, tempVars do something"! !

!VarNode methodsFor: 'var usage' stamp: 'ajh 10/2/2002 00:20'!
usage: usage nowHasRef: referenceNode
	"Do nothing, tempVars do something"! !

!VarNode methodsFor: 'code generation' stamp: 'ajh 9/24/2002 23:54'!
emit: method assignment: assignmentNode forValue: bool

	^ self subclassResponsibility! !

!VarNode methodsFor: 'code generation' stamp: 'ajh 9/27/2002 00:49'!
emitValue: method

	^ self subclassResponsibility! !

!VarNode methodsFor: 'code generation' stamp: 'ajh 9/27/2002 00:49'!
emitValue: method reference: referenceNode

	^ self emitValue: method! !

!VarNode methodsFor: 'evaluating' stamp: 'ajh 6/10/2002 14:30'!
evaluate
	"Get value from env"

	^ self subclassResponsibility! !

!VarNode methodsFor: 'evaluating' stamp: 'ajh 6/4/2002 16:36'!
evaluateAssignTo: object
	"Set value in env"

	^ self subclassResponsibility! !

!VarNode methodsFor: 'converting' stamp: 'ajh 9/30/2002 14:14'!
localTo: blockNode

	^ self! !

!VarNode methodsFor: 'converting' stamp: 'ajh 2/8/2003 18:55'!
referenceNode

	^ VarReferenceNode new var: self! !

!VarNode methodsFor: 'printing' stamp: 'ajh 6/12/2002 11:44'!
printOn: aStream 
	"Refer to the comment in Object|printOn:."

	aStream nextPutAll: ((DialectStream dialect: #ST80
								contents: [:strm | self printOn: strm indent: 0])
							asString).
! !

!VarNode methodsFor: 'printing' stamp: 'ajh 6/10/2002 16:07'!
printOn: aStream indent: level 
	aStream withStyleFor: #variable
		do: [aStream nextPutAll: self name].
! !

!VarNode methodsFor: 'C translation' stamp: 'ajh 9/11/2002 01:41'!
asTranslatorNode

	^ TVariableNode new setName: self name
! !


!InstVarNode methodsFor: 'initializing' stamp: 'ajh 6/15/2002 12:47'!
instVar: index
	"used to access the instance variable from the selfTempVar"

	instVarIndex _ index! !

!InstVarNode methodsFor: 'initializing' stamp: 'ajh 2/3/2003 22:56'!
name: stringOrNil

	name _ stringOrNil! !

!InstVarNode methodsFor: 'initializing' stamp: 'ajh 8/16/2002 19:13'!
selfVar: tempVar
	"this inst var comes from tempVar, which is expected to be the receiverTempVar of the outer methodNode (the selfVar)"

	selfVar _ tempVar! !

!InstVarNode methodsFor: 'initializing' stamp: 'ajh 10/3/2002 11:12'!
usage: usage nowHasDef: assignmentNode

	usage var: selfVar nowHasRef: assignmentNode! !

!InstVarNode methodsFor: 'initializing' stamp: 'ajh 10/3/2002 11:11'!
usage: usage nowHasRef: referenceNode

	usage var: selfVar nowHasRef: referenceNode! !

!InstVarNode methodsFor: 'accessing' stamp: 'ajh 7/29/2002 10:36'!
instVarIndex

	^ instVarIndex! !

!InstVarNode methodsFor: 'accessing' stamp: 'ajh 2/8/2003 17:39'!
name

	^ name ifNil: ['instVar', instVarIndex printString]! !

!InstVarNode methodsFor: 'accessing' stamp: 'ajh 8/16/2002 19:14'!
selfVar

	^ selfVar! !

!InstVarNode methodsFor: 'code generation' stamp: 'ajh 1/31/2003 21:42'!
emit: method assignment: assignmentNode forValue: bool

	assignmentNode valueNode emitValue: method.
	method mapTo: assignmentNode.
	selfVar emitValue: method.
	method setField: instVarIndex.
	bool ifFalse: [method doPop].
	method unMap.
! !

!InstVarNode methodsFor: 'code generation' stamp: 'ajh 1/15/2003 19:59'!
emitValue: method

	| i |
	i _ selfVar emitValue: method.
	method getField: instVarIndex.
	^ i! !

!InstVarNode methodsFor: 'code generation' stamp: 'ajh 10/3/2002 16:33'!
isIndirect
	"polymorphic with TempVarNode"

	^ false! !

!InstVarNode methodsFor: 'evaluating' stamp: 'ajh 6/15/2002 12:47'!
evaluate

	^ (thisContext env at: #receiver) instVarValue: instVarIndex! !

!InstVarNode methodsFor: 'evaluating' stamp: 'ajh 6/15/2002 12:47'!
evaluateAssignTo: object

	^ (thisContext env at: #receiver) instVarAt: instVarIndex put: object! !

!InstVarNode methodsFor: 'comparing' stamp: 'ajh 8/16/2002 19:16'!
= other

	^ other class == self class
		and: [instVarIndex = other instVarIndex
			and: [selfVar = other selfVar]]! !

!InstVarNode methodsFor: 'comparing' stamp: 'ajh 8/16/2002 19:15'!
hash

	^ instVarIndex hash bitXor: selfVar hash! !


!MessageVarNode methodsFor: 'initializing' stamp: 'ajh 9/14/2002 11:20'!
getMessageNode: messageNode

	getMessageNode _ messageNode! !

!MessageVarNode methodsFor: 'initializing' stamp: 'ajh 9/14/2002 18:21'!
name: string

	name _ string! !

!MessageVarNode methodsFor: 'initializing' stamp: 'ajh 9/14/2002 11:20'!
setMessageNode: messageNode
	"'TempVarNode dummy' in messageNode arguments will be substituted with the actual assigned value node when generated"

	messageNode argumentNodes
		detectIndex: [:node | node isDummyTemp]
		ifNone: [self error: 'MessageVarNode setMessage should have ''TempVarNode dummy'' as one of its arguments'].
	setMessageNode _ messageNode.
! !

!MessageVarNode methodsFor: 'accessing' stamp: 'ajh 9/14/2002 10:42'!
canAssign

	^ setMessageNode notNil! !

!MessageVarNode methodsFor: 'accessing' stamp: 'ajh 9/14/2002 10:43'!
name

	^ name! !

!MessageVarNode methodsFor: 'code generation' stamp: 'ajh 1/31/2003 21:41'!
emit: method assignment: assignmentNode forValue: bool

	| message i |
	message _ setMessageNode copy.
	i _ message argumentNodes detectIndex: [:node | node isDummyTemp].
	message argumentNodes at: i put: assignmentNode valueNode.
	message emit: method forValue: bool
! !

!MessageVarNode methodsFor: 'code generation' stamp: 'ajh 9/27/2002 00:49'!
emitValue: method

	^ getMessageNode emitValue: method! !

!MessageVarNode methodsFor: 'evaluating' stamp: 'ajh 9/14/2002 10:43'!
evaluate

	^ getMessageNode evaluate! !

!MessageVarNode methodsFor: 'evaluating' stamp: 'ajh 9/14/2002 10:56'!
evaluateAssignTo: object

	self toDo! !

!MessageVarNode methodsFor: 'comparing' stamp: 'ajh 9/14/2002 10:56'!
= other

	^ other class == self class and: [other name = name]! !

!MessageVarNode methodsFor: 'var usage' stamp: 'ajh 10/17/2002 18:06'!
usage: usage nowHasDef: assignmentNode

	setMessageNode calcVarUsage: usage! !

!MessageVarNode methodsFor: 'var usage' stamp: 'ajh 10/17/2002 18:16'!
usage: usage nowHasRef: referenceNode

	getMessageNode calcVarUsage: usage! !


!PoolVarNode methodsFor: 'initializing' stamp: 'ajh 6/5/2002 06:49'!
poolVariable: aPoolVariable

	poolVariable _ aPoolVariable! !

!PoolVarNode methodsFor: 'accessing' stamp: 'ajh 9/12/2002 12:07'!
canAssign

	^ poolVariable canAssign! !

!PoolVarNode methodsFor: 'accessing' stamp: 'ajh 9/12/2002 02:28'!
name

	^ poolVariable poolName! !

!PoolVarNode methodsFor: 'accessing' stamp: 'ajh 7/29/2002 10:35'!
poolVariable

	^ poolVariable! !

!PoolVarNode methodsFor: 'code generation' stamp: 'ajh 1/31/2003 21:40'!
emit: method assignment: assignmentNode forValue: bool

	assignmentNode valueNode emitValue: method.
	method mapTo: assignmentNode.
	method pushLiteral: poolVariable.
	method send: #setInPool:.
	bool ifFalse: [method doPop].
	method unMap.
! !

!PoolVarNode methodsFor: 'code generation' stamp: 'ajh 9/27/2002 00:49'!
emitValue: method

	| i |
	i _ method pushLiteral: poolVariable.
	method send: #poolValue.
	^ i! !

!PoolVarNode methodsFor: 'evaluating' stamp: 'ajh 6/10/2002 14:31'!
evaluate

	^ poolVariable value! !

!PoolVarNode methodsFor: 'evaluating' stamp: 'ajh 6/4/2002 13:41'!
evaluateAssignTo: object

	^ poolVariable value: object! !

!PoolVarNode methodsFor: 'comparing' stamp: 'ajh 7/29/2002 10:34'!
= other

	^ other class == self class and: [other poolVariable = poolVariable]! !

!PoolVarNode methodsFor: 'comparing' stamp: 'ajh 10/3/2002 23:52'!
hash

	^ poolVariable hash! !


!TempVarNode methodsFor: 'initializing' stamp: 'ajh 6/10/2002 08:40'!
name: string

	name _ string! !

!TempVarNode methodsFor: 'initializing' stamp: 'ajh 10/3/2002 17:20'!
owningBlockNode: blockNode

	owningBlockNode _ blockNode! !

!TempVarNode methodsFor: 'accessing' stamp: 'ajh 2/8/2003 23:23'!
instVarNode: varName at: index

	"self isReceiverVariable ifFalse: [self error: 'self must be receiver var']."
	^ InstVarNode new
		name: varName;
		instVar: index;
		selfVar: self
! !

!TempVarNode methodsFor: 'accessing' stamp: 'ajh 9/13/2002 13:42'!
name

	^ name ifNil: ['t', self identityHash printString]! !

!TempVarNode methodsFor: 'accessing' stamp: 'ajh 10/3/2002 17:20'!
owningBlockNode

	^ owningBlockNode! !

!TempVarNode methodsFor: 'accessing' stamp: 'ajh 1/15/2003 20:08'!
slotOf: var

	^ owningBlockNode indexOfClosureVar: var! !

!TempVarNode methodsFor: 'testing' stamp: 'ajh 9/12/2002 00:35'!
canAssign

	self isReceiverVariable ifTrue: [^ false].
	self isThisContextVariable ifTrue: [^ false].
	^ true! !

!TempVarNode methodsFor: 'testing' stamp: 'ajh 10/2/2002 01:22'!
isOtherTemp

	^ owningBlockNode otherTempVars identityIncludes: self! !

!TempVarNode methodsFor: 'testing' stamp: 'ajh 9/2/2002 23:56'!
isRarg
	"Is receiver or arg"

	^ owningBlockNode isRarg: self! !

!TempVarNode methodsFor: 'testing' stamp: 'ajh 10/3/2002 23:48'!
isReceiverVariable

	^ owningBlockNode receiverTempVar == self! !

!TempVarNode methodsFor: 'testing' stamp: 'ajh 7/29/2002 13:27'!
isSelfPseudoVariable

	^ self isReceiverVariable and: [name = 'self']! !

!TempVarNode methodsFor: 'testing' stamp: 'ajh 7/29/2002 10:42'!
isTemp

	^ true! !

!TempVarNode methodsFor: 'testing' stamp: 'ajh 10/3/2002 23:48'!
isThisContextVariable

	^ owningBlockNode thisContextTempVar == self! !

!TempVarNode methodsFor: 'var usage' stamp: 'ajh 10/9/2002 11:43'!
isIndirect

	^ (self varUseIfNone: [^ false]) isIndirect! !

!TempVarNode methodsFor: 'var usage' stamp: 'ajh 10/9/2002 11:43'!
isUsed

	^ (self varUseIfNone: [^ false]) isUsed! !

!TempVarNode methodsFor: 'var usage' stamp: 'ajh 10/3/2002 11:12'!
usage: usage nowHasDef: assignmentNode

	usage var: self nowHasDef: assignmentNode! !

!TempVarNode methodsFor: 'var usage' stamp: 'ajh 10/3/2002 11:11'!
usage: usage nowHasRef: referenceNode

	usage var: self nowHasRef: referenceNode! !

!TempVarNode methodsFor: 'var usage' stamp: 'ajh 10/2/2002 00:26'!
varUse

	^ owningBlockNode varUsage atVar: self! !

!TempVarNode methodsFor: 'var usage' stamp: 'ajh 2/9/2003 20:41'!
varUseIfNone: noneBlock

	^ (owningBlockNode ifNil: [^ noneBlock value]) varUsage
		atVar: self ifAbsent: noneBlock

	"owningBlockNode is nil only when decompiling block methods without its home method and it has captured vars from it.  See BytecodeDecompiler capturedVarAt:"! !

!TempVarNode methodsFor: 'code generation' stamp: 'ajh 1/31/2003 21:39'!
emit: method assignment: assignmentNode forValue: bool

	assignmentNode valueNode emitValue: method.
	method mapTo: assignmentNode.
	self emitStore: method.
	bool ifFalse: [method doPop].
	method unMap.
! !

!TempVarNode methodsFor: 'code generation' stamp: 'ajh 1/15/2003 20:21'!
emitDirectValue: method
	"Push value to be captured by embedded block"

	(method isLocalTemp: self) ifTrue: [
		method pushTemp: self.
		self isThisContextVariable ifTrue: [
			"Push contextTag to be captured"
			method send: #contextTag].
	] ifFalse: [
		method pushTemp: method receiverArg.
		method getField: (method receiverArg slotOf: self).
	]! !

!TempVarNode methodsFor: 'code generation' stamp: 'ajh 10/2/2002 01:55'!
emitDirectValue: method reference: referenceNode

	^ self emitDirectValue: method! !

!TempVarNode methodsFor: 'code generation' stamp: 'ajh 1/15/2003 20:33'!
emitStore: method

	(method isLocalTemp: self) ifFalse: [  "self is a closure var"
		method receiverArg emitValue: method.
		^ self isIndirect
			ifTrue: [method getField: (method receiverArg slotOf: self).
					method send: #storeInTempHolder:]
			ifFalse: [method setField: (method receiverArg slotOf: self)].
	].

	"self is a local var"
	self isIndirect
		ifTrue: [method pushTemp: self.
				method send: #storeInTempHolder:]
		ifFalse: [method storeTemp: self].
! !

!TempVarNode methodsFor: 'code generation' stamp: 'ajh 1/15/2003 20:30'!
emitValue: method

	(method isLocalTemp: self) ifTrue: [
		"self is a local var"
		method pushTemp: self.
	] ifFalse: [
		"self is a closure var"
		method receiverArg emitValue: method.
		method getField: (method receiverArg slotOf: self).
	].
	self isIndirect ifTrue: [method send: #tempHolderValue].
! !

!TempVarNode methodsFor: 'evaluating' stamp: 'ajh 6/10/2002 14:31'!
evaluate

	^ thisContext env at: self ifAbsent: [nil]! !

!TempVarNode methodsFor: 'evaluating' stamp: 'ajh 6/4/2002 16:35'!
evaluateAssignTo: object

	^ thisContext env at: self put: object! !

!TempVarNode methodsFor: 'printing' stamp: 'ajh 9/13/2002 13:42'!
printOn: aStream indent: level

	aStream withStyleFor: #temporaryVariable
			do: [aStream nextPutAll: self name]! !

!TempVarNode methodsFor: 'decompiling' stamp: 'ajh 8/2/2002 22:25'!
fieldOffset
	"Special for Decompiler2 only.  Assumes name is 't1', or 't2', etc, and the index in the name means the position of the temp in the stack"

	^ self name allButFirst asNumber - 1! !

!TempVarNode methodsFor: 'decompiling' stamp: 'ajh 7/29/2002 11:53'!
isImplicit

	^ (owningBlockNode otherTempVars includes: self) not! !

!TempVarNode methodsFor: 'decompiling' stamp: 'ajh 9/30/2002 13:30'!
removeTempFromBlock

	owningBlockNode otherTempVars remove: self ifAbsent: [^ false "already removed"].
	^ true! !

!TempVarNode methodsFor: 'decompiling' stamp: 'ajh 1/31/2003 13:42'!
scopeIn: decompiler
	"For decompiling"

	^ decompiler scopeNumberOfVar: self! !

!TempVarNode methodsFor: 'decompiling' stamp: 'ajh 1/31/2003 13:45'!
scopeIn: decompiler set: level
	"For decompiling old methods.  Note scope of temporary variables.
	Currently only the following distinctions are made:
		0	outer level: args and user-declared temps
		1	block args and doLimiT temps
		-1	a block temp that is no longer active
		-2	a block temp that held limit of to:do:"

	decompiler scopeNumberOfVar: self is: level! !


!TempVarNode class methodsFor: 'as yet unclassified' stamp: 'ajh 2/8/2003 18:55'!
dummy
	"Use as place holder to be substituted for before generating IR code.  Test for its existence by sending isDummyTemp to any parseNode, and this dummy reference node will respond true"

	^ DummyTempRefNode new var: (self new name: '#dummy')! !


!VarReferenceNode methodsFor: 'initializing' stamp: 'ajh 2/8/2003 18:54'!
var: aVarNode

	var _ aVarNode.
! !

!VarReferenceNode methodsFor: 'accessing' stamp: 'ajh 2/8/2003 18:53'!
var

	^ var! !

!VarReferenceNode methodsFor: 'testing' stamp: 'ajh 6/18/2002 14:19'!
isLeafNode

	^ true! !

!VarReferenceNode methodsFor: 'testing' stamp: 'ajh 2/8/2003 18:53'!
isSelfPseudoVariable

	^ var isSelfPseudoVariable! !

!VarReferenceNode methodsFor: 'var usage' stamp: 'ajh 2/8/2003 18:53'!
calcVarUsage: usage

	var usage: usage nowHasRef: self! !

!VarReferenceNode methodsFor: 'code generation' stamp: 'ajh 2/8/2003 18:53'!
emit: method forValue: bool

	method mapTo: self.
	bool ifTrue: [var emitValue: method reference: self].
	method unMap.
! !

!VarReferenceNode methodsFor: 'evaluating' stamp: 'ajh 2/8/2003 18:53'!
evaluate

	^ var evaluate! !

!VarReferenceNode methodsFor: 'printing' stamp: 'ajh 2/8/2003 18:53'!
printOn: aStream indent: level

	self startSourceRange: aStream position.
	var printOn: aStream indent: level.
	self endSourceRange: aStream position.
! !

!VarReferenceNode methodsFor: 'comparing' stamp: 'ajh 2/8/2003 18:53'!
= other

	^ other class == self class and: [other var = var]! !

!VarReferenceNode methodsFor: 'comparing' stamp: 'ajh 2/8/2003 18:52'!
hash

	^ var hash! !

!VarReferenceNode methodsFor: 'decompiling' stamp: 'ajh 2/8/2003 18:55'!
asSuperNode: behavior

	var isSelfPseudoVariable ifFalse: [self error: 'expected self for super'].
	^ SuperNode new
		var: var;
		varClass: behavior;
		sourceRange: sourceRange;
		bytecodeRange: bytecodeRange;
		yourself! !

!VarReferenceNode methodsFor: 'decompiling' stamp: 'ajh 2/8/2003 18:53'!
scopeIn: decompiler set: level

	var scopeIn: decompiler set: level! !

!VarReferenceNode methodsFor: 'C translation' stamp: 'ajh 2/8/2003 18:53'!
asTranslatorNode

	^ TVariableNode new setName: var name
! !


!DummyTempRefNode methodsFor: 'as yet unclassified' stamp: 'ajh 10/17/2002 18:15'!
calcVarUsage: usage
	"Do nothing"! !

!DummyTempRefNode methodsFor: 'as yet unclassified' stamp: 'ajh 10/17/2002 18:12'!
isDummyTemp
	"see senders and TempVarNode-dummy and its sender"

	^ true! !


!SuperNode methodsFor: 'as yet unclassified' stamp: 'ajh 6/18/2002 12:11'!
isSuperNode

	^ true! !

!SuperNode methodsFor: 'as yet unclassified' stamp: 'ajh 8/12/2002 14:43'!
printOn: aStream indent: level

	aStream withStyleFor: #variable
		do: [aStream nextPutAll: 'super'].
! !

!SuperNode methodsFor: 'as yet unclassified' stamp: 'ajh 9/27/2002 09:13'!
varClass

	^ varClass! !

!SuperNode methodsFor: 'as yet unclassified' stamp: 'ajh 9/27/2002 09:13'!
varClass: behavior

	varClass _ behavior! !


!VarUsageSet methodsFor: 'adding' stamp: 'ajh 10/3/2002 12:57'!
addClosureUsage: usage

	usage do: [:varUse |
		(self atVar: varUse var)
			add: varUse closureUseState
			from: usage blockNode
			inStatement: currentBlockStatement
	].
! !

!VarUsageSet methodsFor: 'adding' stamp: 'ajh 10/3/2002 14:39'!
addUsage: usage

	usage do: [:varUse |
		(self atVar: varUse var)
			addUse: varUse
			inStatement: currentBlockStatement
	].
! !

!VarUsageSet methodsFor: 'adding' stamp: 'ajh 10/3/2002 11:00'!
currentBlockStatement: parseNode
	"parseNode is a top-level statement node in my block node"

	currentBlockStatement _ parseNode! !

!VarUsageSet methodsFor: 'adding' stamp: 'ajh 10/3/2002 12:57'!
var: var nowHasDef: node

	(self atVar: var)
		add: #write
		from: node
		inStatement: currentBlockStatement! !

!VarUsageSet methodsFor: 'adding' stamp: 'ajh 2/1/2003 01:39'!
var: var nowHasRef: node

	| varUse |
	varUse _ self atVar: var.

	"Check for read before write"
	( varUse useState == #none and:
	  [(self isLocalVar: var) and: 
	  [var isOtherTemp]]
	) ifTrue: [UndefinedVarSignal new node: node; signal].

	varUse
		add: #read
		from: node
		inStatement: currentBlockStatement! !

!VarUsageSet methodsFor: 'merging' stamp: 'ajh 10/3/2002 13:40'!
asCapturedVarUsage

	| set |
	set _ VarUsageSet new: self size.
	self do: [:varUse | | copy |
		copy _ varUse copy.
		copy userNode: blockNode.
		set add: copy.
	].
	^ set! !

!VarUsageSet methodsFor: 'merging' stamp: 'ajh 10/3/2002 13:48'!
calcVarUsageForCase: caseActionBlock rest: otherCases otherwise: otherwiseBlock

	| trueBranch case falseBranch |
	trueBranch _ caseActionBlock calcCapturedVarUsage.
	otherCases isEmpty ifTrue: [
		falseBranch _ otherwiseBlock calcCapturedVarUsage.
	] ifFalse: [
		case _ otherCases removeFirst.
		falseBranch _ case receiverNode calcCapturedVarUsage.
		falseBranch
			calcVarUsageForCase: case argumentNodes first
			rest: otherCases
			otherwise: otherwiseBlock.
	].
	^ self addUsage: (trueBranch mergeUsage: falseBranch)! !

!VarUsageSet methodsFor: 'merging' stamp: 'ajh 10/27/2002 16:12'!
mergeUsage: usage

	usage do: [:varUse |
		(self atVar: varUse var) mergeUse: varUse.
	].
	self do: [:varUse |
		(usage includesVar: varUse var)
			ifFalse: [varUse mergeUse: VarUse new].
	].
! !

!VarUsageSet methodsFor: 'merging' stamp: 'ajh 10/3/2002 12:41'!
withEmptyBranch

	^ self mergeUsage: VarUsageSet new! !

!VarUsageSet methodsFor: 'accessing' stamp: 'ajh 10/27/2002 16:32'!
atVar: varKey
	"Return VarUse for varKey, create one if necessary"

	^ self detect: [:varUse | varUse var = varKey]
		ifNone: [self add: (VarUse new var: varKey)]! !

!VarUsageSet methodsFor: 'accessing' stamp: 'ajh 10/27/2002 16:33'!
atVar: varKey ifAbsent: block

	^ self detect: [:varUse | varUse var = varKey]
		ifNone: block! !

!VarUsageSet methodsFor: 'accessing' stamp: 'ajh 10/27/2002 16:33'!
includesVar: varKey

	^ self anySatisfy: [:varUse | varUse var = varKey]! !

!VarUsageSet methodsFor: 'block' stamp: 'ajh 10/3/2002 10:59'!
blockNode

	^ blockNode! !

!VarUsageSet methodsFor: 'block' stamp: 'ajh 10/3/2002 10:59'!
blockNode: aBlockNode
	"This is the blockNode this varUsage is calculated for"

	blockNode _ aBlockNode! !

!VarUsageSet methodsFor: 'block' stamp: 'ajh 10/4/2002 16:38'!
freeVarUsage

	| freeVarUsage |
	freeVarUsage _ VarUsageSet new blockNode: blockNode.
	self do: [:varUse |
		(self isLocalVar: varUse var) ifFalse: [freeVarUsage add: varUse].
	].
	^ freeVarUsage! !

!VarUsageSet methodsFor: 'block' stamp: 'ajh 10/27/2002 16:35'!
freeVars

	| vars |
	vars _ OrderedCollection new.
	self do: [:varUse |
		(self isLocalVar: varUse var) ifFalse: [vars add: varUse var].
	].
	^ vars! !

!VarUsageSet methodsFor: 'block' stamp: 'ajh 10/3/2002 11:01'!
isLocalVar: var

	^ var owningBlockNode == blockNode! !

!VarUsageSet methodsFor: 'block' stamp: 'ajh 10/27/2002 16:36'!
localVars
	"Temps never used won't be in here"

	| vars |
	vars _ OrderedCollection new.
	self do: [:varUse |
		(self isLocalVar: varUse var) ifTrue: [vars add: varUse var].
	].
	^ vars! !

!VarUsageSet methodsFor: 'block' stamp: 'ajh 10/9/2002 11:45'!
localVarsNonArgs

	^ self localVars select: [:varNode | varNode isOtherTemp]! !

!VarUsageSet methodsFor: 'finish block' stamp: 'ajh 10/4/2002 00:17'!
localFinished
	"We finished building var usage for closureNode.  May want to do some checks"

	currentBlockStatement _ nil.
	self removeDeadTemps.
! !

!VarUsageSet methodsFor: 'finish block' stamp: 'ajh 10/2/2002 01:09'!
removeDeadTemps
	"Remove unused temps and move temps only used by a single inner block to that block"! !


!VarUse methodsFor: 'accessing' stamp: 'ajh 10/1/2002 20:04'!
firstUse

	^ firstUse! !

!VarUse methodsFor: 'accessing' stamp: 'ajh 10/1/2002 20:04'!
firstUseStatement

	^ firstUseStatement! !

!VarUse methodsFor: 'accessing' stamp: 'ajh 10/1/2002 20:31'!
key
	"For KeyedSets"

	^ var! !

!VarUse methodsFor: 'accessing' stamp: 'ajh 10/1/2002 20:04'!
lastUse

	^ lastUse! !

!VarUse methodsFor: 'accessing' stamp: 'ajh 10/3/2002 11:32'!
lastUseStatement

	^ lastUseStatement! !

!VarUse methodsFor: 'accessing' stamp: 'ajh 10/3/2002 19:52'!
printOn: stream

	var printOn: stream! !

!VarUse methodsFor: 'accessing' stamp: 'ajh 10/1/2002 20:32'!
var

	^ var! !

!VarUse methodsFor: 'testing' stamp: 'ajh 10/1/2002 20:04'!
isDirect

	^ self isIndirect not! !

!VarUse methodsFor: 'testing' stamp: 'ajh 10/3/2002 14:15'!
isFirstUseTopLevelStore
	"Is firstUse a top-level statement (equals firstUseStatement) and is it a VarAssignment.
	 If it is then the temp can be allocated at this assignment instead of in the beginning of the block"

	^ firstUse == firstUseStatement and: [firstUse isMemberOf: VarAssignmentNode]! !

!VarUse methodsFor: 'testing' stamp: 'ajh 10/1/2002 20:04'!
isIndirect
	"Is this var shared between the home context and block closure(s)"

	^ self useState == #indirect! !

!VarUse methodsFor: 'testing' stamp: 'ajh 10/2/2002 01:00'!
isUnassigned

	^ self useState = #none or: [self useState = #read]! !

!VarUse methodsFor: 'testing' stamp: 'ajh 10/2/2002 00:35'!
isUnused

	^ self useState = #none! !

!VarUse methodsFor: 'testing' stamp: 'ajh 10/2/2002 00:34'!
isUsed

	^ self useState ~= #none! !

!VarUse methodsFor: 'private' stamp: 'ajh 10/1/2002 20:54'!
closureUseState
	"Return #readBlock, #writeBlock, or #indirect"

	(#(read readBlock) identityIncludes: useState) ifTrue: [^ #readBlock].
	(#(write wrRdBlock) identityIncludes: useState) ifTrue: [^ #writeBlock].
	(#(writeBlock indirect)  identityIncludes: useState) ifTrue: [^ #indirect].
	self error: useState printString, ', unexpected closureVar useState'
! !

!VarUse methodsFor: 'private' stamp: 'ajh 10/1/2002 20:19'!
initialize

	useState _ #none.
! !

!VarUse methodsFor: 'private' stamp: 'ajh 10/1/2002 20:04'!
useState

	^ useState! !

!VarUse methodsFor: 'private' stamp: 'ajh 10/1/2002 20:04'!
useState: symbol
	"symbol must be one of #(none read readBlock write wrRdBlock writeBlock indirect)"

	useState _ symbol! !

!VarUse methodsFor: 'private' stamp: 'ajh 10/1/2002 20:31'!
var: key

	var _ key! !

!VarUse methodsFor: 'adding' stamp: 'ajh 10/3/2002 12:57'!
add: useSymbol from: node inStatement: rootNode
	"useSymbol must be one of
	#(read write readBlock wrRdBlock writeBlock indirect)"

	firstUse ifNil: [
		firstUse _ node.
		firstUseStatement _ rootNode.
	].
	useState _ AddUseAutomaton state: useState transition: useSymbol.
	lastUse _ node.
	lastUseStatement _ rootNode.
! !

!VarUse methodsFor: 'adding' stamp: 'ajh 10/3/2002 14:39'!
addUse: varUse inStatement: rootNode

	firstUse ifNil: [
		firstUse _ varUse firstUse.
		firstUseStatement _ rootNode.
	].
	useState _ AddUseAutomaton state: useState transition: varUse useState.
	lastUse _ varUse lastUse.
	lastUseStatement _ rootNode.
! !

!VarUse methodsFor: 'merging' stamp: 'ajh 10/3/2002 20:01'!
mergeUse: varUse

	| noUse otherNoUse |
	noUse _ Object new.  "unique object"
	otherNoUse _ Object new.
	firstUse _ (firstUse ifNil: [noUse]) inCollection
			, (varUse firstUse ifNil: [otherNoUse]) inCollection.
	useState _ MergeUseAutomaton state: useState transition: varUse useState.
	lastUse _ (lastUse ifNil: [noUse]) inCollection
			, (varUse lastUse ifNil: [otherNoUse]) inCollection.
! !

!VarUse methodsFor: 'merging' stamp: 'ajh 10/3/2002 13:41'!
userNode: node

	firstUse _ node.
	firstUseStatement _ nil.
	lastUse _ node.
	lastUseStatement _ nil.
! !


!VarUse class methodsFor: 'as yet unclassified' stamp: 'ajh 10/1/2002 20:04'!
initialize

	self initializeAutomata.
! !

!VarUse class methodsFor: 'as yet unclassified' stamp: 'ajh 10/1/2002 20:10'!
initializeAutomata
	"VarUse initializeAutomata"

 AddUseAutomaton _ FiniteAutomaton new
	  actions:	#(none		read		readBlock	write		wrRdBlock	writeBlock	indirect)
	  transitions: #(
"from state"		"to states"
(none			(none		read		readBlock	write		wrRdBlock	writeBlock	indirect))
(read			(read		read		readBlock	write		wrRdBlock	writeBlock	indirect))
(readBlock		(readBlock	readBlock	readBlock	indirect		indirect		indirect		indirect))
(write			(write		write		wrRdBlock	write		wrRdBlock	writeBlock	indirect))
(wrRdBlock		(wrRdBlock	wrRdBlock	wrRdBlock	indirect		indirect		indirect		indirect))
(writeBlock		(writeBlock	indirect		indirect		indirect		indirect		indirect		indirect))
(indirect		(indirect	indirect		indirect		indirect		indirect		indirect		indirect))
).

 MergeUseAutomaton _ FiniteAutomaton new
	  actions:	#(none		read		readBlock	write		wrRdBlock	writeBlock	indirect)
	  transitions: #(
"from state"		"to states"
(none			(none		read		readBlock	write		wrRdBlock	writeBlock	indirect))
(read			(read		read		readBlock	write		wrRdBlock	writeBlock	indirect))
(readBlock		(readBlock	readBlock	readBlock	wrRdBlock	wrRdBlock	writeBlock	indirect))
(write			(write		write		wrRdBlock	write		wrRdBlock	writeBlock	indirect))
(wrRdBlock		(wrRdBlock	wrRdBlock	wrRdBlock	wrRdBlock	wrRdBlock	writeBlock	indirect))
(writeBlock		(writeBlock	writeBlock	writeBlock	writeBlock	writeBlock	writeBlock	indirect))
(indirect		(indirect	indirect		indirect		indirect		indirect		indirect		indirect))
).
! !

!VarUse class methodsFor: 'as yet unclassified' stamp: 'ajh 10/1/2002 20:20'!
new

	^ super new initialize! !


!VariableNode methodsFor: 'accessing' stamp: 'ajh 8/2/2002 22:15'!
referenceNode
	"Compatible with new ParseNodes.  Needed for Decompiler>>convertToDoLoop"! !

VarUse initialize!
MethodContext removeSelector: #tempAt:!
MethodContext removeSelector: #tempAt:put:!
Decompiler2 removeSelector: #lastBytecodeRange!

!Decompiler2 reorganize!
('initialize' asBlock asBlock: initialize keepBasicConstructs outer:)
('var scope' scopeNumberOfVar: scopeNumberOfVar:is:)
('control' checkForBlock: convertToDoLoop jump:if:)
('private' lastTempVarIndex method methodNode outer outermost privNewConstructor)
('instruction decoding' doStore: popIntoLiteralVariable: pushLiteralVariable: storeIntoLiteralVariable:)
!

BlockContext removeSelector: #tempAt:!
BlockContext removeSelector: #tempAt:put:!
BytecodeV3WriteStream initialize!
BlockClosure initialize!
